(* Copyright (C) 2014, Daniel Wyckoff*)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(* Version 0.3 bool2 -- certified -- Coq 8.4 pl4 *)

Require Export BoolAlgBasics.
Require Import InfiniteOperations.
Require Import SetUtilities.
Require Import TypeUtilities.
Require Export Powerset.
Require Import ProofIrrelevanceFacts.
Require Import ProofIrrelevance.
Require Import Description.
Require Import QArith.
Require Import ArithUtilities.
Require Import Equality.
Require Import DecidableDec.
Require Import LogicUtilities.
Require Import FunctionProperties.
Require Import NaryFunctions.


Section power_set_sec.
Variable Xt : Type.
Definition X := Full_set Xt.
Definition PX := Power_set _ X.

(* PSA = Power Set Algebra. *)
(* This definition just builds the Bconst component of the 
   Boolean Algebra.*)
Definition bc_psa := 
  Build_Bconst (Ensemble Xt) PX (@Union Xt) (@Intersection Xt) 
               X (Empty_set Xt) (@Comp Xt).


Lemma full_power_compat : Full_set (Ensemble Xt) = PX.
assert (h1:forall Y:(Ensemble Xt), Included Y X -> 
    In PX Y).
  unfold PX.
  apply Definition_of_Power_set.
assert (h2:forall Y:(Ensemble Xt), Included Y X).
  intro Y. unfold X.
  unfold Included.
  intros x h.
  apply (Full_intro Xt).
assert (h3:forall Y:Ensemble Xt, In PX Y).
  intro Y. apply h1.  apply h2.
apply Extensionality_Ensembles.
unfold Same_set.
split.
unfold Included.
intros S h4.
apply (h3 S).
unfold Included.
intros S h5.
apply Full_intro.
Qed.

Lemma full_power_compat' : BS bc_psa = Full_set (Btype bc_psa).
assert (h1: Btype bc_psa = Ensemble Xt).
unfold bc_psa. reflexivity. 
assert (h2: BS bc_psa = PX).
unfold bc_psa. reflexivity.
rewrite h2. 
assert (h3:Full_set (Btype bc_psa) = Full_set (Ensemble Xt)).
unfold bc_psa. 
reflexivity.
rewrite h3.
rewrite full_power_compat. reflexivity.
Qed.

Definition psa := Build_Bool_Alg bc_psa full_power_compat'
  (@Union_associative Xt) (@assoc_prod_psa Xt) (@Union_commutative Xt)
  (@Intersection_commutative Xt) (@abs_sum_psa Xt) (@abs_prod_psa Xt)
  (@Distributivity Xt) (@dist_prod_psa Xt) (@comp_sum_psa Xt) (@comp_prod_psa Xt).


Lemma btype_bc_psa :  bt psa = Ensemble Xt.
simpl. reflexivity.
Qed.

Definition transfer_psa (P:bt psa) : Ensemble Xt :=
  transfer btype_bc_psa P.

End power_set_sec.



Section field_of_sets_sec.

Record Field_of_Sets :=
{ Xt:Type;
  F:(Ensemble (Ensemble Xt));
  non_empty_F : (exists S:(Ensemble Xt), In F S);
  Union_closed : forall S1 S2:(Ensemble Xt), 
    In F S1 -> In F S2 -> In F (Union S1 S2);
  Int_closed : forall S1 S2:(Ensemble Xt),
    In F S1 -> In F S2 -> In F (Intersection S1 S2);
  Comp_closed : forall S :(Ensemble Xt),
    In F S -> In F (Comp S);
  Ft := {S:(Ensemble Xt) | In F S};
  Union_fos := 
    fun (S1 S2:Ft) => (let S1' := proj1_sig S1 in let S2' := proj1_sig S2 in 
      exist (fun (S':(Ensemble Xt)) => (In F S'))
        (Union S1' S2') (Union_closed S1' S2' (proj2_sig S1) (proj2_sig S2)));
  Int_fos := 
    fun (S1 S2:Ft) => (let S1' := proj1_sig S1 in let S2' := proj1_sig S2 in 
      exist (fun (S':(Ensemble Xt)) => (In F S'))
        (Intersection S1' S2') (Int_closed S1' S2' (proj2_sig S1) (proj2_sig S2)));
  Comp_fos := 
    fun (S:Ft) => (let S' := proj1_sig S in  
      exist (fun (S'':(Ensemble Xt)) => (In F S''))
        (Comp S') (Comp_closed S' (proj2_sig S)))
      }.


Lemma f_fos_inj : 
  forall (fos fos':Field_of_Sets)
         (pf:Xt fos = Xt fos'),
    (@transfer_dep _ (fun T=>Ensemble (Ensemble T)) _ _ pf (F fos)) = F fos' -> fos = fos'.
intros fos fos' h1 h2.
destruct fos; destruct fos'.
simpl in h1. subst.
simpl in h2.
subst.
f_equal; apply proof_irrelevance.
Qed.


Variable fos:Field_of_Sets.

Lemma full_in_F : In (F fos) (Full_set (Xt fos)).
elim (non_empty_F fos).
intros S h1.
assert (h2: In (F fos) (Comp S)).
apply ((Comp_closed fos) S h1).
assert (h3: In (F fos) (Union S (Comp S))).
apply ((Union_closed fos) S (Comp S) h1 h2).
assert (h4: 
  (Union S (Comp S)) = 
    (Full_set (Xt fos))).
apply comp_sum_psa.
rewrite h4 in h3. assumption.
Qed.

Lemma empty_in_F : In (F fos) (Empty_set (Xt fos)).
elim (non_empty_F fos).
intros S h1.
assert (h2: In (F fos) (Comp S)).
apply ((Comp_closed fos) S h1).
assert (h3: In (F fos) (Intersection S (Comp S))).
apply ((Int_closed fos) S (Comp S) h1 h2).
assert (h4:  (Intersection  S (Comp S)) = 
    (Empty_set (Xt fos))).
apply comp_prod_psa.
rewrite h4 in h3. assumption.
Qed.

Lemma assoc_sum_fos : forall N M P:(Ft fos), (Union_fos fos N (Union_fos fos M P)) =
  (Union_fos fos (Union_fos fos N M) P).
intros N M P.
unfold Union_fos; simpl.
assert (h:Union (proj1_sig N) (Union (proj1_sig M) (proj1_sig P)) =
        Union (Union (proj1_sig N) (proj1_sig M)) (proj1_sig P)).
  apply Union_associative.
apply existTexist.
apply subsetT_eq_compat.
apply h.
Qed.

Lemma assoc_prod_fos : forall N M P:(Ft fos), (Int_fos fos N (Int_fos fos M P)) =
  (Int_fos fos (Int_fos fos N M) P).
intros N M P.
unfold Int_fos; simpl.
assert (h:(Intersection (proj1_sig N) 
  (Intersection (proj1_sig M) (proj1_sig P))) = 
  (Intersection (Intersection (proj1_sig N) (proj1_sig M))
     (proj1_sig P))).
  apply assoc_prod_psa.
apply existTexist.
apply subsetT_eq_compat.
apply h.
Qed.

Lemma comm_sum_fos : forall N M:(Ft fos), Union_fos fos N M = Union_fos fos M N.
intros N M.
unfold Union_fos; simpl.
assert (h: (Union (proj1_sig N) (proj1_sig M)) = 
           (Union (proj1_sig M) (proj1_sig N))).
  apply Union_commutative.
apply existTexist.
apply subsetT_eq_compat.
apply h.
Qed.

Lemma comm_prod_fos : forall N M:(Ft fos), Int_fos fos N M = Int_fos fos M N.
intros N M.
unfold Int_fos; simpl.
assert (h: (Intersection (proj1_sig N) (proj1_sig M)) = 
           (Intersection (proj1_sig M) (proj1_sig N))).
  apply Intersection_commutative.
apply existTexist.
apply subsetT_eq_compat.
apply h.
Qed.

Lemma abs_sum_fos  :
  forall N M:(Ft fos), Union_fos fos N (Int_fos fos N M) = N.
intros N M.
unfold Union_fos; unfold Int_fos; simpl.
assert (h1: (Union (proj1_sig N)
     (Intersection (proj1_sig N) (proj1_sig M))) = (proj1_sig N)).
  apply abs_sum_psa.
assert (h2: N = 
  exist (fun S' : Ensemble (Xt fos) => In (F fos) S')
    (proj1_sig N) (proj2_sig N)).
  destruct N; simpl.
  reflexivity.
rewrite h2 at 8.
apply existTexist.
apply subsetT_eq_compat.
apply h1.
Qed.

Lemma abs_prod_fos  :
  forall N M:(Ft fos), Int_fos fos N (Union_fos fos N M) = N.
intros N M.
unfold Union_fos; unfold Int_fos; simpl.
assert (h1: (Intersection (proj1_sig N)
     (Union (proj1_sig N) (proj1_sig M))) = (proj1_sig N)).
  apply abs_prod_psa.
assert (h2: N = 
  exist (fun S' : Ensemble (Xt fos) => In (F fos) S')
    (proj1_sig N) (proj2_sig N)).
  destruct N; simpl.
  reflexivity.
rewrite h2 at 8.
apply existTexist.
apply subsetT_eq_compat.
apply h1.
Qed.

Lemma dist_sum_fos :
  forall N M P:(Ft fos),
    Int_fos fos P (Union_fos fos N M) =
    Union_fos fos (Int_fos fos P N) (Int_fos fos P M).
intros N M P.
unfold Union_fos; unfold Int_fos; simpl.
assert (h: 
  (Intersection (proj1_sig P)
     (Union  (proj1_sig N) (proj1_sig M))) = 
  (Union (Intersection (proj1_sig P) (proj1_sig N))
     (Intersection (proj1_sig P) (proj1_sig M)))).
  apply Distributivity.
apply existTexist.
apply subsetT_eq_compat.
apply h.
Qed.

Lemma dist_prod_fos :
  forall N M P:(Ft fos),
    Union_fos fos P (Int_fos fos N M) =
    Int_fos fos (Union_fos fos P N) (Union_fos fos P M).
intros N M P.
unfold Union_fos; unfold Int_fos; simpl.
assert (h: 
  (Union (proj1_sig P)
     (Intersection  (proj1_sig N) (proj1_sig M))) = 
  (Intersection (Union (proj1_sig P) (proj1_sig N))
     (Union (proj1_sig P) (proj1_sig M)))).
  apply dist_prod_psa.
apply existTexist.
apply subsetT_eq_compat.
apply h.
Qed.

Definition Full_fos := 
  exist (fun S':(Ensemble (Xt fos)) => (In (F fos) S')) (Full_set (Xt fos))
  full_in_F.

Lemma comp_sum_fos:
  forall N:(Ft fos), Union_fos fos N (Comp_fos fos N) = Full_fos.
intro N.
unfold Union_fos; unfold Comp_fos; unfold Full_fos; simpl.
assert (h1: 
  (Union (proj1_sig N) (Comp (proj1_sig N))) = 
    (Full_set (Xt fos))).
  apply comp_sum_psa.
apply existTexist.
apply subsetT_eq_compat.
assumption.
Qed.

Definition Empty_fos :=
  exist (fun S':(Ensemble (Xt fos)) => (In (F fos) S')) (Empty_set (Xt fos))
  empty_in_F.

Lemma comp_prod_fos:
  forall N:(Ft fos), Int_fos fos N (Comp_fos fos N) = Empty_fos.
intro N.
unfold Int_fos; unfold Comp_fos; unfold Empty_fos; simpl.
assert (h1: 
  (Intersection (proj1_sig N) (Comp (proj1_sig N))) = 
    (Empty_set (Xt fos))).
  apply comp_prod_psa.
apply existTexist.
apply subsetT_eq_compat.
assumption.
Qed.

Definition Bc_fos := Build_Bconst (Ft fos) (Full_set (Ft fos)) (Union_fos fos)
  (Int_fos fos) Full_fos Empty_fos (Comp_fos fos).

Lemma fos_full_compat : BS Bc_fos = Full_set (Btype Bc_fos).
unfold Bc_fos; simpl. reflexivity.
Qed.

Definition fos_ba := 
  Build_Bool_Alg Bc_fos fos_full_compat
    assoc_sum_fos assoc_prod_fos comm_sum_fos comm_prod_fos abs_sum_fos
    abs_prod_fos dist_sum_fos dist_prod_fos comp_sum_fos comp_prod_fos.

Lemma fos_psa_ex :
  forall (T:Type), exists! fos:Field_of_Sets, 
    (Xt fos) = T /\ (F fos) = power_set (Full_set (Xt fos)). 
intros T.
pose (power_set (Full_set T)) as F'.
assert (my_non_empty_F : exists S : Ensemble T, In F' S).
  exists (Empty_set _).
  constructor. auto with sets.

assert (my_Union_closed : forall S1 S2 : Ensemble T,
                   In F' S1 -> In F' S2 -> In F' (Union S1 S2)).
intros.
constructor. red. intros. constructor.

assert (my_Int_closed : forall S1 S2 : Ensemble T,
                 In F' S1 -> In F' S2 -> In F' (Intersection S1 S2)).
  intros. constructor. red. intros. constructor.
assert (my_Comp_closed : forall S : Ensemble T,
                  In F' S -> In F' (Comp S)).
  intros. constructor. red. intros. constructor.

pose (Build_Field_of_Sets T F' my_non_empty_F my_Union_closed my_Int_closed my_Comp_closed) as fos'.
exists fos'. simpl in fos'.
red; split; try split; simpl; try reflexivity.
intros B h1. 
destruct h1 as [h1 h2].
simpl.
assert (h3:T = Xt fos'). reflexivity. 
rewrite  h3 in h1.
symmetry in h1. 
assert (h5:@transfer_dep _ (fun T => Ensemble (Ensemble T)) (Xt fos') (Xt B) h1 (F fos') = F B). 
  simpl. unfold F'.    
  pose (fun t:{T:Type & (Ensemble (Ensemble T))} =>
          projT2 t = power_set (Full_set (projT1 t))) as P.
  pose proof (transfer_dep_prop _ _ h1 (F fos') P) as h4.
  unfold P in h4. simpl in h4.
  rewrite <- h2 in h4. unfold F' in h4.
  rewrite <- h4. reflexivity.
eapply f_fos_inj.
apply h5.
Qed.

Definition fos_psa (T:Type) := 
  proj1_sig (constructive_definite_description _ (fos_psa_ex T)).

Lemma fos_psa_compat : 
  forall (T:Type),
    let fos:=fos_psa T in
    (Xt fos) = T /\ (F fos) = power_set (Full_set (Xt fos)).
intros T Fos.
unfold Fos. unfold fos_psa.
destruct constructive_definite_description as [Fos' h1].
simpl.
assumption.
Qed.

Definition proj1_sig_fos_psa 
           {T:Type} (A:{S:Ensemble (Xt (fos_psa T)) |
                                 In (F (fos_psa T)) S}) : 
  Ensemble T.
destruct A as [A h1].
pose proof (fos_psa_compat T) as [h2 h3]. 
clear h1.
rewrite h2 in A.
refine A.
Defined.

Definition proj1_sig_fos_psa_compat : 
  forall {T:Type} (A:{S:Ensemble (Xt (fos_psa T)) |
                      In (F (fos_psa T)) S}),
    let A' := proj1_sig_fos_psa A in
    A' = transfer_dep   
                      match (fos_psa_compat T) with
                        | conj P _ => P 
                      end
                      (proj1_sig A). 
intros T A A'. unfold A'. 
symmetry. 
rewrite <- transfer_dep_eq_iff.
unfold proj1_sig_fos_psa.
destruct (fos_psa_compat T) as [h1 h2]. 
destruct A as [A h3].
simpl.
destruct h1.
simpl.
reflexivity.
Qed.

Lemma proj1_sig_fos_psa_compat' : 
  forall {T:Type} (A:{S:Ensemble (Xt (fos_psa T)) |
                      In (F (fos_psa T)) S}),
         existT Ensemble (Xt (fos_psa T)) (proj1_sig A) =
         existT Ensemble T (proj1_sig_fos_psa A).
intros T A.
pose proof (proj1_sig_fos_psa_compat A) as h1.
symmetry in h1.
rewrite <- transfer_dep_eq_iff in h1.
assumption.
Qed.


Definition exist_fos_psa 
           {T:Type} (A:Ensemble T) : 
  {S:Ensemble (Xt (fos_psa T)) |In (F (fos_psa T)) S}.
pose proof (fos_psa_compat T) as [h2 h3].
rewrite <- h2 in A.
pose proof (full_inclusion A) as h4.
rewrite h3.
assert (h5:In (power_set (Full_set (Xt (fos_psa T)))) A).
  constructor. red; intros; constructor.
refine (exist _ A h5).
Defined.

Lemma exist_fos_psa_compat : 
  forall {T:Type} (A:Ensemble T),
    let A' := exist_fos_psa A in
    proj1_sig A' = transfer_dep_r
                      match (fos_psa_compat T) with
                        | conj P _ => P 
                      end A.
intros T A A'.
destruct (fos_psa_compat T) as [h1 h2].
pose proof h1 as h1'. symmetry in h1'.
rewrite <- (transfer_dep_transfer_dep_r_compat h1' h1).
symmetry.
rewrite <- transfer_dep_eq_iff.
unfold A'.
unfold exist_fos_psa.
destruct (fos_psa_compat T) as [h3 h4]. unfold eq_rect_r.
rewrite h4. simpl.
rewrite h3. simpl.
reflexivity.
Qed.

Lemma exist_fos_psa_compat' : 
  forall {T:Type} (A:Ensemble T),
    existT Ensemble (Xt (fos_psa T)) (proj1_sig (exist_fos_psa A)) =
    existT Ensemble T A.
intros T A.
unfold exist_fos_psa.
destruct (fos_psa_compat T) as [h1 h2]. 
rewrite h2, h1.
unfold eq_rect_r.
rewrite <- eq_rect_eq.
simpl.
reflexivity.
Qed.

Lemma proj1_sig_exist_fos_psa_int : 
  forall {T:Type} (A B:Ensemble T),
    proj1_sig (exist_fos_psa (Intersection A B)) = 
    Intersection (proj1_sig (exist_fos_psa A))
                 (proj1_sig (exist_fos_psa B)).
intros T A B. 
unfold exist_fos_psa.
destruct (fos_psa_compat T) as [h1 h2].
unfold eq_rect_r.
rewrite h2.
do 3 rewrite <- eq_rect_eq.
simpl.
rewrite h1.
do 3 rewrite <- eq_rect_eq.
reflexivity.
Qed.

Lemma proj1_sig_exist_fos_psa_union : 
  forall {T:Type} (A B:Ensemble T),
    proj1_sig (exist_fos_psa (Union A B)) = 
    Union (proj1_sig (exist_fos_psa A))
                 (proj1_sig (exist_fos_psa B)).
intros T A B. 
unfold exist_fos_psa.
destruct (fos_psa_compat T) as [h1 h2].
unfold eq_rect_r.
rewrite h2.
do 3 rewrite <- eq_rect_eq.
simpl.
rewrite h1.
do 3 rewrite <- eq_rect_eq.
reflexivity.
Qed.

Lemma proj1_sig_exist_fos_psa_comp : 
  forall {T:Type} (A:Ensemble T),
    proj1_sig (exist_fos_psa (Comp A)) = 
    Comp (proj1_sig (exist_fos_psa A)).
intros T A.
unfold exist_fos_psa.
destruct (fos_psa_compat T) as [h1 h2].
unfold eq_rect_r.
rewrite h2.
do 2 rewrite <- eq_rect_eq.
simpl.
rewrite h1.
do 2 rewrite <- eq_rect_eq.
reflexivity.
Qed.


Lemma in_proj1_sig_exist_fos_psa_iff : 
  forall {T:Type} (A:Ensemble T) (x:Xt (fos_psa T)),
    In (proj1_sig (exist_fos_psa A)) 
       x <->
    In A (transfer (match (fos_psa_compat T)
           with | conj pf _ => pf end)
          x).
intros T A x.
destruct (fos_psa_compat T) as [h1l h1r].
rewrite <- (transfer_dep_undoes_transfer_dep_r h1l A) at 2.
rewrite <- transfer_in.
rewrite exist_fos_psa_compat.
destruct (fos_psa_compat T) as [h2 h3].
pose proof (proof_irrelevance _ h2 h1l) as h4; rewrite h4; tauto.
Qed.


Lemma in_proj1_sig_exist_fos_psa_iff_r : 
  forall {T:Type} (A:Ensemble T) (x:T),
    In (proj1_sig (exist_fos_psa A)) 
       (transfer_r 
          (match (fos_psa_compat T)
           with | conj pf _ => pf end)
          x) <->
    In A x.
intros T A x.
pose proof (transfer_in_r (match (fos_psa_compat T) with
                             |conj pf _ => pf end) A x) as h1.
rewrite h1.  
pose proof (exist_fos_psa_compat A) as h3.
rewrite h3.
reflexivity.
Qed.



End field_of_sets_sec.

Section finite_cofinite_sec.


Variable Xt : Type.
Definition F_fc := 
  [S:(Ensemble Xt) | Finite S \/ Finite (Comp S)].

Lemma Union_closed_fc : forall S1 S2:(Ensemble Xt), 
    In F_fc S1 -> In F_fc S2 -> In F_fc (Union S1 S2).
intros S1 S2 h1 h2. 
unfold F_fc. unfold F_fc in h1. unfold F_fc in h2.
inversion h1 as [h3].
inversion h2 as [h4].
destruct h3 as [h5 | h6].
  (*left*)
  destruct h4 as [h7 | h8].
    (*left*)
    constructor.  left.
    apply Union_preserves_Finite; assumption.
    (*right*)
    pose proof (comp_union _ S1 S2) as h9.
    assert (h10: Included (Intersection (Comp S1) (Comp S2))
      (Comp S2)).
      auto with sets.
    pose proof (Finite_downward_closed _ _ h8 _ h10) as h11.
    rewrite <- h9 in h11.
    constructor. right; assumption.
  (*right*)
  destruct h4 as [h12 | h13].
    (*left*)
    pose proof (comp_union _ S1 S2) as h14.
    assert (h15: Included (Intersection (Comp S1) (Comp S2))
      (Comp S1)).
      auto with sets.
    pose proof (Finite_downward_closed _ _ h6 _ h15) as h16.
    rewrite <- h14 in h16.
    constructor. right; assumption.
    (*right*)
    constructor.  right.
    pose proof (comp_union _ S1 S2) as h17.
    assert (h18: Included (Intersection (Comp S1) (Comp S2))
      (Comp S1)).
      auto with sets.
    rewrite <- h17 in h18.
    apply Finite_downward_closed with (Comp S1); 
      assumption.
Qed.

Lemma Comp_closed_fc : forall S :(Ensemble Xt),
    In F_fc S -> In F_fc (Comp S).
intros S h1.
unfold F_fc in h1.
inversion h1 as [h2].
destruct h2 as [h3 | h4].
(*left*)
unfold F_fc.
constructor. right.
pose proof (Complement_Complement _ S) as h5.
rewrite h5. assumption.
(*right*)
unfold F_fc.
constructor. left. assumption.
Qed.


Lemma Int_closed_fc : forall S1 S2:(Ensemble Xt),
    In F_fc S1 -> In F_fc S2 -> In F_fc (Intersection S1 S2).
intros S1 S2 h1 h2. 
pose proof (Complement_Complement _ S1) as h3.
pose proof (Complement_Complement _ S2) as h4.
rewrite <- h3. rewrite <- h4.
pose proof (comp_union _ 
  (Comp S1) (Comp S2)) as h5.
rewrite <- h5.
pose proof (Union_closed_fc 
  (Comp S1) (Comp S2)) as h6.
pose proof (Comp_closed_fc S1 h1) as h7.
pose proof (Comp_closed_fc S2 h2) as h8.
pose proof (h6 h7 h8) as h9.
apply Comp_closed_fc. assumption.
Qed.

Lemma non_empty_F_fc : (exists S:(Ensemble Xt), In F_fc S).
unfold F_fc.
exists (Empty_set Xt).
constructor.
left. constructor.
Qed.

Lemma full_in_F_fc : In F_fc (Full_set _).
pose proof non_empty_F_fc as h1.
elim h1. intros S h2.
pose proof (Comp_closed_fc _ h2) as h3.
pose proof (Union_closed_fc _ _ h2 h3) as h4.
pose proof (comp_sum_psa S) as h5.
rewrite <- h5; assumption.
Qed.

Lemma empty_in_F_fc : In F_fc (Empty_set _).
pose proof non_empty_F_fc as h1.
elim h1. intros S h2.
pose proof (Comp_closed_fc _ h2) as h3.
pose proof (Int_closed_fc _ _ h2 h3) as h4.
pose proof (comp_prod_psa S) as h5.
rewrite <- h5; assumption.
Qed.


Definition fin_cof_fos := Build_Field_of_Sets Xt F_fc non_empty_F_fc
  Union_closed_fc Int_closed_fc Comp_closed_fc.
Definition fin_cof_ba := fos_ba fin_cof_fos.
End finite_cofinite_sec.

Section Count_Cocount.
Variable Xt:Type.
Definition F_cc := 
  [S:(Ensemble Xt) | Countable S \/ Countable (Comp S)].

Lemma Union_closed_cc : forall S1 S2:(Ensemble Xt), 
    In F_cc S1 -> In F_cc S2 -> In F_cc (Union S1 S2).
intros S1 S2 h1 h2.
inversion h1 as [h3].
inversion h2 as [h4].
unfold F_cc.
constructor.
destruct h3 as [h5 | h6].
  (*left*)
  destruct h4 as [h7|h8].
    (*left*)
    left. 
    apply countable_union2; trivial. 
    (*right*)
    right.
    pose proof (comp_union _ S1 S2) as h9.
    rewrite h9.
    apply (countable_downward_closed 
      (Intersection (Comp S1) (Comp S2))
      (Comp S2)).
    assumption. auto with sets.
  (*right*)
  destruct h4 as [h10|h11].
    (*left*)
    right.
    pose proof (comp_union _ S1 S2) as h12.
    rewrite h12.
    apply (countable_downward_closed
      (Intersection (Comp S1) (Comp S2))
      (Comp S1)).
    assumption. auto with sets.
    (*right*)
    right.
    pose proof (comp_union _ S1 S2) as h13.
    rewrite h13.
    apply countable_intersection; trivial.
Qed. 

Lemma Comp_closed_cc : forall S :(Ensemble Xt),
    In F_cc S -> In F_cc (Comp S).
unfold F_cc.
intros S h1.
inversion h1 as [h2].
constructor.
pose proof (Complement_Complement _ S) as h3.
rewrite h3.
tauto.
Qed.

Lemma Int_closed_cc : forall S1 S2:(Ensemble Xt),
    In F_cc S1 -> In F_cc S2 -> In F_cc (Intersection S1 S2).
intros S1 S2 h1 h2.
pose proof (Complement_Complement _ S1) as h3.
pose proof (Complement_Complement _ S2) as h4.
rewrite <- h3. rewrite <- h4.
pose proof (comp_union _ 
  (Comp S1) (Comp S2)) as h5.
rewrite <- h5.
pose proof (Union_closed_cc 
  (Comp S1) (Comp S2)) as h6.
pose proof (Comp_closed_cc S1 h1) as h7.
pose proof (Comp_closed_cc S2 h2) as h8.
pose proof (h6 h7 h8) as h9.
apply Comp_closed_cc. assumption.
Qed.

Lemma non_empty_F_cc : (exists S:(Ensemble Xt), In F_cc S).
unfold F_cc.
exists (Empty_set Xt).
constructor.
left.
apply Finite_impl_Countable.
constructor.
Qed.

Definition cnt_ccnt_fos := Build_Field_of_Sets Xt F_cc non_empty_F_cc
  Union_closed_cc Int_closed_cc Comp_closed_cc.
Definition cnt_ccnt_ba := fos_ba cnt_ccnt_fos.


End Count_Cocount.

(*For now, these will only be done with rational points and intervals, which should be sufficient for my purposes.*)
Section IntervalAlgebras.
Require Import ListUtilities.
Open Scope Q.

(*Extended rationals, i.e. with infinity endpoints.*)
Inductive Qx : Type :=
|neg_inf : Qx
|pos_inf : Qx
|q_qx : Q->Qx.

Inductive qx_lt : Qx->Qx->Prop :=
|inf2_lt : qx_lt neg_inf pos_inf
|lt_inf : forall a:Q, qx_lt (q_qx a) pos_inf
|inf_lt : forall a:Q, qx_lt neg_inf (q_qx a)
|q_qx_lt : forall a b:Q, a < b -> qx_lt (q_qx a) (q_qx b).

Inductive qx_eq : Qx->Qx->Prop :=
|q_qx_eq:forall a b:Q, a == b -> qx_eq (q_qx a) (q_qx b)
|neg_inf_eq :qx_eq neg_inf neg_inf
|pos_inf_eq: qx_eq pos_inf pos_inf.

Delimit Scope qx_scope with Qx.
Open Scope qx_scope.
Notation "x < y" := (qx_lt x y) : qx_scope.
Notation "x == y" := (qx_eq x y) : qx_scope.
Notation "x <= y" := (x < y \/ x == y) : qx_scope.

Inductive qx_is_q : Qx -> Prop :=
| qx_is_q_intro : forall a:Q, qx_is_q (q_qx a).



Lemma q_qx_fst :
  forall (p:Q*Q),
    q_qx (fst p) == 
    fst (q_qx (fst p), q_qx (snd p))%Q.
  intros; constructor; apply Qeq_refl.
Qed.

Lemma q_qx_snd :
  forall (p:Q*Q),
    q_qx (snd p) == 
    snd (q_qx (fst p), q_qx (snd p))%Q.
  intros; constructor; apply Qeq_refl.
Qed.


Lemma qx_lt_irrefl : 
  forall (x:Qx),
    ~ x < x. 
intros x h1; dependent induction h1.
apply Qlt_irrefl in H; auto.
Qed.

Lemma qx_eq_refl :
  forall (x:Qx),
    x == x.
intro x.
destruct x; try constructor; auto.
constructor.
Qed.

Lemma sym_qx_eq : 
  forall (x y:Qx),
    x == y -> y == x.
intros x y h1.
destruct x, y; try induction h1; try constructor; 
try (rewrite H; try constructor); try constructor.
Qed.

Lemma trans_qx_eq : 
  forall (x y z:Qx),
    x == y -> y == z ->
    x == z.
intros x y z h1 h2.
destruct x, y; try inversion h1; try inversion h2; try constructor.
subst.
inversion h2. subst.
rewrite H1. assumption.
Qed.



Lemma qx_le_refl : 
  forall (x:Qx),
    x <= x.
intro x. right. apply qx_eq_refl.
Qed.


Lemma trans_qx_lt : Transitive _ qx_lt.
red. 
intros x y z h1 h2. 
destruct x, y, z; try dependent induction h1; try dependent induction h2; try constructor.
eapply Qlt_trans; auto.
apply H. apply H0.
Qed.


Lemma trans_qx_le_lt : forall x y z:Qx, x<=y -> y<z -> x<z.

intros x y z h1 h2.
destruct x, y, z; try dependent induction h1; try dependent induction h2; try constructor.
dependent induction H.
inversion H.
dependent induction H. 
inversion H.
dependent induction H.
inversion H.
dependent induction H.
dependent induction H.
dependent induction H.
inversion H.
eapply Qlt_trans; auto. 
inversion H. subst.   apply H3. apply H0.
inversion H.  subst. 
rewrite <- H3 in H0. assumption.
Qed.

Lemma trans_qx_lt_le : forall x y z:Qx, x<y -> y<=z -> x<z.
intros x y z h1 h2.
destruct x, y, z; try dependent induction h1; try dependent induction h2; try constructor.
dependent induction H.
inversion H.
dependent induction H.
inversion H.
dependent induction H.
inversion H.
dependent induction H.
inversion H.
dependent induction H0.
inversion H0.
inversion H0. subst.
eapply Qlt_trans. apply H. apply H3.
inversion H0. subst.
rewrite H3 in H. assumption.
Qed.


Lemma trans_qx_eq_le : 
  forall x y z:Qx, 
    x == y -> y <= z -> x <= z.
intros x y z h1 h2.
destruct x, y, z; try apply qx_le_refl; try inversion h1; try inversion h2.
left. assumption. left. constructor.
inversion H. subst. left. constructor. left. constructor.
inversion H. inversion H. left. assumption.
inversion H. inversion H2. inversion H2. left. constructor.
left. constructor. subst.
left. constructor. inversion H2. subst.
rewrite H1. assumption. subst.
inversion H2. subst.
right. constructor. rewrite H1.
assumption.
Qed.


Lemma trans_qx_le_eq : 
  forall x y z:Qx, 
    x <= y -> y == z -> x <= z.
intros x y z h1 h2.
destruct x, y, z; try apply qx_le_refl; try inversion h1; try inversion h2.
left. assumption. left. constructor. left. constructor.
left. constructor. inversion H.
destruct h1; inversion H0.
subst.
destruct h1; inversion H0. subst.
inversion H. inversion H. inversion H. 
left. constructor. left. constructor. subst.
left. constructor.
rewrite <- H2. inversion H.
assumption.
subst. right.
constructor.
inversion H. subst.
rewrite <- H2.
assumption.
Qed.

Lemma trans_qx_eq_lt : 
  forall x y z:Qx, 
    x == y -> y < z -> x < z.
intros x y z h1 h2.
destruct x, y, z; try inversion h2; try constructor; try inversion h1.
subst.
rewrite H4.
assumption.
Qed.

Lemma trans_qx_lt_eq : 
  forall x y z:Qx, 
    x < y -> y == z -> x < z.
intros x y z h1 h2.
destruct x, y, z; try inversion h1; try inversion h2; try constructor.
subst.
rewrite <- H4.
assumption.
Qed.

 
Lemma trans_qx_le : 
  forall x y z:Qx,
    x <= y -> y <= z -> x <= z.
intros x y z h1 h2. 
destruct h2 as [h2 | h2].
left. eapply trans_qx_le_lt. apply h1. apply h2.
eapply trans_qx_le_eq. apply h1.
apply h2.
Qed.

Lemma q_qx_eq_iff :
  forall (p q:Q),
    (q_qx p) == (q_qx q) <-> (p == q)%Q.
  intros; split; intro h1; try constructor; auto; try inversion h1; subst; auto.
Qed.


Lemma q_qx_le_iff :
  forall (p q:Q),
    (q_qx p) <= (q_qx q) <-> (p <= q)%Q.
intros p q. split.
intro h1. 
destruct h1 as [h1|h1].
inversion h1. subst.
apply Qlt_impl_Qle.
assumption.
inversion h1. subst.
rewrite H1. apply Qle_refl.
intro h1.
apply Qle_lt_or_eq in h1.
destruct h1 as [h1 | h1].
left. constructor; auto.
right. constructor. assumption.
Qed.


Lemma q_qx_lt_iff :
  forall (p q:Q),
    (q_qx p) < (q_qx q) <-> (p < q)%Q.
  intros p q; split; intro h1.
  inversion h1; auto.
  constructor; auto.
Qed.



Lemma q_qx_dec :
  forall (p q:Q),
    {q_qx p < q_qx q} + {q_qx p == q_qx q} + {q_qx q < q_qx p}.
intros p q.
pose proof (Q_dec p q) as h1.
destruct h1 as [[h1 | h1]|h1].
left; constructor; constructor; auto.
right; constructor; auto. 
left; right; constructor; auto.
Qed.


Lemma qx_dec :
  forall (p q:Qx),
    {p < q} +  {p == q} + {q < p}.
intros p q. 
destruct p, q; try apply q_qx_dec; try (left; right;constructor); try (left; left; constructor); try (right; constructor).
Qed.

Lemma qx_dec' : 
  forall (p q:Qx),
    {p < q} + {q <= p}.
intros p q.
destruct (qx_dec p q) as [h1 | h1].
destruct h1 as [h1 | h1].
left; auto.
right. right. apply sym_qx_eq. assumption.
right. left. assumption.
Qed.


Lemma qx_is_q_dec :
  forall q:Qx,
    {qx_is_q q} + {~qx_is_q q}.
  intro q.
  destruct (qx_dec neg_inf q) as [h2 | h2].
  destruct h2 as [h2 | h2].
  destruct (qx_dec q pos_inf) as [h3 | h3].
  destruct h3 as [h3 | h3].
  left.
  destruct q.
  apply qx_lt_irrefl in h2; contradiction.
  apply qx_lt_irrefl in h3; contradiction.
  constructor.
  right.
  intro h4.
  destruct h4.
  inversion h3.
  right.
  inversion h3.
  right.
  intro h3.
  destruct h3.
  inversion h2.
  right.
  inversion h2.
Qed.


Lemma not_qx_lt_iff : 
  forall (p q:Qx),
    ~ p < q <-> q <= p.
intros p q.
split.
intro  h1.
pose proof (qx_dec p q) as h2.
destruct h2 as [h2 | h2]. destruct h2 as [h2 | h2].
contradiction. right. apply sym_qx_eq.
assumption.
left; auto.
intros h2 h3.
destruct h2 as [h2 | h2].
pose proof (trans_qx_lt _ _ _ h2 h3) as h4.
apply qx_lt_irrefl in h4; auto.
pose proof (trans_qx_eq_lt _ _ _ h2 h3) as h4.
apply qx_lt_irrefl in h4; auto.
Qed.

Lemma not_qx_le_iff : 
  forall (p q:Qx),
    ~ p <= q <-> q < p.
intros p q.
split.
intro  h1.
pose proof (qx_dec p q) as h2.
destruct h2 as [h2 | h2]. destruct h2 as [h2 | h2].
contradict h1. left. assumption.
contradict h1. right. assumption.
assumption.
intros h2 h3.
pose proof (trans_qx_lt_le _ _ _ h2 h3) as h4.
apply qx_lt_irrefl in h4; auto.
Qed.



Definition qx_min (p q:Qx) :=
  if (qx_dec p q) then p else q.

Definition qx_max (p q:Qx) :=
  if (qx_dec p q) then q else p.


Lemma qx_min_le :
  forall (p q:Qx),
    qx_min p q <= p /\ qx_min p q <= q.
intros p q.
unfold qx_min.
destruct qx_dec as [h1 | h1].
split. right. apply qx_eq_refl.
destruct h1 as [h1 | h1].
left. assumption. right. assumption.
split. left. assumption.
right. apply qx_eq_refl.
Qed.

Lemma le_qx_max : 
  forall (p q:Qx),
    p <= qx_max p q /\ q <= qx_max p q.
intros p q.
unfold qx_max.
destruct qx_dec as [h1 | h1].
destruct h1 as [h1 | h1]. split.
left. assumption. right. apply qx_eq_refl.
split. right. apply h1. right. apply qx_eq_refl.
split.
right. apply qx_eq_refl.
left. assumption.
Qed.

Lemma qx_2lt_min : 
  forall (p q q':Qx),
    p < q -> p < q' ->
    p < qx_min q q'.
intros p q q' h1 h2.
unfold qx_min.
destruct qx_dec; auto.
Qed.


Lemma qx_2lt_min_min : 
  forall (p p' q q':Qx),
    p < q -> p' < q' ->
    qx_min p p' < qx_min q q'.
intros p p' q q' h1 h2.
unfold qx_min.
destruct qx_dec as [h3 | h3]; destruct qx_dec as [h4 | h4]; auto.
destruct h3 as [h3 | h3].
eapply trans_qx_lt. apply h3. apply h2.
eapply trans_qx_eq_lt. apply h3. assumption.
destruct h4 as [h4 | h4].
eapply trans_qx_lt. apply h3. apply h1.
apply sym_qx_eq in h4.
eapply trans_qx_lt_eq. apply h2.
assumption.
Qed.

Lemma qx_2lt_max : 
  forall (p p' q:Qx),
    p < q -> p' < q ->
    qx_max p p' < q.
intros p p' q h1 h2.
unfold qx_max.
destruct qx_dec; auto.
Qed.

Lemma qx_2le_max : 
  forall (p p' q:Qx),
    p <= q -> p' <= q ->
    qx_max p p' <= q.
intros p p' q h1 h2.
unfold qx_max.
destruct qx_dec; auto.
Qed.



Lemma not_pos_inf_lt : 
  forall q:Qx,
    ~pos_inf < q. 
intros q h1.
inversion h1.
Qed.

Lemma not_lt_neg_inf :
  forall q:Qx,
    ~q < neg_inf.
intros q h1.
inversion h1.
Qed.

Lemma lt_neg_inf :
  forall q:Q,
    neg_inf < (q_qx q).
  intro q.
  constructor.
Qed.


Lemma le_neg_inf :
  forall q:Qx,
    neg_inf <= q.
  intro q; destruct q.
  apply qx_le_refl.
  constructor.
  constructor.
  constructor.
  constructor.
Qed.


Lemma lt_pos_inf :
  forall q:Q,
   (q_qx q) < pos_inf.
  intro q.
  constructor.
Qed.

Lemma le_pos_inf :
  forall q:Qx,
    q <= pos_inf.
  intro q; destruct q.
  constructor.
  constructor.
  apply qx_le_refl.
  constructor.
  constructor.
Qed.


Lemma qx_lt_dec : 
  forall (p q:Qx),
    p < q ->
    {exists p', p = q_qx p'} + {p = neg_inf}.
intros p q h1.
destruct p. right; auto.
apply not_pos_inf_lt in h1; contradiction.
left.
exists q0.
reflexivity.
Qed.

Lemma qx_lt_dec' :
  forall (p q:Qx),
    p < q ->
    {exists q', q = q_qx q'} + {q = pos_inf}.
intros p q h1.
destruct q. apply not_lt_neg_inf in h1.
contradiction.
right; auto.
left.
exists q.
reflexivity.
Qed.


(*Half-open interval.*)
Definition ho_invl (a b:Qx) : Ensemble Q := 
  [x:Q | a <= q_qx x /\ q_qx x < b].

Definition ho_invl' (p:Qx*Qx) :=
  ho_invl (fst p) (snd p).

Definition hoq_invl (a b:Q) :=
  ho_invl (q_qx a) (q_qx b).

Definition hoq_invl' (p:Q*Q) :=
  hoq_invl (fst p) (snd p).

Lemma hoq_invl_compat' :
  forall (p:Q*Q),
    hoq_invl' p =
    ho_invl' (q_qx (fst p), q_qx (snd p)).
  auto.
Qed.


Lemma inh_ho_invl_iff : 
  forall (a b:Qx),
    Inhabited (ho_invl a b) <->
    a < b.
intros a b. split.
intro h1.
destruct h1 as [x h1].
destruct h1 as [h1].
destruct h1 as [h1 h2].
eapply trans_qx_le_lt.
apply h1. apply h2.
intro h1.
destruct a, b; try inversion h1.
apply Inhabited_intro with 0.
constructor.
split; constructor. constructor.
subst.
apply Inhabited_intro with (q - 1).
constructor. split; try constructor. constructor.
apply Qminus_pos_Qlt. constructor.
subst.
apply Inhabited_intro with q.
constructor. split; try right; constructor. constructor.
subst.
apply Inhabited_intro with q. 
constructor.
split. right. constructor. constructor.
assumption.
Qed.

Lemma inh_hoq_invl_iff : 
  forall (a b:Q),
    Inhabited (hoq_invl a b) <->
    Qlt a  b.
  intros; rewrite <- q_qx_lt_iff.
  apply inh_ho_invl_iff.
Qed.

Lemma in_ho_invl_lr_iff :
  forall (q r s:Q),
    Inn (ho_invl' (q_qx q, q_qx r)) s <->
    Qle q s /\ Qlt s r.
  intros q r s; split; intro h1. 
  inversion h1.
  simpl in H.
  destruct H.
  inversion H.
  inversion H0.
  inversion H1; subst.
  split.
  apply Qlt_le_weak; auto.
  assumption.
  inversion H0; inversion H1; subst.
  split.
  rewrite H7.
  apply Qle_refl.
  assumption.

  destruct h1 as [h1 h2].
  econstructor.
  simpl; split; try rewrite q_qx_le_iff; auto.
  constructor; auto.
Qed.

Lemma in_hoq_invl_lr_iff :
  forall (q r s:Q),
    Inn (hoq_invl' (q, r)) s <->
    Qle q s /\ Qlt s r.
  intros; rewrite <- in_ho_invl_lr_iff; tauto.
Qed.



Lemma in_ho_invl_l_iff :
  forall (q s:Q),
    Inn (ho_invl' (q_qx q, pos_inf)) s <->
    Qle q s.
  intros q s; split; intro h1.
  destruct h1 as [h1].
  simpl in h1.
  destruct h1 as [h1 h2].
  rewrite q_qx_le_iff in h1; auto.
  constructor.
  simpl.
  split.
  rewrite q_qx_le_iff; auto.
  constructor.
Qed.


Lemma in_ho_invl_r_iff :
  forall (r s:Q),
    Inn (ho_invl' (neg_inf, q_qx r)) s <->
    Qlt s r.
  intros r s; split; intro h1.
  destruct h1 as [h1].
  simpl in h1.
  destruct h1 as [h1 h2].
  rewrite q_qx_lt_iff in h2; auto.
  constructor.
  simpl.
  split.
  apply le_neg_inf.
  rewrite q_qx_lt_iff; auto.
Qed.


Lemma in_ho_invl :
  forall (s:Q),
    Inn (ho_invl' (neg_inf, pos_inf)) s.
  intro s.
  constructor.
  simpl; split.
  apply le_neg_inf.
  apply lt_pos_inf.
Qed.


Lemma ho_invl_neg_inf_eq :
  forall x:Qx,
  ho_invl x neg_inf = Empty_set _.
intro x.
apply Extensionality_Ensembles; red; split; red; intros a h1; try contradiction.
destruct h1 as [h1].
destruct h1 as [h1 h2].
inversion h2.
Qed.

Lemma ho_invl_pos_inf_eq :
  forall x:Qx,
  ho_invl pos_inf x = Empty_set _.
intro x.
apply Extensionality_Ensembles; red; split; red; intros a h1; try contradiction.
destruct h1 as [h1].
destruct h1 as [h1 h2]. 
inversion h1 as [h3|h3]; inversion h3.
Qed.



Lemma ho_invl_neg_inf_eq_neg_inf :
  forall (a b c:Qx),
    Inhabited (ho_invl neg_inf b) ->
    ho_invl neg_inf b = ho_invl a c ->
    a = neg_inf.
intros a b c h0 h1.
destruct (qx_dec a neg_inf) as [[h2|h2]|h2].
inversion h2.
inversion h2. reflexivity.
destruct a; auto.
rewrite ho_invl_pos_inf_eq in h1.
destruct b.
rewrite ho_invl_neg_inf_eq in h0.
destruct h0; contradiction.
assert (h3:Inn (ho_invl neg_inf pos_inf) 0).
  constructor. split; auto.
  left. constructor. constructor.
rewrite  h1 in h3.
contradiction.
assert (h4:Inn (ho_invl neg_inf (q_qx q)) (q-1)).
  constructor. split; auto.
  left. constructor. constructor.   
  apply Qminus_pos_Qlt.
  constructor.
rewrite h1 in h4.
contradiction.
pose proof h0 as h0'.
rewrite h1 in h0'.
rewrite inh_ho_invl_iff in h0'.
assert (h3:Inn (ho_invl (q_qx q) c) q).
  constructor. split; auto; try right; try constructor; try constructor.
rewrite <- h1 in h3.
destruct h3 as [h3]. destruct  h3 as [h3 h4].
assert (h5:Inn (ho_invl neg_inf b) (q - 1)).
  constructor. split. left. constructor.
  assert (h5:(q -1 < q)%Q).
    apply Qminus_pos_Qlt.
    constructor.
  assert (h6:q_qx (q - 1) < q_qx q).
    constructor. assumption.
  eapply trans_qx_lt.
  apply h6. apply h4.
rewrite h1 in h5.
destruct h5 as [h5].
destruct h5 as [h5 h6].
destruct h5 as [h5 | h5].
inversion h5. subst.
assert (h7:(0 < 1)%Q).
  constructor.
pose proof (Qminus_pos_Qlt 1 q h7) as h8.
pose proof (Qlt_trans _ _ _ H1 h8) as h9.
apply Qlt_irrefl in h9.
contradiction.
inversion h5. subst.
assert (h7:(q + 0 == q + -(1))%Q).
  rewrite <- H1.
  rewrite Qplus_0_r.
  constructor.
rewrite Qplus_inj_l in h7 at 1.
inversion h7.
Qed.

Lemma ho_invl_neg_inf_pos_inf_eq :
  ho_invl neg_inf pos_inf = Full_set Q.
apply Extensionality_Ensembles; red; split; red; intros x h1; try constructor.
split.
constructor. constructor.
constructor.
Qed.


Lemma ho_invl_pos_inf_eq_pos_inf :
  forall (a b c:Qx),
    Inhabited (ho_invl a pos_inf) ->
    ho_invl a pos_inf = ho_invl b c ->
    c = pos_inf.
intros a b c h0 h1.
destruct (qx_dec a pos_inf) as [[h2|h2]|h2].
inversion h2.  subst.
rewrite ho_invl_neg_inf_pos_inf_eq in h1.
destruct b, c; auto. 
rewrite ho_invl_neg_inf_eq in h1.
assert (h3:Inn (Full_set Q) 0).
  constructor.
rewrite h1 in h3.
contradiction.
assert (h3:Inn (Full_set Q) q).
  constructor.
rewrite h1 in h3.
destruct h3 as [h3]. destruct h3 as [h3 h4].
inversion h4.
subst.
apply Qlt_irrefl in H1.
contradiction.
rewrite ho_invl_pos_inf_eq in h1.
assert (h3:Inn (Full_set Q) 0).
  constructor.
rewrite h1 in h3. contradiction.
rewrite ho_invl_pos_inf_eq in h1.
assert (h3:Inn (Full_set Q) 0).
  constructor.
rewrite h1 in h3. contradiction.
rewrite ho_invl_neg_inf_eq in h1.
assert (h3:Inn (Full_set Q) 0).
  constructor.
rewrite h1 in h3. contradiction.
assert (h3:Inn (Full_set Q) q0).
  constructor.
rewrite h1 in h3. 
destruct h3 as [h3]. destruct h3 as [h3 h4].
inversion h4. subst.
apply Qlt_irrefl in H1. contradiction.
subst.
pose proof h0 as h0'. rewrite h1 in h0'.
rewrite inh_ho_invl_iff in h0'.
destruct b, c; auto.
apply qx_lt_irrefl in h0'. contradiction.
assert (h3:Inn (ho_invl (q_qx a0) pos_inf) a0).
  constructor. split. right. constructor. constructor.
  constructor.
rewrite h1 in h3. 
destruct h3 as [h3]. destruct h3 as [h3 h4].
assert (h5:Inn (ho_invl (q_qx a0) pos_inf) (q + 1)).
  constructor.
  split. constructor. constructor.
  inversion h4. subst.
  assert (h5:(q < q + 1)%Q).
    apply Qlt_plus_pos. constructor.
eapply Qlt_trans.
apply H1. apply h5.
constructor.
rewrite h1 in h5.
destruct h5 as [h5]. destruct h5 as [h5].
inversion H. subst.
assert (h6:(0 < 1)%Q). constructor.
pose proof (Qlt_plus_pos 1 q h6) as h7.
pose proof (Qlt_trans _ _  _ h7 H2) as h8.
apply Qlt_irrefl in h8.
contradiction.
inversion h0'. inversion h0'. inversion h0'.
assert (h3:Inn (ho_invl (q_qx q) (q_qx q0)) q).
  constructor; auto. split.
  right. constructor. constructor. assumption.
rewrite <- h1 in h3.
destruct h3 as [h3]. destruct h3 as [h3 h4].
pose proof (trans_qx_le_lt _ _ _ h3 h0') as h5.
assert (h6:Inn (ho_invl (q_qx a0) pos_inf) q0).
  constructor. split. left. assumption.
  constructor.
rewrite h1 in h6.
destruct h6 as [h6]. destruct h6 as [h6 h7]. inversion h7.
subst. apply Qlt_irrefl in H1. contradiction.
inversion h2. subst.
rewrite ho_invl_pos_inf_eq in h0.
destruct h0; contradiction.
inversion h2.
Qed.

Lemma ho_invl_neg_inf_eq2 :
  forall (q r:Qx),
    ho_invl neg_inf q =
    ho_invl neg_inf r ->
    q == r.
intros q r h1.
destruct q, r; try constructor; auto.
rewrite ho_invl_neg_inf_eq in h1.
rewrite ho_invl_neg_inf_pos_inf_eq in h1.
assert (h2:Inn (Full_set Q) 0).
  constructor.
rewrite <- h1 in h2. contradiction.
rewrite ho_invl_neg_inf_eq in h1.
assert (h2:Inn (ho_invl neg_inf (q_qx q)) (q-1)).
  constructor. split. left. constructor.
  constructor. apply Qminus_pos_Qlt. constructor.
rewrite <- h1 in h2.
contradiction.
rewrite ho_invl_neg_inf_pos_inf_eq in h1.
rewrite ho_invl_neg_inf_eq in h1.
assert (h2:Inn (Full_set Q) 0).
  constructor.
rewrite h1 in h2. contradiction. 
apply ho_invl_pos_inf_eq_pos_inf in h1.
rewrite h1. constructor. 
rewrite ho_invl_neg_inf_pos_inf_eq. 
apply Inhabited_intro with 0. constructor.
rewrite ho_invl_neg_inf_eq in h1.
assert (h2:Inn (ho_invl neg_inf (q_qx q)) (q-1)).
  constructor. split. left. constructor.
  constructor. apply Qminus_pos_Qlt. constructor.
rewrite h1 in h2. contradiction.
symmetry in h1.
apply ho_invl_pos_inf_eq_pos_inf in h1.
discriminate h1.
rewrite ho_invl_neg_inf_pos_inf_eq.
apply Inhabited_intro with 0. constructor.
destruct (Q_dec q q0) as [[h2|h2]|h2].
assert (h3:Inn (ho_invl neg_inf (q_qx q0)) q).
  constructor. split. left. constructor.
  constructor. assumption.
rewrite <- h1 in h3.
destruct h3 as [h3]. destruct  h3 as [h3 h4].
inversion h4. subst. apply Qlt_irrefl in H1.
contradiction.

assert (h3:Inn (ho_invl neg_inf (q_qx q)) q0).
  constructor. split. left. constructor.
  constructor. assumption.
rewrite h1 in h3.
destruct h3 as [h3]. destruct  h3 as [h3 h4].
inversion h4. subst. apply Qlt_irrefl in H1.
contradiction.
assumption.
Qed.



Lemma ho_invl_pos_inf_eq2 :
  forall (q r:Qx),
    ho_invl q pos_inf =
    ho_invl r pos_inf ->
    q == r.
intros q r h1.
destruct q, r; try constructor; auto.
rewrite ho_invl_pos_inf_eq in h1.
rewrite ho_invl_neg_inf_pos_inf_eq in h1.
assert (h2:Inn (Full_set Q) 0).
  constructor.
rewrite h1 in h2. contradiction.
apply ho_invl_neg_inf_eq_neg_inf in h1.
discriminate h1.
rewrite ho_invl_neg_inf_pos_inf_eq.
apply Inhabited_intro with 0. constructor.
rewrite ho_invl_pos_inf_eq in h1.
rewrite ho_invl_neg_inf_pos_inf_eq in h1.
assert (h2:Inn (Full_set Q) 0).
  constructor.
rewrite <- h1 in h2. contradiction.
rewrite ho_invl_pos_inf_eq in h1.
assert (h2:Inn (ho_invl (q_qx q) pos_inf) q).
  constructor. split. right. constructor. constructor.
  constructor.
rewrite <- h1 in h2.
contradiction.
symmetry in h1.
apply ho_invl_neg_inf_eq_neg_inf in h1.
discriminate h1.
rewrite ho_invl_neg_inf_pos_inf_eq.
apply Inhabited_intro with 0. constructor.
rewrite ho_invl_pos_inf_eq in h1.
assert (h2:Inn (ho_invl (q_qx q) pos_inf) q).
  constructor. split. right. constructor. constructor.
  constructor.
rewrite h1 in h2.
contradiction.

destruct (Q_dec q q0) as [[h2|h2]|h2].  
assert (h3:~Inn (ho_invl (q_qx q0) pos_inf) q).
  intro h4.
  destruct h4 as [h4]. destruct h4 as [h4 h5].
  rewrite q_qx_le_iff in h4.
  pose proof (Qle_lt_trans _ _ _ h4 h2) as h6.
  apply Qlt_irrefl in h6. contradiction.
rewrite <- h1 in h3.
contradict h3.
constructor. split. right. constructor. constructor. constructor.

assert (h3:~Inn (ho_invl (q_qx q) pos_inf) q0).
  intro h4.
  destruct h4 as [h4]. destruct h4 as [h4 h5].
  rewrite q_qx_le_iff in h4.
  pose proof (Qle_lt_trans _ _ _ h4 h2) as h6.
  apply Qlt_irrefl in h6. contradiction.
rewrite h1 in h3.
contradict h3.
constructor. split. right. constructor. constructor. constructor.

assumption.
Qed.



Lemma ho_invl_qq_union :
  forall (p q r:Q),
    (p < q)%Q -> (q < r)%Q ->
    Union (ho_invl (q_qx p) (q_qx q)) (ho_invl (q_qx q) (q_qx r)) =
    ho_invl (q_qx p) (q_qx r).
intros p q r hp hq.
pose proof (Qlt_trans _ _ _ hp hq) as h0.
apply Extensionality_Ensembles; red; split; red; intros x h1.
destruct h1 as [x h1 | x h1]. 
destruct h1 as [h1]. destruct h1 as [h1 h2].
constructor. split; auto.
inversion h2. subst.
constructor. 
eapply Qlt_trans. apply H1. apply hq.
destruct h1 as [h1].  destruct h1 as [h1 h2].
constructor. split.  
left. eapply trans_qx_lt_le.
assert (h3:q_qx p < q_qx q). constructor; auto.
apply h3. assumption. assumption.
destruct h1 as [h1]. destruct h1 as [h1 h2]. 

destruct (qx_dec (q_qx x) (q_qx q)) as [[h3|h3]|h3].
left. constructor. split; auto.
inversion h3. subst. 
right. constructor. split. 
right. constructor. rewrite H1. constructor.
assumption.
right. constructor. split.
left. assumption. assumption.
Qed.


Lemma hoq_invl_qq_union :
  forall (p q r:Q),
    (p < q)%Q -> (q < r)%Q ->
    Union (hoq_invl p q) (hoq_invl q r) =
    hoq_invl p r.
  unfold hoq_invl; intros; apply ho_invl_qq_union; auto.
Qed.


Lemma ho_invl_qq_eq :
  forall q r q' r':Q,
    (q < r)%Q -> (q' < r')%Q ->
    ho_invl (q_qx q) (q_qx r) = ho_invl (q_qx q') (q_qx r') ->
    (q == q')%Q /\ (r == r')%Q.
intros q r q' r' h1 h2 h3.
pose proof (Q_dec r r') as h5. 
assert (h7:(q' <= q)%Q).
  assert (h8:Inn (ho_invl (q_qx q') (q_qx r')) q).
    rewrite <- h3. constructor. split.
    right. constructor. constructor.
    constructor. assumption.
  destruct h8 as [h8]. destruct h8 as [h8 h9].
  rewrite q_qx_le_iff in h8.
  assumption.
assert (h7':(q <= q')%Q).
  assert (h8:Inn (ho_invl (q_qx q) (q_qx r)) q').
    rewrite h3. constructor. split.
    right. constructor. constructor.
    constructor. assumption.
  destruct h8 as [h8]. destruct h8 as [h8 h9].
  rewrite q_qx_le_iff in h8.
  assumption.
pose proof (Qle_antisym _ _ h7 h7') as h0.
split. rewrite h0. constructor.
destruct h5 as [[h5 | h5] | h5].  
pose proof (Qlt_trans _ _ _ h1 h5) as h6. clear h7 h7'.
assert (h7:Inn (ho_invl (q_qx q') (q_qx r')) r).
  constructor. split.
  rewrite q_qx_le_iff.
  rewrite h0. apply Qlt_impl_Qle. assumption.
  constructor. assumption.
rewrite <- h3 in h7.
destruct h7 as [h7]. destruct h7 as [h7 h8].
inversion h8. subst.
apply Qlt_irrefl in H1. contradiction.

assert (h9:Inn (ho_invl (q_qx q) (q_qx r)) r').
  constructor. split.
  rewrite q_qx_le_iff.
  rewrite <- h0. apply Qlt_impl_Qle. assumption.
  constructor. assumption.
rewrite h3 in h9.
destruct h9 as [h9]. destruct h9 as [h9 h10].
inversion h10. subst.
apply Qlt_irrefl in H1. contradiction.
assumption.
Qed.

Lemma hoq_invl_qq_eq :
  forall q r q' r':Q,
    (q < r)%Q -> (q' < r')%Q ->
    hoq_invl q r = hoq_invl q' r' ->
    (q == q')%Q /\ (r == r')%Q.
  unfold hoq_invl; intros; apply ho_invl_qq_eq; auto.
Qed.

Lemma ho_invl_pair_inj_ : 
  forall (p r:Qx*Qx),
  Inhabited (ho_invl' p) ->
  Inhabited (ho_invl' r) ->
    ho_invl' p = ho_invl' r ->
    fst p == fst r /\
    snd p == snd r.
intros p r hp hr h2.  
destruct p as [pa pb], r as [ra rb]. simpl.
unfold ho_invl' in h2, hp, hr. simpl in h2, hp, hr.
rewrite inh_ho_invl_iff in hp, hr.
pose proof (qx_dec pa ra) as h3.
destruct h3 as [[h3 |h3] | h3]. 
pose proof (qx_lt_dec _ _ h3) as h4.
destruct h4 as [h4 | h5].
destruct h4 as [p h4].
assert (h5:~Inn (ho_invl ra rb) p).
  intro h6. destruct h6 as [h6].
  destruct h6 as [h6 h7].
  rewrite <- h4 in h6.
  pose proof (trans_qx_lt_le _  _ _ h3 h6) as h8.
  apply qx_lt_irrefl in h8; auto. 
rewrite <- h2 in h5.
contradict h5.
constructor.
rewrite <- h4.
split; auto.
right. 
apply qx_eq_refl. 
subst.  
apply ho_invl_neg_inf_eq_neg_inf in h2.
subst. split; auto. constructor.
apply qx_lt_irrefl in h3. contradiction.
destruct pb.
apply qx_lt_irrefl in hp. contradiction.
apply Inhabited_intro with 0.
constructor. split. left. constructor. constructor.
apply Inhabited_intro with (q - 1).
constructor. split.
left. constructor. constructor. 
apply Qminus_pos_Qlt. constructor. 
split; auto.  
inversion h3. subst.
destruct pb, rb; try constructor; auto.
inversion hp. inversion hp. inversion hr.
apply ho_invl_pos_inf_eq_pos_inf in h2. 
discriminate h2.
apply Inhabited_intro with a. constructor.
split. right. constructor. constructor. constructor.
rewrite ho_invl_neg_inf_eq in h2.
rewrite <- inh_ho_invl_iff in hp.
rewrite h2 in hp.
destruct hp; contradiction.
symmetry in h2.
apply ho_invl_pos_inf_eq_pos_inf in h2. 
discriminate h2. 
apply Inhabited_intro with b. constructor.
split. right. constructor. constructor. constructor. 
apply ho_invl_qq_eq in h2.
destruct h2 as [h2a h2b].
assumption.
inversion hp. subst. assumption.
inversion hr. subst. assumption.
subst.
apply ho_invl_neg_inf_eq2 in h2.
assumption.
subst.
inversion hp.
pose proof (qx_lt_dec _ _ h3) as h4.
destruct h4 as [h4 | h5].
destruct h4 as [p h4]. subst.
pose proof (qx_lt_dec _ _ hr) as h5.
destruct h5 as [h5 | h6].
destruct h5 as [p' h5].
injection h5. intro h6.
subst.
assert (h6:Inn (ho_invl (q_qx p') rb) p').
  constructor. split. right. constructor. constructor.
  assumption.
rewrite <- h2 in h6.
destruct h6 as [h6]. destruct h6 as [h6 h7].
pose proof (trans_qx_lt_le _ _ _ h3 h6) as h8.
apply qx_lt_irrefl in h8. contradiction.
discriminate h6.
subst.
symmetry in h2.
apply ho_invl_neg_inf_eq_neg_inf in h2.
subst.
split. constructor.
inversion h3. 
destruct rb.
inversion hr.
rewrite ho_invl_neg_inf_pos_inf_eq.
apply Inhabited_intro with 0. constructor.
apply Inhabited_intro with (q - 1).
constructor. split.
left. constructor. constructor.
apply Qminus_pos_Qlt.
constructor.
Qed.

Lemma ho_invl_pair_inj : 
  forall (q r s t:Qx),
  Inhabited (ho_invl' (q, r)) ->
  Inhabited (ho_invl' (s, t)) ->
    ho_invl' (q, r) = ho_invl' (s, t) ->
    q == s /\ r == t.
  intros q r s t h1 h2 h3.
  pose proof (ho_invl_pair_inj_ (q, r) (s, t) h1 h2 h3) as h4.
  apply h4.
Qed.


Lemma hoq_invl_pair_inj_ : 
  forall (p q:Q*Q),
  Inhabited (hoq_invl' p) ->
  Inhabited (hoq_invl' q) ->
    hoq_invl' p = hoq_invl' q ->
    (fst p == fst q)%Q /\
    (snd p == snd q)%Q.
  intros p r hp hr h1.
  do 2 rewrite <- q_qx_eq_iff.
  pose proof (ho_invl_pair_inj_ (q_qx (fst p), q_qx (snd p))
                                (q_qx (fst r), q_qx (snd r)) hp hr h1) as h2.
  simpl in h2; auto.
Qed.


Lemma eq_endpnts_not_disjoint : 
  forall (p r:Qx*Qx),
    Inhabited (ho_invl' p) ->
    Inhabited (ho_invl' r) ->
    snd p == fst r ->
    Union (ho_invl' p) (ho_invl' r) = ho_invl' ((fst p), (snd r)).
intros p r h1 h2 h3.
destruct p as [pa pb], r as [ra rb]. simpl in h3.
destruct pb, ra.
destruct h1 as [x h1]. destruct h1 as [h1].
simpl in h1. destruct h1 as [h1 h4]. inversion h4.
inversion h3. inversion h3. inversion h3.
destruct h2 as [x h2]. destruct h2 as [h2]. simpl in h2.
destruct h2 as [h2]. destruct h2 as [h2|h2]. inversion h2.
inversion h2. inversion h3. inversion h3. inversion h3.
inversion h3. 
subst. simpl.
apply Extensionality_Ensembles; red; split; red; intros x h4; simpl.
destruct h4 as [x h4|x h4].
simpl in h4. 
unfold ho_invl' in h1, h2.
rewrite inh_ho_invl_iff in h1, h2.
simpl in h1, h2.
destruct h4 as [h4]. destruct h4 as [h4 h5].
constructor. split; auto.
eapply trans_qx_lt. apply h5.
eapply trans_qx_eq_lt. apply h3. apply h2.
simpl in h4.
destruct h4 as [h4]. destruct h4 as [h4 h5].
constructor.
unfold ho_invl' in h1, h2.
rewrite inh_ho_invl_iff in h1, h2. simpl in h1, h2. split.
pose proof (trans_qx_eq_le _ _ _ h3 h4) as h6.
left.
eapply trans_qx_lt_le. apply h1. assumption. assumption. 
simpl in h4.  
destruct h4 as [h4]. destruct h4 as [h4 h5].
destruct (qx_dec (q_qx x) (q_qx q)) as [h6 | h6].
destruct h6 as [h6 | h6].
left. constructor. split; auto.
pose proof (trans_qx_eq _ _ _ h6 h3) as h7.
right. constructor. split.
right. apply sym_qx_eq. assumption. assumption.
right. constructor.
split.
apply sym_qx_eq in h3.
left.
eapply trans_qx_eq_lt. apply h3. 
assumption.
assumption.
Qed.


Lemma disjoint_ho_invl_iff : 
  forall (p r:Qx*Qx),
    Inhabited (ho_invl' p) ->
    Inhabited (ho_invl' r) ->
    (disjoint (ho_invl' p) (ho_invl' r) <->
    snd p <= fst r \/ snd r <= fst p).
intros p r hp hr. split. 
intro h1. apply NNPP.
intro h2.
apply not_or_and in h2.
destruct h2 as [h2 h3].
rewrite not_qx_le_iff in h2, h3.
unfold ho_invl' in hp, hr. 
rewrite inh_ho_invl_iff in hp, hr.  
pose proof (qx_2lt_max _ _ _ h3 hr) as h5.
apply qx_lt_dec in h5. 
destruct h5 as [h5 | h5].
destruct h5 as [q h5]. 
pose proof (le_qx_max (fst p) (fst r)) as h6.
destruct h6 as [h6 h7].
contradict h1. unfold disjoint.
apply Inhabited_not_empty.
apply Inhabited_intro with q.
constructor. constructor. split.
rewrite <- h5; auto.
rewrite <- h5.
unfold qx_max.
destruct qx_dec; auto.
constructor. split. rewrite <- h5; auto.
rewrite <- h5.
unfold qx_max. destruct qx_dec; auto.
pose proof (le_qx_max (fst p) (fst r)) as h6.
rewrite h5 in h6.
destruct h6 as [h6 h7]. destruct h6 as [h6 | h6].
inversion h6.
destruct h7 as [h7 | h7]. inversion h7.
contradict h1. unfold disjoint.
apply Inhabited_not_empty.
pose proof (qx_min_le (snd p) (snd r)) as h8.
destruct h8 as [h8 h9].
destruct h8 as [h8 | h8]. destruct h9 as [h9 | h9]. 
pose proof (qx_lt_dec (qx_min (snd p) (snd r)) _ h8) as h10.
destruct h10 as [h10 | h10]. destruct h10 as [q h10].
apply Inhabited_intro with (q -1). constructor.
constructor. split.
assert (h11:neg_inf < q_qx (q - 1)). constructor. 
pose proof (trans_qx_eq_lt _ _ _ h6 h11) as h12.
left.
assert (h13:1 > 0). constructor.
pose proof (Qminus_pos_Qlt 1 q h13) as h14. 
assumption.
pose proof (qx_min_le (snd p) (snd r)) as h11.
destruct h11 as [h11 h12].
assert (h13:1 > 0). constructor.
assert (h15:q_qx (q - 1) < q_qx q). 
  pose proof (Qminus_pos_Qlt 1 q h13) as h14. 
  constructor; auto.
rewrite <- h10 in h15.
eapply trans_qx_lt_le. apply h15; auto.
assumption.
constructor. split.
left.
eapply trans_qx_eq_lt. apply h7. constructor.
assert (h13:1 > 0). constructor.
assert (h15:q_qx (q - 1) < q_qx q). 
  pose proof (Qminus_pos_Qlt 1 q h13) as h14. 
  constructor; auto.
pose proof (qx_min_le (snd p) (snd r)) as h11.
destruct h11 as [h11 h12].
eapply trans_qx_lt. apply h15. rewrite <- h10. assumption.
unfold qx_min in h10.
destruct qx_dec as [h11 | h11].
rewrite h10 in h2.
apply sym_qx_eq in h7.
pose proof (trans_qx_eq_lt _ _ _ h7 h2) as h12.
inversion h12.
rewrite h10 in h3.
inversion h3.
assert (h10:snd r < snd p).
  apply sym_qx_eq in h9.
  eapply trans_qx_eq_lt. apply h9. assumption. 
pose proof (qx_lt_dec _  _ h10) as h11.
destruct h11 as [h11 | h11].
destruct h11 as [q h11].
apply Inhabited_intro with (q - 1).
constructor. constructor. split.
left. 
eapply trans_qx_eq_lt. apply h6. constructor.
assert (h13:1 > 0). constructor.
assert (h15:q_qx (q - 1) < q_qx q). 
  pose proof (Qminus_pos_Qlt 1 q h13) as h14. 
  constructor; auto.
rewrite <- h11 in h15.
eapply trans_qx_lt. apply h15. assumption.
constructor. split.
left. eapply trans_qx_eq_lt. apply h7. constructor.
rewrite h11. constructor. apply Qminus_pos_Qlt. constructor.
rewrite h11 in hr.
apply sym_qx_eq in h7.
pose proof (trans_qx_eq_lt _ _ _ h7 hr) as h12.
inversion h12. 
destruct h9 as [h9 | h9].
assert (h10:snd p < snd r).
  apply sym_qx_eq in h8.
  eapply trans_qx_eq_lt. apply h8. assumption.
pose proof (qx_lt_dec _  _ h10) as h11.
destruct h11 as [h11 | h11].
destruct h11 as [q h11].
apply Inhabited_intro with (q - 1).
constructor. constructor. split. 
left. 
eapply trans_qx_eq_lt. apply h6. constructor.
assert (h13:1 > 0). constructor.
assert (h15:q_qx (q - 1) < q_qx q). 
  pose proof (Qminus_pos_Qlt 1 q h13) as h14. 
  constructor; auto. 
rewrite <- h11 in h15.
assumption.
constructor. split. 
left. eapply trans_qx_eq_lt. apply h7. constructor.
assert (h13:1 > 0). constructor.
assert (h15:q_qx (q - 1) < q_qx q). 
  pose proof (Qminus_pos_Qlt 1 q h13) as h14. 
  constructor; auto.
rewrite <- h11 in h15.
eapply trans_qx_lt. apply h15. assumption.
rewrite h11 in hp.
inversion hp.
assert (h10:snd p == snd r).
  apply sym_qx_eq in h8.
  pose proof (trans_qx_eq _ _ _ h8 h9).
  assumption. 
destruct p as [pa pb], r as [ra rb].
simpl in hp, hr, h2, h3, h5, h6, h7, h8, h9, h10.
destruct pa, pb, ra, rb; try inversion h10; try inversion hp; try inversion h2; try inversion h7; try inversion h6; subst.
apply Inhabited_intro with 0.
constructor; constructor; simpl. split.
left; constructor. constructor. split.
left; constructor. constructor. 
apply Inhabited_intro with (q - 1).
constructor.
constructor; simpl; split. left; constructor.
assert (h13:1 > 0). constructor.
assert (h15:q_qx (q - 1) < q_qx q). 
  pose proof (Qminus_pos_Qlt 1 q h13) as h14. 
  constructor; auto.
assumption.
constructor; simpl; split. left; constructor.
assert (h13:1 > 0). constructor.
pose proof (Qminus_pos_Qlt 1 q h13) as h14. 
constructor.
rewrite <- H1.
assumption.


intro h1. red.
apply Extensionality_Ensembles; red; split; red; intros x h2; try contradiction.
destruct h2 as [x h2 h3].
destruct h2 as [h2], h3 as [h3].
destruct h2 as [h2a h2b], h3 as [h3a h3b].
destruct h1 as [h1 | h1].
pose proof (trans_qx_le_lt _ _ _ h3a h2b) as h4.
pose proof (trans_qx_le_lt _ _ _ h1 h4) as h5.
apply qx_lt_irrefl in h5.
contradiction.
pose proof (trans_qx_le_lt _ _ _ h2a h3b) as h4.
pose proof (trans_qx_lt_le _ _ _ h4 h1) as h5.
apply qx_lt_irrefl in h5.
contradiction.
Qed.

Lemma disjoint_fst_snd_compat_le : 
  forall (p q:Qx*Qx),
    Inhabited (ho_invl' p) -> Inhabited (ho_invl' q) ->
    disjoint (ho_invl' p) (ho_invl' q) ->
    (fst p <= fst q <-> snd p <= snd q).
intros p q h1 h2 h3.
rewrite disjoint_ho_invl_iff in h3; auto.
apply inh_ho_invl_iff in h1.  apply inh_ho_invl_iff in h2.
split.
intro h4.
destruct h3 as [h3 | h3].
left.
eapply trans_qx_le_lt. apply h3. assumption.
pose proof (trans_qx_le_lt _ _ _ h4 h2) as h5.
pose proof (trans_qx_lt_le _ _ _ h5 h3) as h6.
apply qx_lt_irrefl in h6. contradiction.
intro h4.
destruct h3 as [h3 | h3].
left. 
eapply trans_qx_lt_le. apply h1. assumption.
pose proof (trans_qx_le _ _ _ h4 h3) as h5.
pose proof (trans_qx_le_lt _ _ _ h5 h1) as h6.
apply qx_lt_irrefl in h6.
contradiction.
Qed.

Lemma disjoint_fst_snd_compat_lt : 
  forall (p q:Qx*Qx),
    Inhabited (ho_invl' p) -> Inhabited (ho_invl' q) ->
    disjoint (ho_invl' p) (ho_invl' q) ->
    (fst p < fst q <-> snd p < snd q).
intros p q h1 h2 h3.
rewrite disjoint_ho_invl_iff in h3; auto.
apply inh_ho_invl_iff in h1.  apply inh_ho_invl_iff in h2.
split.
intro h4. 
destruct h3 as [h3 | h3].
eapply trans_qx_le_lt. apply h3. assumption.
pose proof (trans_qx_lt_le _ _ _ h2 h3) as h5.
pose proof (trans_qx_lt _ _ _ h5 h4) as h6.
apply qx_lt_irrefl in h6. contradiction.
intro h4.
destruct h3 as [h3 | h3].
eapply trans_qx_lt_le. apply h1. assumption.
pose proof (trans_qx_lt_le _ _ _ h4 h3) as h5.
pose proof (trans_qx_lt _ _ _ h5 h1) as h6.
eapply qx_lt_irrefl in h6.
contradiction.
Qed.


Definition overlapping_ho_pair_join (p r:Qx*Qx) : Qx*Qx:=
  ((qx_min (fst p) (fst r)), (qx_max (snd p) (snd r))).
 
Definition overlapping_ho_pair_meet (p r:Qx*Qx) : Qx *Qx :=
  ((qx_max (fst p) (fst r)), (qx_min (snd p) (snd r))).


Lemma ho_invl_union : 
  forall (p r:Qx*Qx),
    Inhabited (ho_invl' p) ->
    Inhabited (ho_invl' r) ->
    ~disjoint (ho_invl' p) (ho_invl' r) ->
    Union (ho_invl' p) (ho_invl' r) = 
    ho_invl' (overlapping_ho_pair_join p r).
intros p r h1 h2 h3.
simpl.
apply Extensionality_Ensembles; red; split; red; intros x h4.  
destruct h4 as [x h4 | x h4].
destruct h4 as [h4]. constructor.
pose proof (qx_min_le (fst p) (fst r)) as h5.
destruct h5 as [h5 h6].
destruct h4 as [h4a h4b].
split.
eapply trans_qx_le. apply h5. apply h4a.
pose proof (le_qx_max (snd p) (snd r)) as h7.
destruct h7 as [h7 h8].
eapply trans_qx_lt_le. apply h4b. assumption.
destruct h4 as [h4]. destruct h4 as [h4 h5].
constructor. split.
pose proof (qx_min_le (fst p) (fst r)) as h6.
destruct h6 as [h6 h7].
eapply trans_qx_le. apply h7. assumption.
pose proof (le_qx_max (snd p) (snd r)) as h6.
destruct h6 as [h6 h7].
eapply trans_qx_lt_le. apply h5. assumption. 
rewrite disjoint_ho_invl_iff in h3; auto.
apply not_or_and in h3.
destruct h3 as [h3a h3b].
rewrite  not_qx_le_iff in h3a.
rewrite  not_qx_le_iff in h3b. 
apply inh_ho_invl_iff in h1. apply inh_ho_invl_iff in h2. 
destruct h4 as [h4]. simpl in h4.
destruct h4 as [h4 h5].
destruct (classic (Inn (ho_invl' (fst r, snd p)) x)) as [h6 | h6].
destruct h6 as [h6]. simpl in h6.
destruct h6 as [h6 h7].
unfold qx_min in h4. unfold qx_max in h5.

destruct (qx_dec (snd p) (snd r)) as [h8 | h8].
right. constructor; auto.
destruct qx_dec as [h9 | h9].
apply sumbool_to_or in h9.
left. constructor. split; auto.
clear h7. clear h6.
destruct (classic (Inn (ho_invl' r) x)) as [h6 | h6].
right; auto.
unfold ho_invl', ho_invl in h6.
apply sat_nin in h6.
apply not_and_or in h6.
rewrite not_qx_le_iff in h6.
destruct h6 as [h6 | h6].
pose proof (trans_qx_le_lt _ _ _ h4 h6) as h7.
apply qx_lt_irrefl in h7. contradiction.
rewrite not_qx_lt_iff in h6.
pose proof (trans_qx_lt_le _ _ _ h3b h6) as h10. 
left. constructor.
split; auto. 
unfold ho_invl', ho_invl in h6. simpl in h6.
apply sat_nin in h6.
apply not_and_or in h6.
rewrite not_qx_le_iff in h6. rewrite not_qx_lt_iff in h6.
unfold qx_max in h5.
destruct qx_dec as [h7 | h7].
apply sumbool_to_or in h7.
unfold qx_min in h4.
destruct qx_dec as [h8 | h8].
apply sumbool_to_or in h8.
destruct h6 as [h6 | h6].
pose proof (trans_qx_lt _ _ _ h6 h3a) as h9.
left. constructor; split; auto.
right. constructor. split.
left. eapply trans_qx_lt_le. apply h3a. apply h6.
assumption.
destruct h6 as [h6 | h6].
pose proof (trans_qx_lt_le _ _ _ h6 h4) as h9.
apply qx_lt_irrefl in h9. contradiction.
right. constructor; split; auto.
unfold qx_min in h4.
destruct qx_dec as [h8 | h8].
left; split; auto.
destruct h6 as [h6 | h6].
pose proof (trans_qx_lt_le _ _ _ h6 h4) as h9.
apply qx_lt_irrefl in h9. contradiction.
pose proof (trans_qx_lt_le _ _ _ h5 h6) as h9.
apply qx_lt_irrefl in h9.
contradiction.
Qed.

Lemma  strict_included_neg_inf_q :
  forall q:Q,
    Strict_Included (ho_invl' (neg_inf, q_qx q)) (ho_invl' (neg_inf, pos_inf)).
intro q.
red; split.
red; intros x h1.
destruct h1 as [h1]. destruct h1 as [h1 h2].
simpl in h1, h2.
constructor. split; auto. simpl.
constructor.
intro h1.
assert (h2:Inn (ho_invl' (neg_inf, pos_inf)) q).
  constructor. simpl. split. left; constructor. constructor.
rewrite <- h1 in h2.
destruct h2 as [h2]. simpl in h2.
destruct h2 as [h2 h3].
apply qx_lt_irrefl in h3.
assumption.
Qed.

Lemma strict_included_neg_inf_minus : 
  forall q:Q,
    Strict_Included (ho_invl' (neg_inf, q_qx (q - 1)))
                    (ho_invl' (neg_inf, q_qx q)).
intro q.
red; split.
red; intros x h1.
destruct h1 as [h1]. simpl in h1. destruct h1 as [h1 h2].
assert (h3:q_qx (q - 1) < q_qx q).
  constructor.
  apply Qminus_pos_Qlt.
  constructor.
constructor. simpl.
split; auto.
eapply trans_qx_lt. apply h2. assumption.
intro h1.
assert (h2:Inn (ho_invl' (neg_inf, q_qx q)) (q - 1)).
  constructor. simpl; split. left. constructor.
  constructor.
  apply Qminus_pos_Qlt.
  constructor.
rewrite <- h1 in h2.
inversion h2 as [h3].
destruct h3 as [h3 h4]. simpl in h4.
apply qx_lt_irrefl in h4.
assumption.
Qed.

Lemma strict_included_Sq_pos_inf : 
  forall (p:Q),
    Strict_Included (ho_invl' (q_qx p, q_qx (p+1)))
                    (ho_invl' (q_qx p, pos_inf)).
intro p.
red; split.
red; intros x h1.
destruct h1 as [h1]. simpl in h1. destruct h1 as [h1 h2].
constructor. simpl. split; auto. constructor. 
intro h1.
apply ho_invl_pair_inj_ in h1.
simpl in h1.
destruct h1 as [h1 h2].
inversion h2.
apply Inhabited_intro with p.
constructor. simpl. split.
right; constructor. constructor.
constructor. 
apply Qlt_plus_pos. constructor.
apply Inhabited_intro with p.
constructor. simpl. split. right. constructor; constructor.
constructor.
Qed.

Lemma strict_included_midpoint : 
  forall (p q:Q),
    (p < q)%Q ->
    Strict_Included (ho_invl' (q_qx p, q_qx (p + (q - p) / (2 # 1))))
                    (ho_invl' (q_qx p, q_qx q)).
intros p q h1.
pose proof (midpoint_Qlt p q h1) as h4. 
red. split.
red. intros x h2.
destruct h2 as [h2]. simpl in h2.
destruct h2 as [h2 h3].
constructor. simpl. split; auto.
eapply trans_qx_lt. apply h3. 
constructor.  assumption.
intro h2.
apply ho_invl_pair_inj_ in h2.
simpl in h2. destruct h2 as [h2 h3].
inversion h3. subst.
rewrite H1 in h4.
apply Qlt_irrefl in h4.
assumption.
unfold ho_invl'.
rewrite inh_ho_invl_iff.
simpl. constructor.
apply Qlt_midpoint.
assumption.
unfold ho_invl'.
rewrite inh_ho_invl_iff. simpl.
constructor. 
assumption.
Qed.


Lemma strict_included_midpoint_q : 
  forall (p q:Q),
    (p < q)%Q ->
    Strict_Included (hoq_invl' (p, (p + (q - p) / (2 # 1))))
                    (hoq_invl' (p, q)).
  intros; rewrite hoq_invl_compat'; apply strict_included_midpoint; auto.
Qed.


Lemma ho_invl_ex_strict_included : 
  forall p:Qx*Qx,
    Inhabited (ho_invl' p) ->
  exists q:Qx*Qx,
    Strict_Included (ho_invl' q) (ho_invl' p) /\
    Inhabited (ho_invl' q).
intros p h1.
destruct p as [pa pb].
unfold ho_invl' in h1.
destruct pa, pb; simpl in h1; try rewrite ho_invl_neg_inf_eq in h1.
apply Inhabited_not_empty in h1.
contradict h1; auto.
exists (neg_inf, q_qx 0). 
split.
apply strict_included_neg_inf_q.
unfold ho_invl'; rewrite inh_ho_invl_iff; simpl; constructor.
exists (neg_inf, (q_qx (q - 1))).
split.
apply strict_included_neg_inf_minus.
unfold ho_invl'; rewrite inh_ho_invl_iff; simpl; constructor.
apply Inhabited_not_empty in h1.
contradict h1. auto.
rewrite ho_invl_pos_inf_eq in h1.
apply Inhabited_not_empty in h1.
contradict h1. auto.
rewrite ho_invl_pos_inf_eq in h1.
apply Inhabited_not_empty in h1.
contradict h1. auto.
apply Inhabited_not_empty in h1.
contradict h1. auto. 
exists (q_qx q, q_qx (q+1)).
split.
apply strict_included_Sq_pos_inf. 
unfold ho_invl'; rewrite inh_ho_invl_iff; simpl; constructor.
apply Qlt_plus_pos. constructor.
apply inh_ho_invl_iff in h1.
inversion h1. subst.
exists (q_qx q, q_qx (Qplus q (Qdiv (Qminus q0 q) (Qmake 2 1)))).
split.
apply strict_included_midpoint; auto.
apply inh_ho_invl_iff. simpl.
constructor.
apply Qlt_midpoint.
assumption.
Qed.



Definition ho_pairs_to_ho_invls_ens : list (Qx*Qx) -> Ensemble Q :=
  (fun l => 
     (FamilyUnion 
        (Im (list_to_set l)
            (fun p => ho_invl' p)))).

Lemma ho_pairs_to_ho_invls_ens_nil : 
  ho_pairs_to_ho_invls_ens nil = Empty_set _.
unfold ho_pairs_to_ho_invls_ens; simpl.
rewrite image_empty.
rewrite empty_family_union.
reflexivity.
Qed.


Lemma ho_pairs_to_ho_invls_ens_cons :
  forall (l:list (Qx*Qx)) (p:Qx*Qx),
    ho_pairs_to_ho_invls_ens (p::l) = 
    Union (ho_invl' p)  (ho_pairs_to_ho_invls_ens l).
intros l p.
unfold ho_pairs_to_ho_invls_ens.
simpl.
rewrite Im_add.
rewrite family_union_add.
unfold ho_invl.
reflexivity.
Qed.

Lemma ho_pairs_to_ho_invls_ens_sing : 
  forall (p:Qx*Qx),
    ho_pairs_to_ho_invls_ens (p::nil) =
    ho_invl' p.
intro p.
rewrite (ho_pairs_to_ho_invls_ens_cons nil p).
rewrite ho_pairs_to_ho_invls_ens_nil.
rewrite Union_commutative.
rewrite empty_union.
reflexivity.
Qed.



Lemma ho_pairs_to_ho_invls_ens_preserves_app :
  forall (lp lq:list (Qx*Qx)), 
    ho_pairs_to_ho_invls_ens (lp ++ lq) =
    Union (ho_pairs_to_ho_invls_ens lp) (ho_pairs_to_ho_invls_ens lq).
intros lp lq.
unfold ho_pairs_to_ho_invls_ens.
rewrite list_to_set_app.
rewrite im_union.
rewrite familyunion_union_eq.
reflexivity.
Qed.

Lemma inh_ho_pairs_to_ho_invls_ens_impl : 
  forall l:list (Qx*Qx),
    Inhabited (ho_pairs_to_ho_invls_ens l) ->
    exists p:Qx*Qx,
      In p l /\ Inhabited (ho_invl' p).
intros l h1.
destruct h1 as [q h1].
destruct h1 as [S q h3 h4].
destruct h3 as [p h3]. subst.
exists p. split.
rewrite <- list_to_set_in_iff in h3.
assumption.
apply Inhabited_intro with q; auto.
Qed.


Inductive union_of_ho_invls : Ensemble Q -> Prop :=
  | union_of_ho_invls_intro : 
      forall l:list (Qx*Qx),
        union_of_ho_invls (ho_pairs_to_ho_invls_ens l).


Lemma ho_invl_incl_union_of_ho_invls : 
  forall (l:list (Qx * Qx)) (p:Qx*Qx),
    In p l ->
    Included (ho_invl' p) (ho_pairs_to_ho_invls_ens l).
intros l p h1.
red; intros x h2.
unfold ho_pairs_to_ho_invls_ens.
eapply family_union_intro.
apply Im_intro with p.
rewrite <- list_to_set_in_iff; auto.
apply eq_refl.
assumption.
Qed.


Lemma union_of_ho_invls_ex_strict_included  : 
  forall S:Ensemble Q,
    Inhabited S ->
    union_of_ho_invls S ->
  exists S':Ensemble Q,
    Strict_Included S' S /\
    union_of_ho_invls S' /\
    Inhabited S'.
intros S h1 h2.
destruct h2 as [l].
destruct 
  (classic (exists p:Qx*Qx, In p l /\ ho_invl' p <>
                                      ho_pairs_to_ho_invls_ens l
  /\ Inhabited (ho_invl' p))) as [h2 | h2].
destruct h2 as [p h2]. destruct h2 as [h2 [h3 h3']].
exists (ho_invl' p). split.
split.
red; intros x h4.
unfold ho_pairs_to_ho_invls_ens.
apply family_union_intro with (ho_invl' p).
apply Im_intro with p. rewrite <- list_to_set_in_iff.
assumption. reflexivity. assumption. assumption.  
split.
rewrite <- ho_pairs_to_ho_invls_ens_sing.  
constructor. assumption.
destruct (nil_dec l) as [h3 | h3].
subst.
apply Inhabited_not_empty in h1. 
rewrite ho_pairs_to_ho_invls_ens_nil in h1.
contradict h1; auto. 
pose proof (inh_ho_pairs_to_ho_invls_ens_impl _ h1) as ha.
destruct ha as [p [ha hb]].
pose proof (not_ex_all_not _ _ h2 p) as h6.
simpl in h6.
apply not_and_or in h6. 
destruct h6 as [|h6]; try contradiction.
apply not_and_or in h6. 
destruct h6 as [h6|h6]; try contradiction.
apply NNPP in h6.
rewrite <- h6.
rewrite <- h6 in h1.
pose proof (ho_invl_ex_strict_included _ h1) as h7.
destruct h7 as [q [h7 h8]].
exists (ho_invl' q).
split; auto. split.
rewrite <- ho_pairs_to_ho_invls_ens_sing. 
constructor.
assumption.
Qed.



Definition int_alg_t := {S:Ensemble Q | union_of_ho_invls S}.

Lemma int_alg_t_plus_sig : 
  forall (A B:int_alg_t),
    union_of_ho_invls (Union (proj1_sig A) (proj1_sig B)).
intros A B.
destruct A as [A h1], B as [B h2].
destruct h1 as [la], h2 as [lb]. simpl.
rewrite <- ho_pairs_to_ho_invls_ens_preserves_app.
constructor.
Qed.

Definition int_alg_t_plus (A B:int_alg_t) :=
  exist _ _ (int_alg_t_plus_sig A B).

Lemma intersection_ho_invls : 
  forall (p q:Qx*Qx),
    let S := Intersection (ho_invl' p) (ho_invl' q) in
    S = Empty_set _  \/ S = ho_invl' (overlapping_ho_pair_meet p q).
intros p q S.
cut (S <> Empty_set Q -> S = ho_invl' (overlapping_ho_pair_meet p q)).
tauto.
intro h1.
unfold S in h1.
assert (h2:~disjoint (ho_invl' p) (ho_invl' q)).
unfold disjoint. assumption.
destruct (ens_eq_dec _ (ho_invl' p) (Empty_set _)) as [h3 | h3].
contradict h2. red. rewrite h3. rewrite empty_intersection.
reflexivity.
destruct (ens_eq_dec _ (ho_invl' q) (Empty_set _)) as [h4 | h4].
contradict h2. rewrite h4. red. rewrite Intersection_commutative.
rewrite empty_intersection. reflexivity.
apply not_empty_Inhabited in h3.  apply not_empty_Inhabited in h4.
rewrite disjoint_ho_invl_iff in h2; auto.
apply not_or_and in h2.
destruct h2 as [h2 h5].
rewrite not_qx_le_iff in h2, h5.
apply Extensionality_Ensembles; red; split; red; intros x h6.
constructor.
unfold overlapping_ho_pair_meet.  
destruct h6 as [x h6 h7].
destruct h6 as [h6], h7 as [h7].
destruct h6 as [h6a h6b], h7 as [h7a h7b].
split; simpl.
apply qx_2le_max; auto.
apply qx_2lt_min; auto.
destruct h6 as [h6]. destruct h6 as [h6 h7]. 
unfold overlapping_ho_pair_meet, qx_max, qx_min in h6, h7.
simpl in h6, h7. 
destruct qx_dec as [h8 | h8].
destruct qx_dec as [h9 | h9].
constructor; constructor.
split; auto. 
apply sumbool_to_or in h8.
apply sumbool_to_or in h9.
eapply trans_qx_le. apply h8. assumption.
apply sumbool_to_or in h8.
apply sumbool_to_or in h9.
split; auto.
eapply trans_qx_lt_le. apply h7. assumption.
apply sumbool_to_or in h8.
constructor; constructor.
split; auto.
eapply trans_qx_le. apply h8. assumption.
eapply trans_qx_lt. apply h7. assumption. 
split; auto.
destruct qx_dec as [h9 | h9].
apply sumbool_to_or in h9.
constructor; constructor.
split; auto. split; auto.
left.
eapply trans_qx_lt_le. apply h8. assumption.
eapply trans_qx_lt_le. apply h7. assumption.
constructor; constructor.
split; auto.
eapply trans_qx_lt. apply h7. assumption.
split; auto.
left.
eapply trans_qx_lt_le.
apply h8. assumption.
Qed.

Lemma union_of_ho_invls_union : 
  forall (p:Qx*Qx) (S:Ensemble Q),
    union_of_ho_invls S ->
    union_of_ho_invls (Union (ho_invl' p) S).
intros p S h1.
destruct h1.
rewrite <- ho_pairs_to_ho_invls_ens_cons.
constructor.
Qed.


Lemma union_of_ho_invls_intersection_ho_invl_int_alg_t : 
  forall (p:Qx*Qx) (S:int_alg_t),
    union_of_ho_invls (Intersection (ho_invl' p) (proj1_sig S)).
intros p S.
destruct S as [S h1]. simpl.
destruct h1 as [l].
induction l as [|a l h1]; simpl. 
rewrite ho_pairs_to_ho_invls_ens_nil.
rewrite Intersection_commutative. rewrite empty_intersection.
rewrite <- ho_pairs_to_ho_invls_ens_nil.
constructor.
rewrite ho_pairs_to_ho_invls_ens_cons.
rewrite Distributivity.
pose proof (intersection_ho_invls p a) as h2. simpl in h2.
destruct h2 as [h2 | h2].
rewrite h2.
rewrite empty_union.
assumption.
rewrite h2.
apply union_of_ho_invls_union; auto.
Qed.

Lemma union_of_ho_invls_family : 
  forall (F:Family Q),
    Finite F ->
    (forall S, Inn F S -> union_of_ho_invls S) ->
    union_of_ho_invls (FamilyUnion F).
intros F h1.
induction h1 as [|F h2 h3 A h4]; simpl; auto.
intros h2.
rewrite empty_family_union.
rewrite <- ho_pairs_to_ho_invls_ens_nil.
constructor.
intros h5.
assert (h6:(forall S : Ensemble Q, Inn F S -> union_of_ho_invls S)).
  intros S h6. apply h5. left. assumption.
specialize (h3 h6).
rewrite family_union_add.
assert (h7:Inn (Add F A) A). right. constructor; auto.
specialize (h5 _ h7).
inversion h3 as [l h8].
destruct h5 as [l'].
rewrite <- ho_pairs_to_ho_invls_ens_preserves_app.
constructor.
Qed.


Lemma int_alg_t_times_sig : 
  forall (R S:int_alg_t),
    union_of_ho_invls (Intersection (proj1_sig R) (proj1_sig S)).
intros R S.
destruct R as [R h1].
simpl.
destruct h1 as [l].
unfold ho_pairs_to_ho_invls_ens.
rewrite Intersection_commutative.
rewrite dist_union_infnt_1.
unfold Int_Fam_Ens.
rewrite im_im. 
assert (h1:forall A,
             Inn (Im (list_to_set l)
           (fun x : Qx * Qx => Intersection (proj1_sig S) (ho_invl' x))) A ->
             union_of_ho_invls A).
  intros A h2.
  destruct h2 as [A h2]. subst.
  rewrite Intersection_commutative.
  apply union_of_ho_invls_intersection_ho_invl_int_alg_t. 
apply union_of_ho_invls_family.
apply finite_image.
apply list_to_set_finite.
assumption.
Qed.

Definition int_alg_t_times (R S:int_alg_t) :=
  exist _ _ (int_alg_t_times_sig R S).

Lemma union_of_ho_invls_full : 
  union_of_ho_invls (Full_set Q).
assert (h1:Full_set Q = ho_pairs_to_ho_invls_ens ((neg_inf, pos_inf) :: nil)).
  unfold ho_pairs_to_ho_invls_ens.
  apply Extensionality_Ensembles; red; split; red; intros x h1.
  eapply family_union_intro.
  simpl.
  rewrite Im_add.
  rewrite image_empty.
  rewrite add_empty_sing.
  unfold ho_invl'.  simpl.
  rewrite ho_invl_neg_inf_pos_inf_eq.
  econstructor. assumption.
  constructor.
rewrite h1; constructor.
Qed.


Lemma union_of_ho_invls_empty : 
  union_of_ho_invls (Empty_set Q).
assert (h1:Empty_set Q = ho_pairs_to_ho_invls_ens nil).
  rewrite ho_pairs_to_ho_invls_ens_nil.
  reflexivity.
rewrite h1.
constructor.
Qed.

Lemma union_of_ho_invls_ho_invl : 
  forall (p q:Qx),
    union_of_ho_invls (ho_invl p q).
intros p q.
assert (h1:ho_invl p q = ho_pairs_to_ho_invls_ens ((p, q)::nil)).
unfold ho_pairs_to_ho_invls_ens.
simpl.
rewrite Im_add.
rewrite image_empty.
rewrite add_empty_sing.
rewrite family_union_sing.
reflexivity.
rewrite h1.
constructor.
Qed.

Lemma union_of_ho_invls_union_ho_invl : 
  forall (p p' q q':Qx),
    union_of_ho_invls (Union (ho_invl p q) (ho_invl p' q')).
intros p p' q q'.
assert (h1: Union (ho_invl p q) (ho_invl p' q') = 
            ho_pairs_to_ho_invls_ens ((p, q)::(p', q')::nil)).
unfold ho_pairs_to_ho_invls_ens.
simpl.
rewrite <- couple_add_add.
rewrite im_couple.
rewrite family_union_couple.
rewrite Union_commutative.
reflexivity.
rewrite h1.
constructor.
Qed.



Lemma comp_ho_invl_neg_inf_q_qx : 
  forall q:Q,
    Comp (ho_invl neg_inf (q_qx q)) = 
    ho_invl (q_qx q) pos_inf.
intro q.
apply Extensionality_Ensembles; red; split; red; intros x h1.
red in h1. red in h1.
unfold ho_invl in h1.
apply sat_nin in h1.
apply not_and_or in h1.
rewrite not_qx_le_iff in h1.
destruct h1 as [h1 | h1].
inversion h1.
apply not_qx_lt_iff in h1.
constructor; constructor; auto. constructor.
inversion h1 as [h2].
destruct h2 as [h2 h3].
red. red.
intro h4.
inversion h4 as [h5].
destruct h5 as [h5 h6].
pose proof (trans_qx_lt_le _ _ _ h6 h2) as h7.
apply qx_lt_irrefl in h7.
contradiction.
Qed.

Lemma comp_ho_invl_q_qx_pos_inf :
  forall q:Q,
    Comp (ho_invl (q_qx q) pos_inf) =
    ho_invl neg_inf (q_qx q).
intro q.
rewrite <- Complement_Complement.
f_equal.
symmetry.
apply comp_ho_invl_neg_inf_q_qx.
Qed.

Lemma comp_ho_invls_q_qx_q_qx : 
  forall q q':Q,
    Comp (ho_invl (q_qx q) (q_qx q')) = 
    Union (ho_invl neg_inf (q_qx q)) (ho_invl (q_qx q') pos_inf).
intros q q'.
apply Extensionality_Ensembles; red; split; red; intros x h1.
red in h1. red in h1. unfold ho_invl in h1.
apply sat_nin in h1.
apply not_and_or in h1. 
rewrite not_qx_le_iff, not_qx_lt_iff in h1.
destruct h1 as [h1 | h1].
left. constructor. split; auto.
left. constructor.
right. constructor. split; auto. constructor.
red. red. intro h2.
destruct h2 as [h2]. destruct h2 as [h2 h3]. 
destruct h1 as [x h1 | x h1].
destruct h1 as [h1]. destruct h1 as [h1 h1'].
pose proof (trans_qx_lt_le _ _ _ h1' h2) as h4.
apply qx_lt_irrefl in h4.
assumption.
destruct h1 as [h1]. destruct h1 as [h1 h1'].
pose proof (trans_qx_le_lt _ _ _ h1 h3) as h4.
apply qx_lt_irrefl in h4.
assumption.
Qed.


Lemma union_of_ho_invls_comp : 
  forall p:Qx*Qx,
    union_of_ho_invls (Comp (ho_invl' p)).
intro p. unfold ho_invl'.
destruct p as [pa pb].
destruct pa, pb; simpl; 
try rewrite ho_invl_neg_inf_eq; try rewrite ho_invl_pos_inf_eq; 
try rewrite ho_invl_neg_inf_pos_inf_eq;
try rewrite complement_empty; try rewrite complement_full;
try apply union_of_ho_invls_empty;
try apply union_of_ho_invls_full;
try rewrite comp_ho_invls_q_qx_q_qx;
try rewrite comp_ho_invl_q_qx_pos_inf;
try rewrite comp_ho_invl_neg_inf_q_qx;
try apply union_of_ho_invls_ho_invl;
try apply union_of_ho_invls_union_ho_invl.
Qed.

Lemma int_alg_t_comp_sig : 
  forall S:int_alg_t,
    union_of_ho_invls (Comp (proj1_sig S)).
intros S.
destruct S as [S h1]. destruct h1.
simpl.
induction l as [|a l h1].
rewrite ho_pairs_to_ho_invls_ens_nil.
rewrite complement_empty.
apply union_of_ho_invls_full.
rewrite ho_pairs_to_ho_invls_ens_cons.
rewrite comp_union.
pose proof (union_of_ho_invls_comp a) as h2.
pose proof (int_alg_t_times_sig (exist _ _ h2) (exist _ _ h1)) as h3.
simpl in h3.
assumption.
Qed.

Definition int_alg_t_comp (R:int_alg_t) :=
  exist _ _ (int_alg_t_comp_sig R).


Definition int_alg_bc :=
  Build_Bconst int_alg_t (Full_set int_alg_t)
               int_alg_t_plus int_alg_t_times
               (exist _ _ union_of_ho_invls_full)
               (exist _ _ union_of_ho_invls_empty)
               int_alg_t_comp.

Lemma int_alg_und_set :
   BS int_alg_bc = Full_set (Btype int_alg_bc).
  auto.
Qed.


Lemma int_alg_assoc_sum : 
  forall n m p : Btype int_alg_bc,
     (n + (m + p))%Bool_Alg = (n + m + p)%Bool_Alg.
intros; simpl.
unfold int_alg_t_plus.
apply proj1_sig_injective. simpl.
rewrite Union_associative.
reflexivity.
Qed.


Lemma int_alg_assoc_prod :
   forall n m p : Btype int_alg_bc,
        (n * (m * p))%Bool_Alg = (n * m * p)%Bool_Alg.
intros; simpl.
unfold int_alg_t_times.
apply proj1_sig_injective. simpl.
rewrite Intersection_associative.
reflexivity.
Qed.

Lemma int_alg_comm_sum :
  forall n m : Btype int_alg_bc, (n + m)%Bool_Alg = (m + n)%Bool_Alg.
intros; simpl.
unfold int_alg_t_plus.
apply proj1_sig_injective. simpl.
rewrite Union_commutative.
reflexivity.
Qed.

Lemma int_alg_comm_prod : 
  forall n m : Btype int_alg_bc, (n * m)%Bool_Alg = (m * n)%Bool_Alg.
intros; simpl.
unfold int_alg_t_times.
apply proj1_sig_injective. simpl.
rewrite Intersection_commutative.
reflexivity.
Qed.


Lemma int_alg_abs_sum : 
  forall n m : Btype int_alg_bc, (n + n * m)%Bool_Alg = n.
intros; simpl.
unfold int_alg_t_plus, int_alg_t_times.
apply proj1_sig_injective. simpl.
rewrite abs_sum_psa.
reflexivity.
Qed.


Lemma int_alg_abs_prod : 
  forall n m : Btype int_alg_bc, (n * (n + m))%Bool_Alg = n.
intros; simpl.
unfold int_alg_t_plus, int_alg_t_times.
apply proj1_sig_injective. simpl.
rewrite abs_prod_psa.
reflexivity.
Qed.


Lemma int_alg_dist_sum : 
  forall n m p : Btype int_alg_bc,
    (p * (n + m))%Bool_Alg = (p * n + p * m)%Bool_Alg.
intros; simpl.
unfold int_alg_t_times, int_alg_t_plus.
apply proj1_sig_injective. simpl.
rewrite Distributivity.
reflexivity.
Qed.


Lemma int_alg_dist_prod : 
  forall n m p : Btype int_alg_bc,
    (p + n * m)%Bool_Alg = ((p + n) * (p + m))%Bool_Alg.
intros; simpl.
unfold int_alg_t_times, int_alg_t_plus.
apply proj1_sig_injective. simpl.
rewrite dist_prod_psa.
reflexivity.
Qed.



Lemma int_alg_comp_sum : 
  forall n : Btype int_alg_bc, (n + - n)%Bool_Alg = 1%Bool_Alg.
intros; simpl.
unfold int_alg_t_plus.
apply proj1_sig_injective; simpl.
apply comp_sum_psa.
Qed.      


Lemma int_alg_comp_prod : 
  forall n : Btype int_alg_bc, (n * - n)%Bool_Alg = 0%Bool_Alg.
intros; simpl.
unfold int_alg_t_plus.
apply proj1_sig_injective; simpl.
apply comp_prod_psa.
Qed.      



Definition int_alg_ba := 
  Build_Bool_Alg int_alg_bc int_alg_und_set 
                 int_alg_assoc_sum int_alg_assoc_prod
                 int_alg_comm_sum int_alg_comm_prod
                 int_alg_abs_sum int_alg_abs_prod
                 int_alg_dist_sum int_alg_dist_prod
                 int_alg_comp_sum int_alg_comp_prod.


Lemma int_alg_ba_le_iff : 
  forall (x y:bt int_alg_ba),
    le x y <-> Included (proj1_sig x) (proj1_sig y).
intros x y.
destruct x as [R h1], y as [S h2].
simpl.
split.
intro h3. red in h3.
simpl in h3.
unfold int_alg_t_plus in h3.
apply exist_injective in h3. simpl in h3.
rewrite <- inclusion_iff_union in h3.
assumption.
intro h3.
red. simpl. unfold int_alg_t_plus.
apply proj1_sig_injective; simpl.
rewrite <- inclusion_iff_union.
assumption.
Qed.


Lemma int_alg_ba_lt_iff : 
  forall (x y:bt int_alg_ba),
    lt x y <-> Strict_Included (proj1_sig x) (proj1_sig y).
intros x y.
split.
intro h1.
destruct h1 as [h1 h2].
rewrite int_alg_ba_le_iff in h1.
split; auto.
intro h3. apply proj1_sig_injective in h3.
subst. contradict h2; auto.
intro h1.
destruct h1 as [h1 h2]. 
rewrite <- int_alg_ba_le_iff in h1.
red; split; auto.
intro h3. subst.
contradict h2; auto.
Qed.



Section CountableIntAlg.

Lemma countable_qx : 
  CountableT Qx.
pose proof Q_countable as h1.
pose proof Q_infinite as h2.
pose proof (inf_countable_index _ h1 h2) as h3.
destruct h3 as [f h3].
pose proof h3 as h3'.
apply bijective_impl_invertible in h3'.
pose (proj1_sig (function_inverse _  h3')) as f'.
pose proof (invertible_impl_inv_invertible _  h3') as h4.
pose (fun q:Qx =>
        match q with 
          | neg_inf => 0%nat
          | pos_inf => 1%nat
          | q_qx q' => S (S (f' q'))
        end) as g.
assert (h5:injective g).
  red; intros x y h5.
  unfold g in h5.
  destruct x, y; auto; try omega.
do 2 apply S_inj in h5.
unfold f' in h5.
apply invertible_impl_bijective in h4.
destruct h4 as [h4a h4b].
apply h4a in h5.
subst.
reflexivity.
econstructor. apply h5.
Qed.

Lemma countable_qxp : 
  CountableT (Qx*Qx).
apply countable_product; apply countable_qx.
Qed.

Lemma countable_int_alg_t : 
  CountableT int_alg_t.
pose (fun x:all_tuples (Qx*Qx) => (exist _ _ (union_of_ho_invls_intro (nprod_to_list _ _ (projT2 x))))) as f.
pose proof (countable_all_tuples_of_countable_type _ countable_qxp) as h1.
apply (surj_countable f h1).
red.
intro y.
destruct y as [S h2].
destruct h2 as [l].
unfold f.
exists (existT _ (length l) (nprod_of_list _ l)).
apply proj1_sig_injective. simpl.
rewrite nprod_to_list_nprod_of_list_eq.
reflexivity.
Qed.

Lemma countable_int_alg_ba : 
  countable_ba int_alg_ba. 
apply countable_int_alg_t.
Qed.


End CountableIntAlg.


End IntervalAlgebras.

