(* Copyright (C) 2019, Daniel Wyckoff *)
(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(* This file correspdons to Chapter 2 in Part II of Dietlinde Lau's "Function
Algebras."  The subject is "The Galois Connection between function-
and relation- algebras." *)
 
Require Import ProofIrrelevance.
Require Import FunctionalExtensionality.
Require Import NPeano.
Require Import EqDec.
Require Import FunctionProperties.
Require Import FunctionProperties2.
Require Import ArithUtilities.
Require Import LogicUtilities.
Require Import SetUtilities.
Require Import ListUtilities.

  (*Returns the function that maps j to k, k to j, and fixes all else.
A very complicated function for a simple transposition.  
The more obvious way to do it by destructing on boolean equality 
requires two passes, and perhaps is less efficient, although I include 
that afterwards too.  *)
Fixpoint transpose_seg_fun {n j k} : j < n -> k < n -> seg_fun n nat :=
  match n
  with
    | 0 => fun pf => False_rect _ (lt_n_0 _ pf)
    | S n' =>
      match j with
        | 0 =>
          match k with
            | 0 =>
              fun _ _ i _ => i
            | S k' =>
              fun _ pfk i =>
                let pfk' := lt_S_n _ _ pfk in
                match i with
                  | 0 =>
                    fun _ => S k'
                  | S i' =>
                    match i' with
                      | 0 =>
                        fun _ => 
                          match k' with
                            | 0 => 0 
                            | S _ => 1
                          end
                      | S i'' => 
                        fun pfi =>
                          let pfi' := lt_S_n _ _ pfi in
                          let pf0 := Olt _ _ pfi' in
                          let t := transpose_seg_fun pf0 pfk' (S i'') pfi' in
                          match t with
                            | 0 => 0
                            | S _ => S (S i'')
                          end
                    end
                end
          end
        | S j' =>
          match k with
            | 0 =>
              fun pfj _ i =>
                let pfj' := lt_S_n _ _ pfj in
                match i with
                  | 0 =>
                    fun _ => S j'
                  | S i' =>
                    match i' with
                      | 0 =>
                        fun _ => 
                          match j' with
                            | 0 => 0 
                            | S _ => 1
                          end
                      | S i'' =>
                        fun pfi =>
                          let pfi' := lt_S_n _ _ pfi in
                          let pf0 := Olt _ _ pfi' in
                          let t := transpose_seg_fun pfj' pf0 
                                                     (S i'') pfi' in
                          match t with
                            | 0 => 0
                            | S _ => S (S i'')
                          end
                    end
                end
            | S k' =>
              fun pfj pfk i =>
                let pfj' := lt_S_n _ _ pfj in
                let pfk' := lt_S_n _ _ pfk in
                match i with
                  | 0 =>
                    fun _ =>  0
                  | S i' =>
                    fun pfi => 
                      let pfi' := lt_S_n _ _ pfi in
                      let pf0 := Olt _ _ pfi' in
                      let t := transpose_seg_fun pfj' pfk' i' pfi' in
                      S t
                end
          end
      end
  end.

  
  (*Seems to be less efficient than above in regards to simplification.*)
Definition transpose_seg_fun'
           n r s :
  seg_fun n nat :=
  fun i (pfi:i < n) =>
    if (beq_nat i r) then
      s else
      if (beq_nat i s) then
        r
      else
        i.

Lemma transpose_seg_fun_functional :
  forall {n r r' s s'}
         (pfr:r < n)
         (pfr':r' < n)
         (pfs:s < n)
         (pfs':s' < n),
    r = r' ->
    s = s' ->
    transpose_seg_fun pfr pfs = 
    transpose_seg_fun pfr' pfs'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

Lemma transpose_seg_fun_functional' :
  forall {n r r' s s' i i'}
         (pfr:r  < n)
            (pfr':r' < n)
            (pfs:s < n)
            (pfs':s' < n)
            (pfi: i < n)
            (pfi':i' < n),
       r = r' ->
       s = s' ->
       i = i' ->
       transpose_seg_fun pfr pfs i pfi =
       transpose_seg_fun pfr' pfs' i' pfi'.
    intros; subst; f_equal; apply proof_irrelevance.
  Qed.


  
Lemma transpose_seg_fun_comm :
  forall {n r s}
         (pfr:r < n)
         (pfs:s < n), 
    transpose_seg_fun pfr pfs = 
    transpose_seg_fun pfs pfr.
  intro n. 
  induction n as [|n ih].
  intros ? ? h1; contradict (lt_n_0 _ h1).
  intro r.
  destruct r as [|r].
  intro s.
  destruct s as [|s].
  intros hr hs.
  assert (hr = hs). apply proof_irrelevance. subst; auto. 
  intros hr hs. 
  simpl.
  apply functional_extensionality_dep.
  intro i.
  destruct i as [|i]; auto.
  destruct i as [|i]; auto.
  apply functional_extensionality.
  intro hi.
  rewrite ih; auto.
  intro s.
  destruct s as [|s].
  intros hr hs.
  simpl.
  apply functional_extensionality_dep.
  intro i.
  destruct i as [|i]; auto.
  destruct i as [|i]; auto.
  apply functional_extensionality.
  intro; rewrite ih; auto.
  intros hr hs.
  simpl.
  apply functional_extensionality_dep.
  intro i.
  destruct i; auto.
  apply functional_extensionality.
  intro hi.
  pose proof (lt_S_n _ _ hi) as hi'.
  specialize (ih r s (lt_S_n _ _ hr) (lt_S_n _ _ hs)).
  rewrite ih; auto.
Qed.



Lemma transpose_seg_fun_same :
  forall {n r} (pfr pfr':r < n),
    transpose_seg_fun pfr pfr' = id_seg n.
  intro n.
  induction n as [|n ih]; simpl.
  intros ? h1; contradict (lt_n_0 _ h1).
  intros r hr. 
  destruct r; auto.
  intro hr'.
  assert (hr = hr').
  apply proof_irrelevance.
  subst.
  apply functional_extensionality_dep.
  intro i.
  destruct i; auto.
  apply functional_extensionality.
  intro hi.
  unfold id_seg.
  f_equal.
  rewrite ih.
  unfold id_seg; auto.
Qed.


Lemma transpose_seg_fun_l :
  forall {n j k}
         (pfj pfj':j < n)
         (pfk:k < n),
    transpose_seg_fun pfj pfk j pfj' = k.
  intro n.
  induction n as [|n ih].
  intros ? ? h1.
  contradict (lt_n_0 _ h1).  
  intro j.
  destruct j as [|j].
  intro k.
  destruct k as [|k]; auto. 
  intro k; destruct k as [|k]; simpl.
  intros hj hk.
  simpl in ih.
  intro h0. 
  destruct j as [|j]; auto.
  rewrite ih; auto.
  intros hj hk.
  f_equal; auto.
Qed.


Lemma transpose_seg_fun_r :
  forall {n j k}
         (pfj:j < n)
         (pfk pfk':k < n),
    transpose_seg_fun pfj pfk k pfk' = j.
  intros; rewrite transpose_seg_fun_comm, transpose_seg_fun_l; auto.
Qed.

Lemma transpose_seg_fun_neq :
  forall {n j k}
         (pfj:j < n)
         (pfk pfk':k < n),
  forall i (pfi:i < n),
    i <> j ->
    i <> k ->
    transpose_seg_fun pfj pfk i pfi = i. 
  intro n.
  induction n as [|n ih]; simpl. 
  intros; omega. 
  intro j.
  destruct j as [|j].
  intro k.
  destruct k as [|k].
  intros; auto.   
  intros h0 hk hk' i hi.
  destruct i as [|i]. 
  intros; omega. 
  intros he he1.
  destruct i as [|i].
  destruct k as [|k].
  contradict he1; auto.
  reflexivity.
  rewrite ih; auto with arith.
  intro k.
  destruct k as [|k].
  intros hk hk' h1 i hi h2 h3.
  destruct i as [|i].
  contradict h3; auto.
  destruct i as [|i].
  destruct j as [|j]. 
  contradict h2; auto.
  reflexivity.
  rewrite ih; auto.
  omega.
  intros hj hk hk' i h1 h2 h3.
  destruct i as [|i]; auto.
  f_equal; auto.
  rewrite ih; auto.
  apply lt_S_n in hk'; auto.
Qed.


Lemma transpose_seg_fun_compat :
  forall {n j k}
         (pfj:j < n)
         (pfk:k < n),
    let f := transpose_seg_fun pfj pfk in
    f j pfj = k /\
    f k pfk = j /\
    forall i (pfi:i < n),
      i <> j ->
      i <> k ->
      f i pfi = i.
  intros n j k hj hk. simpl.
  split.
  rewrite transpose_seg_fun_l; auto.
  split.
  rewrite transpose_seg_fun_r; auto.
  apply transpose_seg_fun_neq; auto.
Qed.


Lemma tranpose_seg_fun_compat' :
  forall {n r s} (pfr:r < n) (pfs:s < n) i (pfi: i < n),
    transpose_seg_fun pfr pfs i pfi =
    transpose_seg_fun' n r s i pfi.
  intros n r s hr hs i hi.
  unfold transpose_seg_fun'.
  destruct (nat_eq_dec i r) as [h1 | h1]; subst.
  rewrite beq_nat_same, transpose_seg_fun_l; auto.
  pose proof h1 as h1'.
  rewrite <- neq_nat_compat in h1.
  rewrite h1.
  destruct (nat_eq_dec i s) as [h2 | h2]; subst.
  rewrite beq_nat_same, transpose_seg_fun_r; auto.
  pose proof h2 as h2'.
  rewrite <- neq_nat_compat in h2.
  rewrite h2.
  rewrite transpose_seg_fun_neq; auto.
Qed.


Lemma transpose_same_seg_fun_iff :
  forall {n r i m}
         (pfr:r < n)
         (pfi:i < n),
    transpose_seg_fun pfr pfr i pfi = m <->
    i = m.
  intro n.
  induction n as [|n ih].
  intros ? ? ? h1.
  contradict (lt_n_O _ h1).
  intros r i m.
  destruct r as [|r], i as [|i], m as [|m]; split; intro h1;
  auto with arith; simpl in h1; try discriminate;
  try apply S_inj in h1; f_equal; auto;
  try rewrite ih in h1; auto; subst.
  pose proof (lt_S_n _ _ pfr) as hr.
  pose proof (lt_S_n _ _ pfi) as hi.
  simpl; f_equal.
  rewrite ih; auto.
Qed.



Lemma lt_transpose_seg_fun : 
  forall {n r s i}
         (pfr:r < n)
         (pfs:s < n)
         (pf:i < n),
         transpose_seg_fun pfr pfs i pf < n.
  intros n r s i h1 h2.
  pose proof (transpose_seg_fun_compat h1 h2) as h4.
  simpl in h4.
  destruct h4 as [h4 [h5 h6]].
  destruct (eq_nat_dec i r) as [h7 | h7].
  rewrite h7.
  intro h1'.
  assert (h8:h1 = h1'). apply proof_irrelevance.
  rewrite <- h8, h4; auto.
  destruct (eq_nat_dec i s) as [h8 | h8].
  rewrite h8.
  intro h2'.
  assert (h9:h2 = h2'). apply proof_irrelevance.
  rewrite <- h9, h5; auto.
  intros; rewrite h6; auto.
Qed.

Lemma transpose_seg_fun_S :
  forall {n r s i}
         (pfr:r < n)
         (pfr':r < S n)
         (pfs:s < n)
         (pfs':s < S n)
         (pfi:i < n)
         (pfi':i < S n),
    transpose_seg_fun pfr' pfs' i pfi'  =
    transpose_seg_fun pfr pfs i pfi.  
  intro n; induction n as [|n ih]. 
  intros; omega.  
  intros r s i hr hs hi hi' hr' hs'.
  destruct r as [|r], s as [|s], i as [|i]; auto.
  pose proof hs' as h1.
  apply lt_S_n in h1.
  destruct (nat_eq_dec i s) as [|h2]; subst.
  rewrite transpose_seg_fun_r.
  rewrite transpose_seg_fun_r; auto.
  rewrite transpose_seg_fun_neq; auto.
  rewrite transpose_seg_fun_neq; auto.
  destruct (nat_eq_dec i r) as [|h2]; subst.
  rewrite transpose_seg_fun_l.
  rewrite transpose_seg_fun_l; auto.
  rewrite transpose_seg_fun_neq; auto.
  rewrite transpose_seg_fun_neq; auto.
  destruct (nat_eq_dec i r) as [|h2]; subst.
  rewrite transpose_seg_fun_l.
  rewrite transpose_seg_fun_l; auto.
  destruct (nat_eq_dec i s) as [|h2']; subst.
  rewrite transpose_seg_fun_r.
  rewrite transpose_seg_fun_r; auto.
  rewrite transpose_seg_fun_neq; auto.
  rewrite transpose_seg_fun_neq; auto.
Qed.


Lemma transpose_seg_fun_Sn :
  forall {n s}
         (pfn pfn':n < S n)
         (pfs:s < S n),
    transpose_seg_fun pfn pfs n pfn' = s.
  intros; rewrite transpose_seg_fun_l; auto.
Qed.

Lemma transpose_seg_fun_Sn' :
  forall {n s}
         (pfn pfn':n < S n)
         (pfr:s < S n),
    transpose_seg_fun pfr pfn' n pfn' = s.
  intros; rewrite transpose_seg_fun_r; auto.
Qed.


Lemma transpose_seg_fun0_eq_iff :
  forall {n s m a} (pf0:0 < n) (pfm:m < n) (pfs:s < n),
    transpose_seg_fun pf0 pfs m pfm = a <->
    (a = 0 /\ s = m) \/
    (a = s /\ m = 0) \/ 
    (a = m /\ m <> 0 /\ m <> s).
  intro n; induction n as [|n ih].
  intros ? ? ? h1; contradict (lt_irrefl _ h1).
  
  intros s m a; destruct s as [|s], m as [|m], a as [|a];
  intros h1 h2 h3; split; intro h4; subst; try tauto;
  try rewrite tranpose_seg_fun_l in h4;
  try rewrite transpose_seg_fun_r in h4;
  try discriminate; try destruct h4 as [[? ?] | [[? ?] | [? ?]]];
  try discriminate.
   
  right; right; auto with arith. 
  apply S_inj in H; subst.
  rewrite transpose_seg_fun_same; auto.
  destruct H0 as [h4]; contradict h4; auto.
  rewrite transpose_seg_fun_l in h4.
  apply S_inj in h4; subst.
  right; left; auto.
  apply S_inj in H; subst.
  rewrite transpose_seg_fun_l; auto.
  simpl in h4.
  destruct m as [|m], s as [|s].
  left; auto.
  discriminate.
  rewrite transpose_seg_fun_same in h4.
  simpl in h4; discriminate.  
  pose proof (lt_S_n _ _ h2) as h5.
  pose proof (lt_S_n _ _ h3) as h6.
  destruct (zerop (transpose_seg_fun (Olt _ _ (lt_S_n _ _ h2))
                               (lt_S_n _ _ h3)
                               (S m)
                               (lt_S_n _ _ h2))) as [h7 | h7].
  rewrite h7 in h4.
  rewrite ih in h7.
  destruct h7 as [h7 | [h7 | h7]].
  destruct h7 as [? h7].
  apply S_inj in h7; subst.
  left; auto.
  destruct h7; discriminate.
  destruct h7; discriminate.
  rewrite (S_pred _ _ h7) in h4.
  discriminate.
  apply S_inj in H0; subst.
  simpl.
  destruct m as [|m]; auto.
  rewrite transpose_seg_fun_r; auto.
  simpl in h4.
  destruct m as [|m], s as [|s].
  discriminate.
  apply S_inj in h4; subst. 
  right; right; auto with arith.
  destruct (zerop (transpose_seg_fun (Olt (S m) n (lt_S_n (S m) n h2)) 
                                     (lt_S_n 0 n h3) (S m) (lt_S_n (S m) n h2))) as [h5 | h5].
  rewrite h5 in h4.
  discriminate.
  rewrite (S_pred _ _ h5) in h4.
  apply S_inj in h4; subst.
  right; right; auto.
  destruct (zerop (transpose_seg_fun (Olt (S m) n (lt_S_n (S m) n h2))
                                     (lt_S_n (S s) n h3) (S m) (lt_S_n (S m) n h2))) as [h5 | h5].
  rewrite h5 in h4.
  discriminate.
  rewrite (S_pred _ _ h5) in h4.
  apply S_inj in h4; subst.
  destruct (nat_eq_dec s m) as [|h6]; subst.
  rewrite transpose_seg_fun_r in h5.
  contradict (lt_irrefl _ h5).
  right; right; auto.
  apply S_inj in H; subst.
  rewrite transpose_seg_fun_neq; auto.
  destruct H0; auto.
Qed.


(*Wyckoff*)
Lemma inj_transpose_seg_fun :
  forall {n r s}
         (pfr:r < n)
         (pfs:s < n),
    inj_seg (transpose_seg_fun pfr pfs).
  intros n r s hr hs; red; intros i j h1 h2 h3. 
  destruct (eq_nat_dec i r) as [h4 | h4], (eq_nat_dec j r) as [h5 | h5].
  subst; auto.
  destruct (eq_nat_dec j s) as [h6 | h6].
  subst; auto. 
  rewrite transpose_seg_fun_l, transpose_seg_fun_r in h3; auto.
  subst.
  rewrite transpose_seg_fun_l, transpose_seg_fun_neq in h3; auto.
  contradict (h6 (eq_sym h3)).
  subst.
  rewrite transpose_seg_fun_l in h3.  
  destruct (eq_nat_dec i s) as [h5 | h5].
  revert h3.
  generalize h1.
  rewrite h5.
  intros hs' h6. 
  rewrite transpose_seg_fun_r in h6; subst.
  contradict h4; auto. 
  rewrite transpose_seg_fun_neq in h3; try subst; auto.
  contradict h5; auto.
  destruct (nat_eq_dec j s) as [h6 | h6]; subst. 
  rewrite transpose_seg_fun_r in h3.
  destruct (eq_nat_dec i s) as [h6 | h6]; auto.
  rewrite transpose_seg_fun_neq in h3; try subst; auto.
  contradict h4; auto.
  destruct (eq_nat_dec i s) as [h9 | h9]; subst.
  rewrite transpose_seg_fun_r, transpose_seg_fun_neq in h3; try subst; auto.
  contradict h5; auto.
  rewrite transpose_seg_fun_neq, transpose_seg_fun_neq in h3; subst; auto.
Qed.


Lemma transpose_seg_fun_endo :
  forall {n r s}
         (pfr:r < n)
         (pfs:s < n),
    seg_fun_endo (transpose_seg_fun pfr pfs) n.
  intros; constructor; intros; apply lt_transpose_seg_fun.
Qed.


Definition perm_seg {n} (f:seg_fun n nat) : Prop :=
  inj_seg_fun_endo f n.


Lemma transpose_seg_fun_perm :
  forall {n r s}
         (pfr:r < n)
         (pfs:s < n),
    perm_seg (transpose_seg_fun pfr pfs).
  intros; constructor; intros.
  apply transpose_seg_fun_endo.
  eapply inj_transpose_seg_fun.
  apply H.
Qed.

Lemma perm_seg_to_S  :
  forall {n} (f:seg_fun (S n) nat)
         (pf0:0 < S n),
    f 0 pf0 = n ->
    perm_seg f ->
    perm_seg (seg_fun_to_S f).
  intros n f h1 h2 h3.
  red in h3; red.
  inversion h3 as [? ? h5 h6]; revert h1 h2; subst.
  inversion h5 as [? ? h7]; subst.
  intros h1 h2.
  constructor.
  constructor.
  intros i hi.
  specialize (h7 (S i) (lt_n_S _ _ hi)).
  apply lt_S_dec in h7. 
  destruct h7 as [|h7]; auto.
  rewrite <- h7 in h2.
  specialize (h6 0 (S i) h1 (lt_n_S _ _ hi) h2); discriminate.
  intros i j hi hj h8.
  unfold seg_fun_to_S in h8.
  apply h6 in h8; auto with arith.
Qed.
       


Lemma perm_seg_to_S' :
  forall {n} (f:seg_fun (S n) nat)
         (pf0:0 < S n),
    f 0 pf0 = n ->
    perm_seg (seg_fun_to_S f) ->
    perm_seg f.
  intros n f h0 h1 h2.
  red in h2; red.
  inversion h2 as [? ? h3 h4 h5 h6].
  eapply inj_seg_fun_endo_to_S' in h2.
  apply h2.
  apply le_n_Sn.
  erewrite f_equal_dep.
  rewrite h1 at 1.
  apply lt_n_Sn.
  reflexivity.
  revert h1; revert h0; inversion h3 as [? ? h7 h8]; subst; intros h0 h1.
  intro h5.
  rewrite in_seg_fun_iff in h5.
  destruct h5 as [i [h5 h6]].
  specialize (h7 i h5).
  rewrite h6 in h7.
  erewrite f_equal_dep in h7.
  rewrite h1 in h7 at 1.
  contradict (lt_irrefl _ h7).
  reflexivity.
Qed.

Lemma perm_seg_fun1_eq :
  forall (f:seg_fun 1 nat) i,
    perm_seg f ->
    forall (pf:i < 1),
      f i pf = 0.
  intros ? ? h1 ?; inversion h1; apply seg_fun_endo1_eq; auto.
Qed.


Lemma perm_seg_left_shift : 
  forall n,
    perm_seg (left_shift_seg_fun n).
  intros; apply inj_seg_fun_endo_shift.
Qed.

Lemma perm_seg_right_shift : 
  forall n,
    perm_seg (right_shift_seg_fun n).
  intros; apply inj_seg_fun_endo_shift.
Qed.


(*Wyckoff*)
(*This transposes values of a given [seg_fun] at [j] and [k] -- 
and leaves all else fixed.*)
Definition transpose_functor
           {U n j k}
           (f:seg_fun n U)
           (pfj:j < n) (pfk:k < n)
: seg_fun n U :=
  fun i pfi =>
    if (beq_nat i j) then
      f k pfk
    else
      if (beq_nat i k) then
        f j pfj else
        f i pfi.


Lemma transpose_functor_functional :
  forall {U} {n j j' k k'}
         (f:seg_fun n U)
         (pfj : j < n)
         (pfj': j' < n)
         (pfk : k < n)
         (pfk': k' < n),
    j = j' ->
    k = k' ->
    transpose_functor f pfj pfk = 
    transpose_functor f pfj' pfk'.
  intros; subst; f_equal; try apply proof_irrelevance.
Qed.

  
Lemma transpose_functor_functional' :
  forall {U} {n j j' k k' i i'}
         (f:seg_fun n U)
         (pfj : j < n)
         (pfj': j' < n)
         (pfk : k < n)
         (pfk': k' < n)
         (pfi : i < n)
         (pfi': i' < n),
    j = j' ->
    k = k' ->
    i = i' ->
    transpose_functor f pfj pfk i pfi = 
    transpose_functor f pfj' pfk' i' pfi'.
  intros; subst; f_equal; try apply proof_irrelevance.
Qed.


(*Wyckoff*)
Lemma transpose_functor_same :
  forall {U:Type} {n:nat}
         (f:seg_fun n U) (r:nat)
         (pfr pfr':r < n),
    transpose_functor f pfr pfr' = f. 
  unfold transpose_functor; intros; 
  apply functional_extensionality_dep; intro x;
  apply functional_extensionality_dep; intro h1.
  destruct (eq_nat_dec x r) as [h2 | h2]; subst.
  rewrite beq_nat_same.
  f_equal; apply proof_irrelevance.
  rewrite <- neq_nat_compat in h2.
  rewrite h2; auto.
Qed.


(*Wyckoff*)
Lemma inj_transpose_functor :
  forall {U:Type} {n:nat}
         (f:seg_fun n U) (r s:nat)
         (pfr:r < n) (pfs:s < n),
    inj_seg f ->
    inj_seg (transpose_functor f pfr pfs).
  intros U n f r s h1 h2 h3.
  red in h3; red.
  intros i j h4 h5 h6.
  unfold transpose_functor in h6. 
  destruct (eq_nat_dec i r) as [h7|h7]; subst.
  rewrite beq_nat_same in h6. 
  destruct (eq_nat_dec j r) as [h8|h8]; subst; auto. 
  rewrite <- neq_nat_compat in h8.
  rewrite h8 in h6.
  destruct (eq_nat_dec j s) as [h9|h9]; subst; auto.
  rewrite beq_nat_same in h6.
  apply h3 in h6; auto.
  rewrite <- neq_nat_compat in h9.
  rewrite h9 in h6.
  apply h3 in h6; subst.
  rewrite beq_nat_same in h9; discriminate.
  rewrite <- neq_nat_compat in h7.
  rewrite h7 in h6.
  destruct (eq_nat_dec i s) as [h9|h9]; subst; auto.
  rewrite beq_nat_same in h6.
  destruct (eq_nat_dec j r) as [h10|h10]; subst; auto.
  rewrite beq_nat_same in h6. 
  apply h3 in h6; auto.
  rewrite <- neq_nat_compat in h10.
  rewrite h10 in h6.
  destruct (eq_nat_dec j s) as [h11|h11]; subst; auto.
  rewrite <- neq_nat_compat in h11.
  rewrite h11 in h6.
  apply h3 in h6; subst.
  rewrite beq_nat_same in h10; discriminate.
  rewrite <- neq_nat_compat in h9.
  rewrite h9 in h6.
  destruct (eq_nat_dec j r) as [h11|h11]; subst; auto.
  rewrite beq_nat_same in h6.
  apply h3 in h6; subst.
  rewrite beq_nat_same in h9; discriminate.
  rewrite <- neq_nat_compat in h11.
  rewrite h11 in h6.
  destruct (eq_nat_dec j s) as [h12|h12]; subst; auto.
  rewrite beq_nat_same in h6.
  apply h3 in h6; subst.
  rewrite beq_nat_same in h7; discriminate.
  rewrite <- neq_nat_compat in h12.
  rewrite h12 in h6.
  apply h3 in h6; auto.
Qed.


(*Wyckoff*)
Lemma inj_transpose_functor' :
  forall {U:Type} {n:nat}
         (f:seg_fun n U) (r s:nat)
         (pfr:r < n) (pfs:s < n),
    inj_seg (transpose_functor f pfr pfs) ->
    inj_seg f.
  intros U n f r s h1 h2 h3; red in h3; red; intros x y h4 h5 h6.
  destruct (eq_nat_dec x r) as [|h7],
                               (eq_nat_dec y s) as [|h8]; subst.
  eapply h3.
  unfold transpose_functor.
  do 2 rewrite beq_nat_same.
  destruct (eq_nat_dec s r) as [h7 | h7]; subst.
  rewrite beq_nat_same; auto.
  rewrite <- neq_nat_compat in h7.
  rewrite h7.
  erewrite f_equal_dep.
  rewrite <- h6 at 1.
  f_equal.
  apply proof_irrelevance.
  reflexivity.
  destruct (eq_nat_dec r y) as [|h9]; subst; auto.
  apply neq_sym in h9.
  rewrite <- neq_nat_compat in h9.
  specialize (h3 s y h2 h5).
  unfold transpose_functor in h3.
  rewrite h9, beq_nat_same in h3.
  rewrite <- neq_nat_compat in h8.
  rewrite h8 in h3.
  destruct (nat_eq_dec s r) as [|hsr]; subst.
  rewrite beq_nat_same in h3.
  apply h3.
  erewrite f_equal_dep.
  rewrite h6 at 1; auto.
  reflexivity.
  rewrite <- neq_nat_compat in hsr.
  rewrite hsr in h3.
  erewrite f_equal_dep in h6.
  apply h3 in h6; subst.
  rewrite beq_nat_same in h8; discriminate.
  reflexivity.
  rewrite <- neq_nat_compat in h7.  
  specialize (h3 x r h4 h1).
  unfold transpose_functor in h3.
  rewrite h7, beq_nat_same in h3.
  destruct (eq_nat_dec x s) as [|hxs]; subst; auto.
  rewrite <- neq_nat_compat in hxs.
  rewrite hxs in h3.
  hfirst h3.
  rewrite h6.
  f_equal; apply proof_irrelevance.
  apply h3 in hf; subst.
  rewrite beq_nat_same in h7; discriminate. 
  rewrite <- neq_nat_compat in h7, h8.
  destruct (eq_nat_dec y r) as [|hyr], (eq_nat_dec x s) as [|hxs]; subst; auto.
  specialize (h3 r s h1 h4).
  unfold transpose_functor in h3.
  rewrite beq_nat_same, h7, beq_nat_same in h3.
  rewrite h3; auto. 
  erewrite f_equal_dep. 
  rewrite h6 at 1.
  f_equal; apply proof_irrelevance.
  reflexivity.
  rewrite <- neq_nat_compat in hxs. 
  specialize (h3 x s h4 h2).
  unfold transpose_functor in h3.
  rewrite h7, hxs, beq_nat_same in h3.
  rewrite beq_nat_sym in h8.
  rewrite h8 in h3.
  rewrite h6 in h3.
  hfirst h3.
  f_equal; apply proof_irrelevance.
  specialize (h3 hf); subst.
  rewrite beq_nat_same in hxs; discriminate.
  rewrite <- neq_nat_compat in hyr. 
  specialize (h3 r y h1 h5).
  unfold transpose_functor in h3.
  rewrite beq_nat_sym in h7.
  hfirst h3.
  rewrite beq_nat_same, hyr, h8.
  rewrite <- h6.
  f_equal; apply proof_irrelevance.
  specialize (h3 hf); subst.
  rewrite beq_nat_same in hyr; discriminate.
  rewrite <- neq_nat_compat in hyr.
  rewrite <- neq_nat_compat in hxs.
  specialize (h3 x y h4 h5).
  unfold transpose_functor in h3.
  hfirst h3.
  rewrite h7, h8, hyr, hxs; auto.
  auto.

  Grab Existential Variables.
  assumption.
  assumption.
Qed.



Definition transposed_seg_list {n r s} (pfr:r < n) (pfs:s < n) :
  list nat :=
  map_seg (transpose_seg_fun pfr pfs).


Lemma in_transposed_seg_list_iff :
  forall {n r s i}
         (pfr:r < n)
         (pfs:s < n),
    (In i (transposed_seg_list pfr pfs) <->
     i < n).
  unfold transposed_seg_list; intros n r s i hr hs; split; intro h1.
  rewrite in_map_seg_iff in h1.
  destruct h1 as [m [h1 ]]; subst.
  apply lt_transpose_seg_fun.
  rewrite in_map_seg_iff.
  destruct (nat_eq_dec i r) as [|hir], (nat_eq_dec i s) as [|his]; subst; subst.
  exists _, hs.
  rewrite transpose_seg_fun_r; auto.
  exists _, hs.
  rewrite transpose_seg_fun_r; auto.
  exists _, hr.
  rewrite transpose_seg_fun_l; auto.
  exists _, h1.
  rewrite transpose_seg_fun_neq; auto.
Qed.  


Lemma length_transposed_seg_list :
  forall {n r s} (pfr:r < n) (pfs:s < n),
    length (transposed_seg_list pfr pfs) = n.
  unfold transposed_seg_list; intros; rewrite length_map_seg; auto.
Qed.


(*Shortened proof than less efficient version using old 
[transpose_seg_fun].  Proof of efficiency in applications on constants!*)
Lemma transposed_seg_list_01 :
  transposed_seg_list lt_0_2 lt_1_2 = 1::0::nil.
  unfold transposed_seg_list, transpose_seg_fun.   
  rewrite map_seg_eq_iff.
  simpl.
  exists eq_refl.
  intros m h1.
  destruct m as [|m]; auto.
  destruct m as [|m]; auto.
  destruct m as [|m]; omega.
Qed.


Lemma transposed_seg_list_0n :
  forall n (pf:0 < n),
    transposed_seg_list (lt_0_Sn n) (lt_n_Sn n) = n::seq 1 (pred n) ++ 0::nil.
  intros n hn. 
  unfold transposed_seg_list.  
  rewrite map_seg_eq_iff.
  ex_goal.
  simpl; rewrite app_length, seq_length; simpl.
  omega. 
  exists hex.
  intros m hm h1.
  unfold h1.
  clear h1.
  destruct (zerop m) as [|h2]; subst.
  rewrite transpose_seg_fun_l.
  symmetry.
  gen0.
  rewrite cons_app.
  intro h1.
  erewrite nth_lt_app1.
  simpl; auto.
  pose proof hm as h3.
  apply le_lt_eq_dec in h3.
  destruct h3 as [h3 | h3].
  apply lt_S_n in h3.
  symmetry.
  gen0.
  rewrite cons_app.
  intro.
  erewrite nth_lt_app1.
  rewrite transpose_seg_fun_neq; auto with arith.
  simpl.
  destruct m as [|m].
  contradict (lt_n_0 _ h2).
  erewrite nth_indep.
  rewrite <- nth_lt_compat. 
  rewrite nth_lt_seq; auto with arith.
  rewrite seq_length; auto with arith. 
  intro; subst; contradict (lt_irrefl _ h2).
  intro; subst; contradict (lt_irrefl _ h3).
  apply S_inj in h3; subst.
  rewrite transpose_seg_fun_r.
  symmetry.
  gen0.
  rewrite cons_app.
  intro. 
  erewrite nth_lt_app2; simpl.
  rewrite seq_length.
  rewrite <- (S_pred _ _ h2), minus_same; auto.

  Grab Existential Variables.
  simpl; rewrite seq_length, <- (S_pred _ _ h2); auto with arith.
  rewrite seq_length.
  omega.
  simpl; rewrite seq_length, <- (S_pred _ _ h3); auto.
  simpl; rewrite seq_length, <- (S_pred _ _ hn); auto.
Qed.


Lemma transposed_seg_list_app_sing :
  forall (n r s:nat)
         (pfr:r < n) (pfs:s < n),
    let pfr' := lt_trans _ _ _ pfr (lt_n_Sn  _) in
    let pfs' := lt_trans _ _ _ pfs (lt_n_Sn  _) in
    transposed_seg_list pfr' pfs' =
    transposed_seg_list pfr pfs ++ n :: nil.
  unfold transposed_seg_list; intros n r s h1 h2.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite app_length, length_map_seg.
  simpl.
  rewrite S_compat; auto.
  exists hex.
  intros m h3.
  intro h4.
  unfold h4.
  clear h4.
  pose proof h3 as ha.
  apply le_lt_eq_dec in ha.
  destruct ha as [ha | ha].
  apply lt_S_n in ha.
  erewrite nth_lt_app1.
  rewrite nth_lt_map_seg. 
  simpl. 
  destruct r as [|r], s as [|s].
  rewrite transpose_seg_fun_same; auto.
  destruct m as [|m].
  rewrite transpose_seg_fun_l; auto.
  destruct m as [|m], s as [|s].
  rewrite transpose_seg_fun_r; auto.
  rewrite transpose_seg_fun_neq; auto.
  rewrite transpose_seg_fun_same; simpl.
  rewrite transpose_seg_fun_neq; auto.
  destruct (nat_eq_dec m s) as [|h4]; subst.
  do 2 rewrite transpose_seg_fun_r; auto.
  rewrite transpose_seg_fun_neq, transpose_seg_fun_neq; auto with arith.
  destruct m as [|m].
  rewrite transpose_seg_fun_r; auto.
  destruct m as [|m], r as [|r].
  rewrite transpose_seg_fun_l; auto.
  rewrite transpose_seg_fun_neq; auto.
  rewrite transpose_seg_fun_same; simpl.
  rewrite transpose_seg_fun_neq; auto.
  destruct (nat_eq_dec m r) as [|h4]; subst.
  rewrite transpose_seg_fun_l, transpose_seg_fun_l; auto.
  rewrite transpose_seg_fun_neq, transpose_seg_fun_neq; auto.
  destruct m as [|m].
  rewrite transpose_seg_fun_neq; auto with arith.
  destruct (nat_eq_dec m r) as [h4 | h4], (nat_eq_dec m s) as [h5 | h5]; subst; subst. 
  do 2 rewrite transpose_seg_fun_same; auto.
  do 2 rewrite transpose_seg_fun_l; auto.
  do 2 rewrite transpose_seg_fun_r; auto.
  rewrite transpose_seg_fun_neq, transpose_seg_fun_neq; auto.
  omega.
  apply S_inj in ha; subst.
  erewrite transpose_seg_fun_neq, nth_lt_app2; simpl.
  rewrite length_map_seg, minus_same; auto with arith.
  auto.
  intro; subst; contradict (lt_irrefl _ h1).
  intro; subst; contradict (lt_irrefl _ h2).

  Grab Existential Variables.
  rewrite length_map_seg; auto with arith.
  auto.
Qed.


Definition transposed_list_nat (l:list nat) (r s:nat)
           (pfr:r < length l) (pfs:s < length l) : list nat :=
  map_seg (transpose_functor (nth_lt l) pfr pfs).
 

Lemma transposed_list_nat_eq_nil_iff :
  forall (l:list nat) (r s:nat)
         (pfr:r < length l) (pfs:s < length l),
    transposed_list_nat l r s pfr pfs = nil <->
    l = nil.
  unfold transposed_list_nat; intros l r s h1 h2; split; intro h3.
  apply map_seg_eq_nil in h3.
  apply length_eq0 in h3; auto.
  subst; simpl; auto.
Qed.



Lemma transposed_list_nat_same :
  forall (l:list nat) (r:nat)
         (pfr pfr':r < length l),
    transposed_list_nat l r r pfr pfr' = l.
  intros l r h1 h2.
  unfold transposed_list_nat.
  rewrite transpose_functor_same.
  apply map_seg_nth_lt.
Qed.
         

Lemma length_transposed_list_nat :
  forall (l:list nat) (r s:nat)
         (pfr:r < length l) (pfs:s < length l),
    length (transposed_list_nat l r s pfr pfs) =
    length l.
  intros l r s h1 h2; unfold transposed_list_nat.
  rewrite length_map_seg; auto.
Qed.

Lemma transposed_list_nat_functional :
  forall (l l':list nat) (pfe:l = l') (r s:nat)
         (pfr:r < length l) (pfs:s < length l),
    transposed_list_nat l r s pfr pfs =
    transposed_list_nat l' r s
                    (lt_length_subst _ _ r pfe pfr)
                    (lt_length_subst _ _ s pfe pfs).
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

Lemma transposed_list_nat_functional' :
  forall (l:list nat)  (r s r' s':nat)
         (pfr:r < length l) (pfs:s < length l)
         (pfer:r = r')
         (pfes:s = s'),
    transposed_list_nat l r s pfr pfs =
    transposed_list_nat l r' s'
                    (eq_lt_trans _ _ _ (eq_sym pfer) pfr)
                    (eq_lt_trans _ _ _ (eq_sym pfes) pfs).
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

Lemma transposed_list_nat_undoes_itself :
  forall l r s
         (pfr: r < length l)
         (pfs: s < length l)
         (pfr':r < length (transposed_list_nat l r s pfr pfs))
         (pfs':s < length (transposed_list_nat l r s pfr pfs)),
    transposed_list_nat
      (transposed_list_nat l r s pfr pfs) r s pfr' pfs' =
    l.
  intros l r s hr hs hr' hs'.
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff. 
  ex_goal.
  rewrite length_map_seg; auto.  
  exists hex.
  intros m hlt hlt'.
  unfold transpose_functor.
  unfold hlt'.
  clear hlt'.
  destruct (eq_nat_dec m r); subst.
  rewrite beq_nat_same.
  rewrite nth_lt_map_seg'.
  rewrite beq_nat_same.
  destruct (eq_nat_dec s r); subst. 
  rewrite beq_nat_same.
  f_equal; apply proof_irrelevance.
  rewrite <- neq_nat_compat in n.
  rewrite n.
  f_equal; apply proof_irrelevance.
  rewrite <- neq_nat_compat in n.
  rewrite n.
  destruct (eq_nat_dec m s); subst.
  rewrite beq_nat_same.
  rewrite nth_lt_map_seg'.
  rewrite beq_nat_same.
  apply nth_lt_functional3.
  rewrite <- neq_nat_compat in n0.
  rewrite n0.
  rewrite nth_lt_map_seg'.
  rewrite n, n0.
  f_equal; apply proof_irrelevance.
Qed.


Lemma transposed_list_nat_eq_iff :
  forall (l l':list nat) (r s:nat)
         (pfr:r < length l) (pfs:s < length l),
    transposed_list_nat l r s pfr pfs = l' <->
    exists (pf:length l = length l'),
      let pfr' := lt_eq_trans _ _ _ pfr pf in
      let pfs' := lt_eq_trans _ _ _ pfs pf in
      l = transposed_list_nat l' r s pfr' pfs'.
  intros l l' r s h1 h2.
  split; intro h3.
  subst.
  ex_goal.
  rewrite length_transposed_list_nat; auto.
  exists hex.
  intros h3 h4.
  rewrite transposed_list_nat_undoes_itself; auto.
  destruct h3 as [h3 h4].
  simpl in h4.
  generalize h1 h2.
  rewrite h4.
  intros h5 h6.
  rewrite transposed_list_nat_undoes_itself; auto.
Qed.


Lemma transposed_list_nat_comm :
  forall (l:list nat) (r s:nat)
         (pfr:r < length l)
         (pfs:s < length l),
    transposed_list_nat l r s pfr pfs =
    transposed_list_nat l s r pfs pfr.
  intros l r s hr hs.
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite length_map_seg; auto.
  exists hex.
  intros m hlt h1.
  rewrite nth_lt_map_seg'.
  unfold transpose_functor.
  unfold h1. clear h1.
  destruct (eq_nat_dec m r) as [h2 | h2], (eq_nat_dec m s) as [h3 | h3];
    subst; try repeat rewrite beq_nat_same; subst;
    try repeat rewrite beq_nat_same; try (f_equal; apply proof_irrelevance).
  rewrite <- neq_nat_compat in h3.
  rewrite h3; auto.
  rewrite <- neq_nat_compat in h2.
  rewrite h2; auto.
  rewrite <- neq_nat_compat in h2, h3.
  rewrite h2, h3; auto.
  f_equal; apply proof_irrelevance.
Qed.

    
Lemma transposed_list_nat1 :
  forall (l:list nat)  (r s:nat)
         (pfr:r < length l) (pfs:s < length l),
    length l = 1 ->
    transposed_list_nat l r s pfr pfs = l.
  intros l r s hr hs h1.
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff.
  exists (eq_refl _).
  intros m h2. simpl.
  unfold transpose_functor.
  destruct (eq_nat_dec m r); subst.
  rewrite beq_nat_same.
  pose proof hr as h3.
  pose proof hs as h4.
  rewrite h1 in h3, h4.
  apply lt1 in h3.
  apply lt1 in h4; subst.
  apply nth_lt_functional3.
  pose proof hr as h3.
  pose proof hs as h4.
  pose proof h2 as h2'.
  rewrite h1 in h2', h3, h4.
  apply lt1 in h2'; apply lt1 in h3; apply lt1 in h4; subst.
  contradict n; auto.
Qed.  

Lemma transposed_seg_list_compat :
  forall (n r s:nat)
         (pfr:r < n)
         (pfs:s < n),
    transposed_seg_list pfr pfs =
    transposed_list_nat (seg_list n) r s
                    (lt_eq_trans _ _ _ pfr
                       (eq_sym (length_seg_list _)))
                    (lt_eq_trans _ _ _ pfs
                                 (eq_sym (length_seg_list _))).
  intros n r s h1 h2.
  unfold transposed_seg_list, transposed_list_nat.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite length_map_seg, length_seg_list; auto.
  exists hex.
  intros i h3.
  simpl.
  rewrite nth_lt_map_seg'.
  symmetry.
  gen0.
  intro h4.
  clear hex.
  unfold transpose_functor.
  destruct (eq_nat_dec i r) as [|h5], (eq_nat_dec i s) as [|h6]; subst;
  subst; try repeat rewrite beq_nat_same;
  try rewrite nth_lt_seg_list;
  try rewrite transpose_seg_fun_r;
  try rewrite transpose_seg_fun_l; auto.
  rewrite <- neq_nat_compat in h5.
  rewrite h5, nth_lt_seg_list; auto.
  pose proof h5 as h5'. pose proof h6 as h6'.
  rewrite <- neq_nat_compat in h5, h6.
  rewrite h5, h6, nth_lt_seg_list.
  rewrite transpose_seg_fun_neq; auto.
Qed.


Lemma in_transposed_list_nat_iff :
  forall (l:list nat) (r s:nat)
         (pfr:r < length l) (pfs:s < length l)
         (i:nat),
    In i (transposed_list_nat l r s pfr pfs) <->
    In i l. 
  unfold transposed_list_nat, transpose_functor; intros l r s h1 h2 i; 
  split; intro h3.
  rewrite in_map_seg_iff in h3.
  destruct h3 as [j [h3 h4]]; subst. 
  destruct (eq_nat_dec j r) as [h4 | h4]; subst.
  rewrite beq_nat_same.
  apply in_nth_lt.
  destruct (eq_nat_dec j s) as [h5 | h5]; subst.
  rewrite <- neq_nat_compat in h4.
  rewrite beq_nat_same, h4.
  apply in_nth_lt.
  rewrite <- neq_nat_compat in h4, h5.
  rewrite h4, h5.
  apply in_nth_lt. 
  rewrite in_map_seg_iff.
  destruct (eq_nat_dec i (nth_lt l s h2)) as [|h4]; subst.
  exists r, h1.
  rewrite beq_nat_same; auto.
  destruct (eq_nat_dec i (nth_lt l r h1)) as [|h5]; subst.
  exists s, h2.
  destruct (eq_nat_dec s r) as [|h5]; subst; auto.
  contradict h4; auto.
  f_equal; apply proof_irrelevance.
  rewrite <- neq_nat_compat in h5.
  rewrite beq_nat_same, h5; auto.
  exists (lind nat_eq_dec _ _ h3).
  exists (lt_lind nat_eq_dec _ _ _).
  destruct (eq_nat_dec (lind nat_eq_dec _ _ h3) r)
    as [|h6],
       (eq_nat_dec
          (lind nat_eq_dec _ _ h3) s)
      as [|h7]; subst; try rewrite beq_nat_same.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
  contradict h5.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
  contradict h4.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
  rewrite <- neq_nat_compat in h6, h7.
  rewrite h6, h7.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
Qed.

Lemma transposed_list_nat_app_sing :
  forall (l:list nat) (r s:nat)
         (pfr:r < length l) (pfs:s < length l),
    transposed_list_nat (l++((length l)::nil)) r s
                   (lt_app _ _ _ pfr)
                   (lt_app _ _ _ pfs) = 
    transposed_list_nat l r s pfr pfs ++ (length l) :: nil.
  intros l r s h1 h2. 
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff.
  ex_goal.
  do 2 rewrite app_length; simpl.
  f_equal.
  rewrite length_map_seg; auto.
  exists hex.   
  unfold transpose_functor; intros m h3.
  symmetry. 
  pose proof h1 as h1'.
  pose proof h2 as h2'.
  apply lt_neq in h1'.
  apply lt_neq in h2'. 
  rewrite <- neq_nat_compat in h1', h2'.
  rewrite beq_nat_sym in h1', h2'.
  destruct (nat_eq_dec m (length l))
    as [h5|h5], 
       (nat_eq_dec r s) as [ha|ha]; subst.
  rewrite h1'.
  erewrite nth_lt_app2.
  simpl.
  erewrite nth_lt_app2; simpl.
  rewrite minus_same, length_map_seg, minus_same; auto.
  erewrite nth_lt_app2.
  simpl.
  rewrite length_map_seg, minus_same; auto.
  rewrite h1', h2'. 
  erewrite nth_lt_app2; simpl.
  rewrite  minus_same; auto.
 
  destruct (eq_nat_dec m s) as [|hms]; subst.
  rewrite beq_nat_same.
  erewrite nth_lt_app1. 
  rewrite nth_lt_map_seg'.
  rewrite beq_nat_same.
  erewrite nth_lt_app1.
  reflexivity.
  rewrite <- neq_nat_compat in hms.
  rewrite hms.
  erewrite nth_lt_app1.
  rewrite nth_lt_map_seg', hms.
  erewrite nth_lt_app1.
  reflexivity. 
  destruct (nat_eq_dec m r)
    as [|hmr],
       (nat_eq_dec m s)
      as [|hms]; subst; try rewrite beq_nat_same; subst.
  erewrite nth_lt_app1.
  rewrite nth_lt_map_seg'.
  rewrite beq_nat_same.
  erewrite nth_lt_app1.
  reflexivity.
  erewrite nth_lt_app1.
  rewrite nth_lt_map_seg'.
  rewrite beq_nat_same.
  erewrite nth_lt_app1.
  reflexivity.
  rewrite <- neq_nat_compat in hmr.
  rewrite hmr.
  erewrite nth_lt_app1.
  rewrite nth_lt_map_seg'.
  rewrite hmr, beq_nat_same.
  erewrite nth_lt_app1.
  reflexivity.
  rewrite <- neq_nat_compat in hmr, hms.
  rewrite hmr, hms.
  erewrite nth_lt_app1.
  rewrite nth_lt_map_seg'.
  rewrite hmr, hms.
  erewrite nth_lt_app1.
  reflexivity.
  
  Grab Existential Variables.
  rewrite length_map_seg; auto. 
  pose proof h3 as h6.
  rewrite app_length in h6.
  simpl in h6.
  omega.
  rewrite length_map_seg.
  omega.
  rewrite length_map_seg.
  omega.
  rewrite length_map_seg.
  omega.
  rewrite length_map_seg; rewrite app_length in h3; simpl in h3.
  omega.
  rewrite length_map_seg.
  omega.
  auto with arith.
  rewrite length_map_seg; auto.
  auto.
  rewrite length_map_seg; auto.
Qed.


Lemma transposed_list_nat_app_sing' :
  forall (l:list nat) (r s:nat)
         (pfr:r < length l) (pfs:s < length l)
         (pfr':r < length (l++((length l)::nil)))
         (pfs':s < length (l++((length l)::nil))),
         transposed_list_nat (l++((length l)::nil)) r s
                         pfr' pfs' = 
    transposed_list_nat l r s pfr pfs ++ (length l) :: nil.
  intros l r s h1 h2 h3 h4.
  assert (h3 = lt_app _ _ _ h1). apply proof_irrelevance.
  assert (h4 = lt_app _ _ _ h2). apply proof_irrelevance.
  subst.
  apply transposed_list_nat_app_sing.
Qed.


Lemma transposed_list_nat_01 :
  forall l a b
         (pfl0:0 < length (a::b::l))
         (pfl1:1 < length (a::b::l)),
    transposed_list_nat (a::b::l) 0 1 pfl0 pfl1 =
    b::a::l.
  intros l a b h1 h2.
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff.
  simpl.
  exists eq_refl.
  intros m h3.
  unfold transpose_functor.
  destruct m as [|m]; subst.
  rewrite beq_nat_same; simpl; auto.
  simpl.
  destruct m as [|m].
  rewrite beq_nat_same; auto.
  pose proof (lt_0_Sn m) as h4.
  apply lt_neq in h4.
  rewrite <- neq_nat_compat, beq_nat_sym in h4.
  rewrite h4.
  apply nth_indep.
  omega.
Qed.


Lemma transposed_list_nat_01' :
  forall l 
         (pfl0:0 < length l)
         (pfl1:1 < length l),
    transposed_list_nat l 0 1 pfl0 pfl1 =
    hd2 l pfl1 :: hd' l (lt_length _ _ pfl0) :: tl2 l.
  intros l h1.
  pose proof (lt_length _ _ h1) as h2.
  pose proof (hd_compat' _ h2) as h3.
  simpl in h3.
  revert h1.
  rewrite h3.
  intros h1 h4; simpl.
  simpl in h1, h4.
  pose proof h4 as h4'.
  apply lt_S_n in h4'.
  pose proof (lt_length _ _ h4') as h5.
  pose proof (hd_compat' _ h5) as h6.
  simpl in h6.
  generalize h1, h4.
  rewrite h6.
  intros; rewrite transposed_list_nat_01.
  unfold hd2, tl2.
  simpl; auto.
Qed.


Lemma transposed_list_nat_S :
  forall (l:list nat)
         (r s:nat)
         (pfr:r < length l)
         (pfs:s < length l)
         (pfsr:S r < S (length (transposed_list_nat l r s pfr pfs)))
         (pfss:S s < S (length (transposed_list_nat l r s pfr pfs))),
    transposed_list_nat (length l :: transposed_list_nat l r s pfr pfs)
                        (S r) (S s) pfsr pfss =
    length l :: l.
  intros l r s hr hs hsr hss.
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff.
  ex_goal.
  simpl.
  f_equal.
  rewrite length_map_seg; auto.
  exists hex.
  intros m h1 h2. 
  unfold transpose_functor, h2.
  clear h2.
  destruct (eq_nat_dec m (S r)) as [h2 | h2],
                                   (eq_nat_dec m (S s)) as [h3 | h3]; subst. 
  apply S_inj in h3; subst.
  rewrite beq_nat_same, nth_lt_cons, nth_lt_map_seg', beq_nat_same.
  rewrite nth_lt_cons.
  apply nth_lt_functional3.
  rewrite beq_nat_same, nth_lt_cons, nth_lt_map_seg', beq_nat_same.
  apply S_inj_neq in h3.
  apply neq_sym in h3.
  rewrite <- neq_nat_compat in h3.
  rewrite h3.
  simpl.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption. 
  rewrite <- neq_nat_compat in h2.
  rewrite h2, beq_nat_same.
  rewrite nth_lt_cons, nth_lt_map_seg', beq_nat_same; simpl.
  rewrite nth_lt_compat.
  apply nth_indep.
  assumption.
  pose proof h2 as h2'.
  pose proof h3 as h3'.
  rewrite <- neq_nat_compat in h2, h3.
  rewrite h2, h3.
  simpl.
  simpl in h1; rewrite length_map_seg in h1.
  destruct m as [|m]; auto.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  rewrite nth_lt_map_seg.
  apply S_inj_neq in h2'.
  apply S_inj_neq in h3'.
  rewrite <- neq_nat_compat in h2', h3'.
  rewrite h2', h3'.
  rewrite nth_lt_compat.
  apply nth_indep.
  auto with arith.
  rewrite length_map_seg; auto with arith.

  Grab Existential Variables.
  auto with arith.
Qed.


Lemma transposed_list_nat_cons_0_S_lind :
  forall (l:list nat) (a:nat) (pfi:In (length l) l)
         (pf0 : 0 < length (a::l))
         (pfs : S (lind nat_eq_dec l (length l) pfi) < length (a::l)),
    let pfs' := lt_S_n _ _ pfs in
    transposed_list_nat (a :: l) 0 (S (lind nat_eq_dec l (length l) pfi))
                        pf0 pfs =
    (length l) :: (list_subst l (lind nat_eq_dec l (length l) pfi) pfs'
                                a).
  intros l a h1 h2 h3.
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff.
  ex_goal.
  simpl.
  f_equal.
  rewrite length_list_subst; auto.
  exists hex.
  intros m h4; simpl.
  unfold transpose_functor.
  destruct (eq_nat_dec m 0) as [|hnm]; subst.
  rewrite beq_nat_same.
  rewrite nth_lt_cons.
  erewrite nth_lt_functional3.
  rewrite lind_compat; auto.
  pose proof hnm as hnm'.
  rewrite <- neq_nat_compat in hnm.
  rewrite hnm.
  destruct (eq_nat_dec m (S (lind nat_eq_dec _ _ h1))); subst.
  rewrite beq_nat_same.
  rewrite nth_lt_0'.
  erewrite nth_indep.
  rewrite <- nth_lt_compat. 
  rewrite nth_lt_list_subst_same; auto.
  rewrite length_list_subst.
  apply lt_lind. 
  destruct m. 
  contradict n; auto.
  contradict hnm'; auto.
  pose proof n as h5.
  rewrite <- neq_nat_compat in h5.
  rewrite h5.
  rewrite nth_lt_cons.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  rewrite nth_lt_list_subst_neq.
  apply nth_lt_functional3.
  omega.
  rewrite length_list_subst; auto with arith.

  Grab Existential Variables.
  rewrite length_list_subst; auto with arith.
  rewrite length_list_subst; auto with arith.
Qed.


  

Lemma transposed_list_nat_0_pred :
  forall l (pfl0:0 < length l) (pfl1:1 < length l)
         (pfp:pred (length l) < length l),
    transposed_list_nat l 0 (pred (length l)) pfl0 pfp = 
    nth_lt _ (pred (length l)) pfp ::
         seq_lind l 1 (pred (pred (length l)))
         pfl1 ++ hd' l (lt_length _ _ pfl0) ::nil.
  intros l h1 h2 h3.
  unfold transposed_list_nat, transpose_functor.
  rewrite map_seg_eq_iff.
  ex_goal.
  simpl.
  rewrite app_length, length_seq_lind.
  rewrite S_compat; auto. 
  omega.
  omega.
  exists hex.
  intros m h4 h5.
  destruct m as [|m]; simpl; auto.
  destruct (zerop (pred (length l))) as [hz | hz].
  rewrite hz at 1.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  erewrite nth_lt_app2.
  simpl.  
  rewrite length_seq_lind.
  rewrite hz; simpl; rewrite  <- minus_n_O; simpl.
  destruct m as [|m]; simpl.
  destruct l as [|l]; simpl.
  omega.
  simpl in hz.
  apply length_eq0 in hz; subst; auto.
  rewrite <- length_tl in hz.
  apply length_eq0 in hz; subst; auto.
  rewrite tl_eq_nil_iff in hz.
  destruct hz as [|h7]; subst.
  simpl in h1; omega.
  destruct h7; subst.
  simpl.
  destruct m; auto.
  rewrite <- pred_of_minus.
  auto with arith.
  (*


  
  destruct (nat_eq_dec m (pred (pred (length l)))) as [h6 | h6]; subst.
  rewrite minus_same, nth_lt0'; auto.
  pose proof (le_minus_iff m (pred (pred (length l)))) as hl.  
  hl hl.
  pose proof (lt_pred _ _ h4) as ha.
  red in ha.
  pose proof (le_pred _ _ ha) as h7.
  simpl in h7; auto.
  rewrite hl in hl0.
  rewrite hl0.
  rewrite nth_lt0'; auto.
  rewrite <- pred_of_minus.
  apply le_pred.
  apply le_pred_n.*)
  rewrite app_length, length_seq_lind.
  simpl.
  omega.
  omega.
  
  rewrite  (S_pred _ _ hz) at 1.
  pose proof (lt_length _ _ h1) as h6.
  pose proof (hd_compat' _ h6) as h7.
  simpl in h7.
  pose proof (nth_lt_functional1 _ _ (S m) h7 h4) as h8.
  hfirst h8.
  simpl.
  rewrite length_tl.
  omega.
  specialize (h8 hf).
  rewrite h8.
  pose proof hf as h9.
  simpl in h9.
  apply lt_S_n in h9.
  rewrite length_tl in h9.
  red in h9. 
  apply mono_pred_le in h9.
  simpl in h9. 
  apply le_lt_eq_dec in h9.
  destruct h9 as [h9 | h9].
  apply lt_neq in h9.
  pose proof h9 as h9'.
  rewrite <- neq_nat_compat in h9.
  rewrite h9; auto. 
  clear h5 h7.
  simpl.
  symmetry.
  erewrite nth_indep.
  erewrite <- nth_lt_compat.
  (*
  rewrite nth_lt_seq_lind.
  
  apply lt_S_n in h9.
  rewrite <- (S_pred _ _ h9); auto.
  simpl in hf.
  rewrite <- pred_of_minus.
  auto with arith.
  
  
  subst.
  rewrite beq_nat_same.
  simpl.
  simpl in hex.
  
  
  erewrite nth_indep.
  rewrite <- nth_lt_compat.*)

  
  erewrite nth_lt_app1.
  erewrite nth_lt_seq_lind.
  simpl.
  rewrite h8 at 1.
  simpl; auto.
  apply nat_eq_dec.
  rewrite app_length, length_seq_lind; simpl.  rewrite  S_compat, <- (S_pred _ _ hz).
  omega.
  rewrite <- pred_of_minus; auto with arith.
  rewrite h9 at 1.
  rewrite beq_nat_same.
  erewrite nth_indep.
  rewrite <- nth_lt_compat.
  erewrite nth_lt_app2; simpl.
  rewrite length_seq_lind, h9, minus_same at 1.
  rewrite nth_lt0'; auto.
  rewrite <- pred_of_minus; auto with arith.
  rewrite app_length; simpl; rewrite length_seq_lind, h9, S_compat; auto with arith.
  rewrite <- pred_of_minus; auto with arith.

  Grab Existential Variables.
  clear h5 h7.
  subst.
  rewrite length_seq_lind; auto with arith.
  rewrite <- pred_of_minus; auto with arith.
  rewrite app_length; simpl;
  rewrite S_compat, length_seq_lind,
  <- (S_pred _ _ hz), h9; auto with arith.
  rewrite <- pred_of_minus; auto with arith.
  rewrite length_seq_lind.
  pose proof (lt_pred _ _ h4) as h10.
  apply lt_dec in h10.
  destruct h10; try contradiction; auto.
  rewrite <- pred_of_minus; auto with arith.
  rewrite app_length, length_seq_lind, S_compat; simpl.
  rewrite <- (S_pred _ _ hz); auto.
  pose proof (lt_pred _ _ h4); auto. 
  rewrite <- pred_of_minus; auto with arith.
  rewrite length_seq_lind; auto with arith.
  rewrite hz; omega.
  rewrite <- pred_of_minus; auto with arith.
  rewrite app_length, length_seq_lind, S_compat, hz; simpl.
  omega.
  rewrite hz; auto with arith.
Qed.


Lemma last_transposed_list_nat_pred :
  forall l r
         (pfr:r < length l)
         (pfp:pred (length l) < length l)
         (d:nat),
    last (transposed_list_nat l r (pred (length l)) pfr pfp) d =
    nth_lt l r pfr.
  intros l r h1 h2 d. 
  rewrite (last_compat _ _
                       (lt_length _ _  
                                  (lt_eq_trans  _ _ _ h1
                                            (eq_sym (length_transposed_list_nat _ _ _ _ _))))).
  rewrite lst_eq.
  unfold transposed_list_nat.
  rewrite nth_lt_map_seg'.
  gen0.
  rewrite length_map_seg.
  intro h3.
  unfold transpose_functor.
  destruct (eq_nat_dec (pred (length l)) r) as [| h4]; subst.
  rewrite beq_nat_same.
  apply nth_lt_functional3.
  rewrite beq_nat_same.
  rewrite <- neq_nat_compat in h4.
  rewrite h4; auto.
Qed.


Lemma lst_transposed_list_nat_pred :
  forall l r
         (pfr:r < length l)
         (pfp:pred (length l) < length l)
         (pfne:transposed_list_nat l r (pred (length l)) pfr pfp <> nil),
    lst (transposed_list_nat l r (pred (length l)) pfr pfp) pfne  =
    nth_lt l r pfr.
  intros; rewrite lst_compat', last_transposed_list_nat_pred; auto.
Qed.


Lemma lst_transposed_list_nat :
  forall l r s
         (pfr:r < length l)
         (pfs:s < length l)
         (pfne:transposed_list_nat l r s pfr pfs <> nil),
    r < pred (length l) ->
    s < pred (length l) ->
    lst (transposed_list_nat l r s pfr pfs) pfne =
    lst l (iffn1 (transposed_list_nat_eq_nil_iff l r s pfr pfs) pfne).
  intros l r  s hr hs hne h1 h2.
  pose proof (lt_length _ _ hr) as hnn.
  unfold transposed_list_nat.
  rewrite lst_map_seg.
  unfold transpose_functor.
  destruct (eq_nat_dec (pred (length l)) r); subst.
  rewrite beq_nat_same.
  apply lt_irrefl in h1; contradiction.
  rewrite <- neq_nat_compat in n.
  rewrite n.
  destruct (eq_nat_dec (pred (length l)) s); subst.
  apply lt_irrefl in h2; contradiction.
  rewrite <- neq_nat_compat in n0.
  rewrite n0.
  rewrite nth_lt_eq_iff.
  exists (in_lst _ _). 
  pose proof (lt_lind nat_eq_dec _ _ (in_lst l hnn)) as h3.
  exists (pred (count_occ nat_eq_dec l
                          (lst l (iffn1 (transposed_list_nat_eq_nil_iff l r s hr hs) hne)))).
  split.
  eapply lt_pred_n.
  apply gt_lt.
  rewrite <- (count_occ_In nat_eq_dec).
  apply in_lst.
  gen0.
  do 2 rewrite lst_eq.
  intro h4.
  apply lind_occ_pred_nth_lt_pred.
Qed.         



Lemma no_dup_transposed_list_nat :
  forall l r s
         (pfr:r < length l)
         (pfs:s < length l),
    NoDup l ->
    NoDup (transposed_list_nat l r s pfr pfs).
  intros l r s h1 h2 h3.
  unfold transposed_list_nat.
  apply no_dup_map_seg_inj.
  apply inj_transpose_functor.
  apply inj_seg_nth_lt; auto.
Qed.


Lemma no_dup_transposed_list_nat' :
  forall l r s
         (pfr:r < length l)
         (pfs:s < length l),
    NoDup (transposed_list_nat l r s pfr pfs) ->
    NoDup l.
  intros l r s h1 h2 h3.
  unfold transposed_list_nat in h3.
  pose proof (no_dup_map_seg_inj' _ h3) as h4.
  pose proof (inj_transpose_functor' _ _ _ _ _ h4) as h5.
  apply inj_seg_nth_lt' in h5; auto.
  apply nat_eq_dec.
Qed.
  
Lemma list_to_set_transposed_list_nat :
  forall l r s
         (pfr:r < length l)
         (pfs:s < length l),
    list_to_set (transposed_list_nat l r s pfr pfs) = list_to_set l.
  intros l r s h1 h2; apply Extensionality_Ensembles; red; split; red; intros x h3; rewrite <- list_to_set_in_iff in h3; rewrite <- list_to_set_in_iff.
  rewrite in_transposed_list_nat_iff in h3; auto.
  rewrite in_transposed_list_nat_iff; auto.
Qed.


Lemma transposed_list_nat_eq : 
  forall (l l':list nat) (r s:nat)
         (pfr:r < length l)
         (pfs:s < length l)
         (pfe:transposed_list_nat l r s pfr pfs = l'),
    let pfr' := lt_eq_trans _ _ _
                            pfr
                            (eq_trans  (eq_sym (length_transposed_list_nat _ _ _ _ _))
                                       (length_subst _ _ pfe)) in
    let pfs' := lt_eq_trans _ _ _
                            pfs
                            (eq_trans (eq_sym (length_transposed_list_nat _ _ _ _ _))
                                      (length_subst _ _ pfe)) in
    l = transposed_list_nat l' r s pfr' pfs'.
  intros.
  pose proof (transposed_list_nat_functional _ _ pfe r s) as h1.
  hfirst h1.
  rewrite length_transposed_list_nat; auto.
  specialize (h1 hf).
  hfirst h1.
  rewrite length_transposed_list_nat; auto.
  specialize (h1 hf0).
  rewrite transposed_list_nat_undoes_itself in h1.
  rewrite h1.
  f_equal; apply proof_irrelevance.
Qed.



Definition cycled_seg_list n k :=
  map_seg (shift_seg_fun n k).


Lemma in_cycled_seg_list_iff :
  forall n k i,
    In i (cycled_seg_list n k) <->
    i < n.
  unfold cycled_seg_list; intros n k i; split; intro h1. 
  rewrite in_map_seg_iff in h1.
  destruct h1 as [m [h1 h2]]; subst. 
  apply lt_shift_seg_fun. 
  rewrite in_map_seg_iff.
  unfold shift_seg_fun.
  assert (h2:0 < n).
  omega. 
  destruct (lt_eq_lt_dec i k) as [[h3 | h3] | h3].
  pose proof (minus_mod_le).
  destruct (le_lt_dec (k - i) n) as [h4 | h4].
  apply le_lt_eq_dec in h4.
  destruct h4 as [h4 | h4].
  exists (n - (k - i)).
  ex_goal.
  apply lt_minus; auto.
  apply lt_le_weak; auto. 
  rewrite lt_iff in h3; auto.  
  exists hex.
  rewrite minus_minus_plus_same_le; auto with arith.
  rewrite plus_mod, mod_same, plus_0_l.
  rewrite mod_mod.
  apply lt_mod_eq; auto.
  subst.
  exists 0, h2.
  rewrite plus_0_l. 
  rewrite <- (lt_mod_eq i (k - i)) at 2; auto.
  symmetry.
  rewrite mod_eq_iff. 
  apply dv_refl.
  assumption.    
  auto with arith.

  exists ((k*(n - 1) + i) mod n), (lt_mod _ _ h2). 
  rewrite plus_mod.
  rewrite mod_mod at 1.
  rewrite <- minus_dist_l, <- plus_mod, mult_1_r, minus_plus_plus_same_le,
  mult_comm, plus_mod, times_mod_l, plus_0_l, mod_mod, lt_mod_eq; auto.
  pose proof (mult_O_le k n) as h5.
  destruct h5 as [|h5]; subst.
  apply lt_irrefl in h2; contradiction.
  rewrite mult_comm; auto.
  subst.
  exists _, h2.
  rewrite plus_0_l, lt_mod_eq; auto.
  exists ((k*(n - 1) + i) mod n), (lt_mod _ _ h2).
  pose proof (lt_trans _ _ _ h3 h1) as h4.
  rewrite <- (lt_mod_eq _ _ h4) at 2.
  rewrite <- minus_dist_l, <- plus_mod, mult_1_r, minus_plus_plus_same_le,
  plus_mod, times_mod_r, plus_0_l, mod_mod, lt_mod_eq; auto.
  pose proof (mult_O_le k n) as h5.
  destruct h5 as [|h5]; subst.
  apply lt_irrefl in h2; contradiction.
  rewrite mult_comm; auto.
Qed.


Lemma length_cycled_seg_list :
  forall k n,
    length (cycled_seg_list n k) = n.
  intros k n.
  unfold cycled_seg_list.
  rewrite length_map_seg; auto.
Qed.


Lemma list_to_set_cycled_seg_list :
  forall k n,
    list_to_set (cycled_seg_list n k) = seg n.
  intros k n; apply Extensionality_Ensembles; red; split; red; intros i h1.
  rewrite <- list_to_set_in_iff, in_cycled_seg_list_iff in h1; constructor; auto.
  destruct h1.
  rewrite <- list_to_set_in_iff.
  rewrite in_cycled_seg_list_iff; auto.
Qed.
 
 
Lemma shift_left_list :
  forall n,
    n > 0 ->
    cycled_seg_list n 1 =
    seq 1 (pred n) ++ 0 :: nil.
  intros n h1. 
  unfold cycled_seg_list.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite app_length, seq_length; simpl.
  omega.
  exists hex.
  intros m h2. 
  simpl.
  pose proof h2 as h3.
  apply lt_dec in h3.
  destruct h3 as [h3 | h3].
  erewrite nth_lt_app1.
  erewrite nth_lt_seq.
  unfold shift_seg_fun.
  rewrite plus_comm.
  rewrite lt_mod_eq; auto.
  omega.
  erewrite nth_lt_app2.
  simpl.
  rewrite seq_length.
  rewrite <- h3, minus_same.
  unfold shift_seg_fun.
  rewrite h3.
  rewrite S_compat.
  rewrite <- (S_pred _ _ h2), mod_same; auto.

  Grab Existential Variables.
  rewrite seq_length, h3; auto.
  rewrite seq_length; auto.
Qed.
 

Lemma shift_right_list :
  forall n,
    n > 0 ->
    cycled_seg_list n (pred n) =
    (pred n) :: seq 0 (pred n).
  intros n h1.
  unfold cycled_seg_list.
  rewrite map_seg_eq_iff.
  ex_goal.
  simpl; rewrite seq_length; auto.
  rewrite <- (S_pred _ _ h1); auto. 
  exists hex.
  intros m h2 h3.
  pose proof h2 as h4.
  destruct m as [|m].
  simpl.
  unfold shift_seg_fun.
  rewrite plus_0_l, lt_mod_eq; auto.
  eapply lt_pred_n.
  apply h4.
  rewrite nth_lt_cons, nth_lt_seq, plus_0_l.
  unfold shift_seg_fun.
  rewrite plus_S_shift_l, <- plus_S_shift_r.
  rewrite <- (S_pred _ _ h1), plus_mod, mod_same, plus_0_r.
  rewrite mod_mod, lt_mod_eq; auto with arith.
Qed.


Inductive permuted_seg_list : list nat -> Prop :=
| permuted_seg_list_id : forall n, permuted_seg_list (seg_list n)
| permuted_seg_list_transpose :
    forall (l':list nat) r s
           (pfr:r < length l') (pfs:s < length l'),
      permuted_seg_list l' ->
      permuted_seg_list (transposed_list_nat l' r s pfr pfs).

Lemma permuted_seg_list_nil :
  permuted_seg_list nil.
  change (permuted_seg_list (seg_list 0)).
  constructor.
Qed.


Lemma permuted_seg_list_0 :
  permuted_seg_list (0::nil).
  change (permuted_seg_list (seg_list 1)).
  constructor.
Qed.



Lemma lt_permuted_seg_list :
  forall (l:list nat) (i:nat),
    permuted_seg_list l ->
    In i l ->
    i < length l.
  intros l i h1. revert i.  
  induction h1 as [i|l r s h1 h2].  
  intros j h4.
  rewrite in_seg_list_iff in h4; auto.
  rewrite length_seg_list; auto.
  intros i h4.  
  rewrite in_transposed_list_nat_iff in h4; auto.
  rewrite length_transposed_list_nat; auto.
Qed.



Lemma permuted_seg_list_app_sing :
  forall (l:list nat),
    permuted_seg_list l ->
    permuted_seg_list (l++((length l)::nil)).
  intros l h1. 
  induction h1. 
  rewrite length_seg_list.
  rewrite <- seg_list_S.
  constructor. 
  rewrite length_transposed_list_nat.
  rewrite <- (transposed_list_nat_app_sing _ _ _ pfr pfs).
  constructor.
  assumption.
Qed.


Lemma no_dup_permuted_seg_list :
  forall (l:list nat),
    permuted_seg_list l ->
    NoDup l.
  intros l h1.
  induction h1.
  apply no_dup_seg_list.
  unfold transposed_list_nat.
  apply no_dup_map_seg_inj.
  apply inj_transpose_functor.
  apply inj_seg_nth_lt.
  assumption.
Qed.
(*
Lemma nth_lt_permuted_seg_list_preserve :
  forall (l:list nat),
    permuted_seg_list l ->
    seg_fun_lt_preserve (nth_lt l).
  intros l h1.
  induction h1.
  apply nth_lt_seg_list_preserve.
  red in IHh1; red.
  intros i h2.
  unfold transposed_list_nat.
  rewrite nth_lt_map_seg', length_map_seg.
  unfold transpose_functor.
  repeat destruct eq_nat_dec; auto.
Qed.
*)
Lemma list_to_set_permuted_seg_list :
  forall l,
    permuted_seg_list l ->
    list_to_set l = seg (length l).
  intros l h1.
  induction h1.
  rewrite list_to_set_seg_list_eq, length_seg_list; auto.
  rewrite length_transposed_list_nat.
  rewrite <- IHh1.
  rewrite list_to_set_transposed_list_nat; auto.
Qed.


Lemma transposed_list_nat_cons_app_sing :
  forall (l:list nat) (m n:nat)
         (pf0 : 0  < length (m :: l ++ n::nil))
         (pfS : S (length l) < length (m :: l ++ n::nil)),
    transposed_list_nat (m :: l ++ n::nil) 0 (S (length l))
    pf0 pfS =
    n :: l ++ m::nil.
  intros l m n h0 hS.
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff. 
  ex_goal. 
  simpl; f_equal.
  simpl; do 2 rewrite app_length; auto.
  exists hex.
  intros i hi.
  unfold transpose_functor.
  destruct (zerop i) as [|hz]; subst.
  rewrite beq_nat_same.
  rewrite nth_lt_S; simpl.
  erewrite nth_lt_app2.
  simpl.
  rewrite minus_same; auto.
  pose proof hz as hz'.
  apply lt_neq in hz.
  rewrite <- neq_nat_compat, beq_nat_sym in hz.
  rewrite hz.
  destruct (eq_nat_dec i (S (length l))) as [|hni]; subst.
  rewrite beq_nat_same.
  rewrite nth_lt0'.
  symmetry.
  gen0.
  rewrite app_comm_cons.
  intro h1.
  erewrite nth_lt_app2.
  simpl. 
  rewrite minus_same; auto.

  pose proof hni as hni'.
  rewrite <- neq_nat_compat in hni.
  rewrite hni.
  assert (h1:0 < i). omega.
  symmetry.
  gen0.
  generalize hi.
  rewrite (S_pred _ _ h1).
  intros h2 h3.
  do 2 rewrite nth_lt_S.
  simpl.
  do 2 erewrite nth_lt_app1.
  f_equal.

  Grab Existential Variables.
  simpl in h3.
  apply lt_S_n in h3.
  rewrite app_length, S_compat in h3.
  simpl in h2.
  
  omega.
  simpl; auto.
  auto.
Qed.

Lemma permuted_seg_list_S :
  forall (l:list nat),
    permuted_seg_list l ->
    permuted_seg_list (l ++ length l::nil).
  intros l h1.
  induction h1; simpl.
  rewrite length_seg_list, <- seg_list_S.
  constructor.
  rewrite length_transposed_list_nat.
  rewrite <- transposed_list_nat_app_sing.
  constructor.
  assumption.
Qed.



Lemma permuted_seg_list_n_seg_list : 
  forall n,
    permuted_seg_list (n::seg_list n).
  intro n. 
  induction n as [|n ih].
  rewrite seg_list0. 
  change (permuted_seg_list (seg_list 1)).
  constructor.
  rewrite seg_list_S.
  pose proof (transposed_list_nat_undoes_itself (S n::seg_list n++n::nil) 0 (S n) (lt_0_Sn _)) as h1.
  hfirst h1; simpl.
  apply lt_n_S.
  rewrite app_length; simpl; auto with arith.
  rewrite length_seg_list, S_compat; auto with arith.
  specialize (h1 hf).
  hfirst h1. 
  rewrite length_transposed_list_nat; simpl.
  rewrite app_length, S_compat.
  auto with arith.
  specialize (h1 hf0).
  hfirst h1. 
  rewrite length_transposed_list_nat; simpl.
  rewrite app_length, S_compat, length_seg_list; auto with arith.
  specialize (h1 hf1).   
  rewrite <- h1.  
  constructor. 
  pose proof (transposed_list_nat_cons_app_sing (seg_list n) (S n) n
                                                (lt_0_Sn _)) as h2.
  hfirst h2.
  simpl. 
  apply lt_n_S.
  rewrite app_length; simpl; rewrite S_compat; auto with arith.
  specialize (h2 hf2).
  match goal with
      |- permuted_seg_list 
           (transposed_list_nat
              _ _ _ ?pf _) => generalize pf end.
  generalize hf. revert h2.
  simpl.
  gen0. 
  rewrite <- (length_seg_list n).
  rewrite length_seg_list.
  intros h3 h4 h5 h6.
  erewrite f_equal_dep in h4. 
  rewrite h4 at 1.
  rewrite length_seg_list.
  rewrite app_comm_cons.
  pose proof (permuted_seg_list_S (n::seg_list n) ih) as h7.
  simpl in h7.
  rewrite length_seg_list in h7.
  rewrite <- app_comm_cons; auto.
  apply proof_irrelevance.
Qed.



Lemma permuted_seg_list_cons :
  forall (l:list nat),
    permuted_seg_list (length l :: seg_list (length l)).
  intros; apply permuted_seg_list_n_seg_list.
Qed.


    
Lemma permuted_seg_list_cons' :
  forall (l:list nat),
    permuted_seg_list l ->
    permuted_seg_list (length l::l).
  intros l h1.
  induction h1; simpl.   
  rewrite length_seg_list.  
  apply permuted_seg_list_n_seg_list.
  rewrite length_transposed_list_nat.
  pose proof (transposed_list_nat_undoes_itself (length l'::transposed_list_nat l' r s pfr pfs) (S r) (S s)) as h2.
  simpl in h2.
  hfirst h2.
  rewrite length_transposed_list_nat; auto with arith.
  specialize (h2 hf).
  hfirst h2.
  rewrite length_transposed_list_nat; auto with arith.
  specialize (h2 hf0).
  hfirst h2.
  rewrite length_transposed_list_nat; auto with arith.
  specialize (h2 hf1).
  hfirst h2.
  rewrite length_transposed_list_nat; auto with arith.
  specialize (h2 hf2).
  rewrite <- h2. 
  constructor.
  rewrite transposed_list_nat_S; auto.
Qed.


Lemma map_seg_shift_fun_0n :
  forall n,
    map_seg (shift_seg_fun 0 n) = nil.
  intro; simpl; auto.
Qed.

Definition cycled_list_nat l k (pf:0 < length l) : list nat :=
  map_seg (shift_functor (nth_lt l) k pf).

Definition left_shift_list_nat l (pf:0 < length l) : list nat :=
  cycled_list_nat l (pred (length l)) pf .

Definition right_shift_list_nat l (pf:0 < length l) : list nat :=
  cycled_list_nat l 1 pf.

Lemma length_cycled_list_nat :
  forall l k (pf:0 < length l),
    length (cycled_list_nat l k pf) = length l.
  intros l k h1.
  unfold cycled_list_nat.
  rewrite length_map_seg; auto.
Qed.

Lemma length_left_shift_list_nat :
  forall l (pf:0 < length l),
    length (left_shift_list_nat l pf) = length l.
  unfold left_shift_list_nat; intros; apply length_cycled_list_nat.
Qed.

Lemma length_right_shift_list_nat :
  forall l (pf:0 < length l),
    length (right_shift_list_nat l pf) = length l.
  unfold right_shift_list_nat; intros; apply length_cycled_list_nat.
Qed.

Lemma nth_lt_cycled_list_nat :
  forall l k i (pf0:0 < length l)
         (pfi:i < length (cycled_list_nat l k pf0)),
    nth_lt (cycled_list_nat l k pf0) i pfi =
    nth_lt l ((i + k) mod (length l)) (lt_mod _ _ pf0).
  intros l k i h1 h2.
  unfold cycled_list_nat.
  rewrite nth_lt_map_seg'.
  unfold shift_functor; auto.
Qed.

Lemma nth_lt_left_shift_list_nat : 
  forall l i (pf0:0 < length l)
         (pfi:i < length (left_shift_list_nat l pf0)),
    nth_lt (left_shift_list_nat l pf0) i pfi =
    nth_lt l ((i + (pred (length l))) mod (length l)) (lt_mod _ _ pf0).
  intros l k i h1.
  unfold left_shift_list_nat.
  rewrite nth_lt_cycled_list_nat.
  unfold shift_functor; auto.
Qed.

Lemma nth_lt_right_shift_list_nat : 
  forall l i (pf0:0 < length l)
         (pfi:i < length (right_shift_list_nat l pf0)),
    nth_lt (right_shift_list_nat l pf0) i pfi =
    nth_lt l ((S i) mod (length l)) (lt_mod _ _ pf0).
  intros l k i h1.
  unfold right_shift_list_nat.
  rewrite nth_lt_cycled_list_nat.
  unfold shift_functor; auto.
  gen0.
  rewrite S_compat.
  intros; apply nth_lt_functional3.
Qed.

Lemma right_shift_list_nat_eq : 
  forall l (pf:0 < length l),
    let pfn := lt_length _ _ pf in
    right_shift_list_nat l pf =
    (tl l) ++ hd' l pfn :: nil.
  intros l h1 h2.
  unfold right_shift_list_nat.
  unfold cycled_list_nat.
  rewrite map_seg_eq_iff.
  simpl.
  unfold h2.
  clear h2.
  ex_goal. 
  rewrite app_length; simpl.
  rewrite length_tl; omega.
  exists hex. 
  intros m h3.
  pose proof h3 as h5.
  apply lt_dec in h5.
  destruct h5 as [h5 | h5].
  unfold shift_functor.
  pose proof (lt_n_S _ _ h5) as h6.
  rewrite <- (S_pred _ _ h3) in h6.
  gen0.
  rewrite S_compat, (lt_mod_eq _ _ h6).
  intro h7.
  symmetry.
  gen0.
  intro.
  rewrite <- length_tl in h5.
  rewrite (nth_lt_app1 _ _ _ h5).
  symmetry.
  generalize h7.
  apply lt_length in h3.
  generalize h5.
  rewrite (hd_compat' _ h3).
  intros h8 h9.
  rewrite nth_lt_cons.
  simpl.
  apply nth_lt_functional3.
  subst.
  unfold shift_functor.
  gen0.

  rewrite S_compat, <- (S_pred _ _ h1).
  intro h4.
  erewrite nth_lt_app2.
  symmetry.
  gen0.
  rewrite length_tl, <- minus_n_n.
  intro h5.
  simpl.
  symmetry.
  generalize h4.
  rewrite mod_same. 
  intro h6. 
  rewrite nth_lt0'.
  apply hd_functional'.

  Grab Existential Variables.
  rewrite length_tl; auto with arith.
Qed.


Lemma right_shift_cons : 
  forall l a (pf0:0 < length (a::l)),
    right_shift_list_nat (a::l) pf0 =
    l ++ a::nil.
  intros l a h1.
  unfold right_shift_list_nat, cycled_list_nat.
  rewrite map_seg_eq_iff.
  ex_goal.
  simpl; rewrite app_length; simpl; rewrite S_compat; auto.
  exists hex.
  intros m h2.
  simpl.
  unfold shift_functor.
  gen0.
  rewrite S_compat, S_mod.
  destruct (nil_dec' l) as [|h3]; subst.
  simpl; simpl in h2.
  intro h3.
  destruct m as [|m]; auto.
  destruct m as [|m]; auto.
  apply not_nil_lt in h3.
  pose proof (lt_n_S _ _ h3) as h4.
  rewrite (lt_mod_eq _ _ h4).
  simpl in h2.
  rewrite (lt_mod_eq _ _ h2), S_compat.
  pose proof h2 as h2'.
  apply lt_dec in h2'.
  simpl in h2'.
  destruct h2' as [h2'|h2']; subst.
  pose proof (lt_n_S _ _ h2') as h5.
  rewrite (lt_mod_eq _ _ h5).
  intro.
  rewrite nth_lt_cons.
  erewrite nth_lt_app1.
  apply nth_lt_functional3.
  rewrite mod_same.
  intros h5.
  rewrite nth_lt_0'.
  symmetry.
  erewrite nth_lt_app2.
  simpl.
  rewrite minus_same; auto.

  Grab Existential Variables.
  apply le_refl.
  assumption.
Qed.


Lemma cycled_seg_list_0n :
  forall n,
    cycled_seg_list 0 n = nil.
  intro; unfold cycled_seg_list;
  apply map_seg_shift_fun_0n.
Qed.

Lemma cycled_seg_list_n0 :
  forall n,
    cycled_seg_list n 0 = seg_list n. 
  unfold cycled_seg_list.
  intro n. rewrite (shift_seg_fun0 n).
  apply map_seg_id_seg.
Qed.

Lemma cycled_seg_list_nS :
  forall n k (pf:0 < length (cycled_seg_list n k)),
    cycled_seg_list n (S k) =
    right_shift_list_nat (cycled_seg_list n k) pf.
  intros n k h1.   
  unfold cycled_seg_list.
  rewrite map_seg_eq_iff.
  ex_goal. 
  rewrite length_right_shift_list_nat, length_map_seg; auto.
  exists hex.
  intros m h2 h3. 
  rewrite nth_lt_right_shift_list_nat, nth_lt_map_seg'.
  symmetry.
  gen0.  
  rewrite length_map_seg.
  intro h4.
  unfold shift_seg_fun.
  pose proof h2 as h5.
  apply lt_dec in h5.
  destruct h5 as [h5 | h5].
  rewrite (lt_mod_eq (S m)) in h4.
  rewrite (lt_mod_eq (S m)).
  rewrite plus_S_shift_r, <- plus_S_shift_l; auto.
  assumption.
  rewrite (S_pred _ _ h2).
  apply lt_n_S; auto.
  rewrite h5.
  rewrite <- (S_pred _ _ h2), mod_same, plus_0_l.
  rewrite plus_S_shift_r, <- plus_S_shift_l; auto.
  rewrite <- (S_pred _ _ h2).
  rewrite plus_mod, mod_same, plus_0_l, mod_mod; auto.
Qed.



Lemma right_shift_seg_list :
  forall n (pf:0 < length (seg_list n)),
    right_shift_list_nat (seg_list n) pf =
    seq 1 (pred n) ++ 0 :: nil.
  intros n h1. 
  pose proof h1 as h1'.
  rewrite length_seg_list in h1'.
  unfold right_shift_list_nat, cycled_list_nat.
  rewrite map_seg_eq_iff.
  ex_goal.
  rewrite length_seg_list, app_length, seq_length; simpl.
  rewrite S_compat. 
  rewrite <- (S_pred _ _ h1'); auto.
  exists hex.
  intros m h2 h3.
  unfold shift_functor.
  rewrite S_compat.
  rewrite (nth_lt_eq_iff nat_eq_dec).
  ex_goal.
  pose proof h2 as h4.
  rewrite length_seg_list in h4.
  apply lt_dec in h4. 
  destruct h4 as [h4 | h4].
  pose proof @nth_lt_app1.
  pose proof (nth_lt_app1 (seq 1 (pred n)) (0::nil) m) as h5.
  hfirst h5.
  rewrite seq_length; auto.
  specialize (h5 hf h3).
  rewrite h5.
  rewrite nth_lt_seq. 
  rewrite plus_comm, S_compat.
  rewrite in_seg_list_iff.
  omega.
  generalize h3.
  rewrite h4.
  intro h5.
  pose proof (nth_lt_app2 (seq 1 (pred n)) (0::nil) (pred n)) as h6.
  hfirst h6.
  rewrite seq_length; auto.
  specialize (h6 hf h5).
  rewrite h6.
  simpl. 
  rewrite seq_length, minus_same. 
  rewrite in_seg_list_iff; auto. 
  exists hex0.
  exists 0. 
  split.
  rewrite count_occ_no_dup; auto.
  apply no_dup_seg_list.
  change (lind nat_eq_dec (seg_list n)  (nth_lt (seq 1 (pred n) ++ 0 :: nil) m h3) hex0 = S m mod length (seg_list n)).
  pose proof h2 as h4.
  rewrite length_seg_list in h4.
  apply lt_dec in h4. 
  destruct h4 as [h4 | h4].
  gen0.
  pose proof (nth_lt_app1 (seq 1 (pred n)) (0::nil) m) as h6.
  hfirst h6.
  rewrite seq_length; auto.
  specialize (h6 hf h3).
  rewrite h6.
  rewrite nth_lt_seq, plus_comm, S_compat.
  intro h5.
  rewrite lind_seg_list, lt_mod_eq; auto.
  rewrite length_seg_list; auto.
  rewrite in_seg_list_iff in h5; auto.
  unfold h3. unfold h3 in hex0.
  clear h3.
  subst.
  gen0.
  erewrite nth_lt_app2. 
  intro h5.
  rewrite lind_eq_iff.
  split.
  rewrite nth_lt_seg_list.
  symmetry.
  gen0.
  rewrite seq_length, <- minus_n_n.
  intro; simpl.
  rewrite <- (S_pred _ _ h1'), length_seg_list, mod_same; auto.
  intro h6.
  rewrite in_firstn_iff in h6.
  destruct h6 as [i [h6 h7]].
  rewrite nth_lt_seg_list in h7.
  symmetry in h7.
  revert h7.
  gen0.
  rewrite seq_length, <- minus_n_n.
  intro h7.
  simpl.
  intro; subst.
  rewrite <- (S_pred _ _ h1'), length_seg_list, mod_same in h6. 
  apply lt_irrefl in h6; auto.
  
  Grab Existential Variables.
  rewrite <- (S_pred _ _ h1'), length_seg_list, mod_same; auto with arith.
  rewrite <- (S_pred _ _ h1'), length_seg_list, mod_same; auto.
  rewrite seq_length; auto.
Qed.



Lemma transposed_list_nat_seq_app_0n :
  forall n (pf:0 < pred n)
         (pfp:pred n < length (seq 1 (pred n) ++ 0 :: n :: nil))
         (pfn: n < length (seq 1 (pred n) ++ 0 :: n :: nil)),
    transposed_list_nat (seq 1 (pred n) ++ 0 :: n :: nil)
                        (pred n)
                        n
                        pfp
                        pfn = 
    seq 1 n ++ 0::nil.
  intros n h1 h2 h3. 
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff.  
  ex_goal. 
  do 2 rewrite app_length; simpl.
  do 2 rewrite seq_length. 
  omega. 
  exists hex.
  intros m h4; simpl.
  pose proof h4 as h4'.
  rewrite app_length, seq_length in h4'.
  simpl in h4'.
  pose proof (O_lt_pred_n_lt_n _ h1) as hc.
  rewrite plus_S_shift_r, S_compat, <- (S_pred _ _ hc) in h4'.
  unfold transpose_functor.
  destruct (eq_nat_dec m (pred n)) as [|hr]; subst.
  rewrite beq_nat_same.
  erewrite nth_lt_app2.
  erewrite nth_lt_app1, nth_lt_seq. 
  simpl.
  rewrite seq_length.
  rewrite n_minus_pred_n; auto. 
  rewrite <- (S_pred _ _ hc); auto.
  pose proof hr as hr'.
  rewrite <- neq_nat_compat in hr.
  rewrite hr.
  destruct (eq_nat_dec m (pred n)) as [|hr0]; subst.
  pose proof (lt_pred_n _ _ hc) as hp.
  apply lt_neq in hp.
  rewrite <- neq_nat_compat in hp.
  rewrite hp.
  do 2 erewrite nth_lt_app2.
  gen0.
  rewrite seq_length, <- minus_n_n.
  intro.
  symmetry.
  simpl.
  rewrite seq_length.
  pose proof hc as h5.
  apply lt_le_weak in h5.
  rewrite le_minus_iff in h5.
  rewrite h5; auto. 
  destruct (eq_nat_dec m n) as [|h5]; subst.
  rewrite beq_nat_same.
  erewrite nth_lt_app2.
  erewrite nth_lt_app2.
  simpl.
  rewrite seq_length, minus_same, seq_length, minus_same; auto.
  pose proof h5 as h5'.
  rewrite <- neq_nat_compat in h5.
  rewrite h5.
  erewrite nth_lt_app1.
  erewrite nth_lt_app1.
  do 2 rewrite nth_lt_seq; auto.
   
  Grab Existential Variables.
  rewrite seq_length.
  
  apply lt_dec in h4'.
  simpl in h4'.
  destruct h4' as [he | he]; subst; auto.
  contradict hr0; auto.
  contradict h5'; auto.
  
  rewrite seq_length.
  apply lt_dec in h4'.
  simpl in h4'.
  destruct h4' as [he | he]; subst; auto.  
  apply lt_dec in he.
  destruct he; subst; auto.
  contradict hr; auto. 
  contradict h5'; auto.
  rewrite seq_length; auto.
  rewrite seq_length; auto.
  rewrite seq_length; auto.
  contradict hr'; auto.
  contradict hr'; auto. 
  rewrite seq_length; auto.
  rewrite seq_length; auto with arith.
Qed.

(*Perhaps erase this, the model for the above, if it's not needed!*)
Lemma transposed_list_nat_seq_app0 :
  forall n (pf:0 < (pred (pred n)))
         (pfpp:pred (pred n) < (length (seq 1 (pred n) ++ 0 :: nil)))
         (pfp: pred n < (length (seq 1 (pred n) ++ 0 :: nil))),
    transposed_list_nat (seq 1 (pred n) ++ 0 :: nil)
                        (pred (pred n))
                        (pred n)
                        pfpp
                        pfp = 
    seq 1 (pred (pred n)) ++ 0::pred n::nil.
  intros n h1 h2 h3. 
  unfold transposed_list_nat.
  rewrite map_seg_eq_iff.  
  ex_goal.
  do 2 rewrite app_length; simpl.
  do 2 rewrite seq_length.
  omega. (*maybe speed up!*) 
  exists hex.
  intros m h4; simpl.
  pose proof h4 as h4'.
  rewrite app_length, seq_length in h4'.
  simpl in h4'.
  rewrite S_compat in h4'.
  pose proof (O_lt_pred_n_lt_n _ h1) as ha.
  pose proof (lt_trans _ _ _ h1 ha) as hb.
  pose proof (O_lt_pred_n_lt_n _ hb) as hc.
  pose proof (lt_trans _ _ _ hb hc) as hd.
  rewrite <- (S_pred _ _ hd) in h4'.
  unfold transpose_functor.
  destruct (eq_nat_dec m (pred (pred n))) as [|hm]; subst.
  rewrite beq_nat_same.
  do 2 erewrite nth_lt_app2.
  gen0.
  rewrite seq_length, <- minus_n_n.
  intro h5.

  symmetry.  
  gen0.
  rewrite seq_length, <- minus_n_n.
  intro; simpl; auto.
  pose proof hm as hm'.
  rewrite <- neq_nat_compat in hm'.
  rewrite hm'.
  destruct (eq_nat_dec m (pred n)) as [|hr0]; subst.
  
  rewrite beq_nat_same.
  erewrite nth_lt_app1, nth_lt_seq. 
  erewrite nth_lt_app2.
  symmetry.
  gen0.
  rewrite seq_length.
  pose proof (n_minus_pred_n (pred n) h1) as h8.
  rewrite h8.
  intro h9.
  simpl.
  pose proof (O_lt_pred_n_lt_n _ h1) as h6.
  pose proof h6 as h7.
  apply lt_le_weak in h6.
  rewrite le_minus_iff in h6.
  rewrite <- (S_pred _ _ h7); auto.
  pose proof hr0 as hr0'.
  rewrite <- neq_nat_compat in hr0'.
  rewrite hr0'.
  do 2 erewrite nth_lt_app1, nth_lt_seq; auto.
 
  Grab Existential Variables.
  rewrite seq_length.
  apply lt_dec in h4'.
  destruct h4' as [h5 | h5]; subst.
  apply lt_dec in h5.
  destruct h5 as [h5 | h5]; subst; auto.
  contradict hm; auto.
  contradict hr0; auto.
  rewrite seq_length.
  apply lt_dec in h4'.
  destruct h4' as [h5 | h5]; subst; auto.
  contradict hr0; auto.
  rewrite seq_length; auto with arith.
  rewrite seq_length; auto.
  rewrite seq_length; auto.
  rewrite seq_length; auto.
Qed.



Lemma permuted_seg_list_cons'' :
  forall (l:list nat) (a:nat)
         (pfi:In (length l) l)
         (pf:a < length l),
    ~In a l ->
    permuted_seg_list (list_subst
                         l _
                         (lt_lind nat_eq_dec _ _ pfi) a) ->
    permuted_seg_list (a::l).
  intros l a h1 h2 h3 h4.
  pose proof
       (transposed_list_nat_undoes_itself
          (a::l) 0
          (S (lind nat_eq_dec _ _ h1))
          (lt_0_Sn _)) as h5.
  hfirst h5.
  simpl.
  apply lt_n_S.
  apply lt_lind.
  specialize (h5 hf).
  hfirst h5.
  rewrite length_transposed_list_nat.
  apply lt_0_Sn.
  specialize (h5 hf0).
  hfirst h5.
  rewrite length_transposed_list_nat.
  simpl.
  apply lt_n_S.
  apply lt_lind.
  specialize (h5 hf1).
  rewrite <- h5.
  constructor.
  rewrite transposed_list_nat_cons_0_S_lind.
  gterm0.
  redterm0 c.
  redterm1 c0.
  rewrite <- (length_list_subst l _ c1) at 1.
  apply permuted_seg_list_cons'.
  gterm0.
  redterm1 c2.
  fold c3.
  assert (h6:c3 = lt_lind _ _ _ _).
    apply proof_irrelevance.
  rewrite h6; auto.
Qed.


(*Transposes the maximum number and the maximally-indexed element, 
then removes the maximum numbe, 
i.e. removes [pred n] with [lind _ (pred n) _ _] .*)
Definition removelast_list_to_set_seg
             (l:list nat) (pfnd:NoDup l)
             (n:nat) (pfn:0 < n)
             (pfe:list_to_set l = seg n) : list nat :=
    let pfi := iff2 (list_to_set_in_iff l _) (in_eq_set _ _ (eq_sym pfe) _ (in_pred_seg _ pfn)) in
    removelast
      (transposed_list_nat l (lind nat_eq_dec l (pred n) pfi)
                           (pred n)
                           (lt_lind nat_eq_dec l (pred n) pfi)
                           (list_to_set_eq_seg_lt' _ _ _ pfnd pfe pfi)).

Lemma removelast_list_to_set_seg_functional :
  forall (l l':list nat)
         (pfnd: NoDup l)
         (pfnd':NoDup l')
         (n:nat) (pfn:0 < n)
         (pfe: list_to_set l = seg n)
         (pfe':list_to_set l' = seg n),
    l = l' ->
    removelast_list_to_set_seg l pfnd n pfn pfe =
    removelast_list_to_set_seg l' pfnd' n pfn pfe'.
  intros; subst; f_equal; apply proof_irrelevance.
Qed.

Lemma length_removelast_list_to_set_seg :
  forall (l:list nat) (pfnd:NoDup l)
         (n:nat) (pfn:0 < n)
         (pfe:list_to_set l = seg n),
    length (removelast_list_to_set_seg l pfnd n pfn pfe) = pred n.
  unfold removelast_list_to_set_seg; intros.
  rewrite length_removelast, length_transposed_list_nat.
  pose proof (list_to_set_eq_seg _ _ pfnd pfe); subst; auto.
Qed.


Lemma no_dup_removelast_list_to_set_seg :
  forall l (pfnd:NoDup l) n (pfn:0 < n) (pfe:list_to_set l = seg n),
    NoDup (removelast_list_to_set_seg l pfnd n pfn pfe).
  intros.
  unfold removelast_list_to_set_seg.
  apply no_dup_removelast.
  apply nat_eq_dec.
  apply no_dup_transposed_list_nat; auto.
Qed.


Lemma seg_list_pred_eq_removelast_to_set_seg :
  forall (l:list nat) (pfnd:NoDup l) (pfl:0 < length l)
         (pfe:list_to_set l = seg (length l)),
    seg_list (pred (length l)) =
    removelast_list_to_set_seg l pfnd (length l) pfl pfe ->
    let pfll :=  iff2 (list_to_set_in_iff l (pred (length l)))
                      (in_eq_set (seg (length l)) (list_to_set l) 
                                 (eq_sym pfe) (pred (length l)) (in_pred_seg (length l) pfl)) in
    l = transposed_list_nat (seg_list (length l))
                        (lind nat_eq_dec l (pred (length l)) pfll)
                        (pred (length l))
                        (lt_eq_trans _ _ _ (lt_lind nat_eq_dec _ _ pfll)
                                     (eq_sym (length_seg_list _)))
                        (lt_eq_trans _ _ _
                                     (lt_pred_n _ _ pfl)
                                     (eq_sym (length_seg_list _))). 
  intros l hnd h1 h2 h3.
  unfold removelast_list_to_set_seg in h3.
  symmetry in h3.
  rewrite removelast_eq_iff in h3.
  erewrite lst_compat' in h3 at 1.
  rewrite last_transposed_list_nat_pred in h3.
  rewrite lind_compat in h3.
  rewrite <- seg_list_S in h3.
  pose proof (transposed_list_nat_functional _ _ h3   (lind nat_eq_dec l (pred (length l))
              (iff2 (list_to_set_in_iff l (pred (length l)))
                 (in_eq_set (seg (length l)) (list_to_set l) 
                    (eq_sym h2) (pred (length l)) (in_pred_seg (length l) h1))))
                                         (pred (length l))) as h4.
  hfirst h4.
  rewrite length_transposed_list_nat.
  apply lt_lind.
  specialize (h4 hf).
  hfirst h4. 
  rewrite length_transposed_list_nat.
  eapply lt_pred_n.
  apply h1.
  specialize (h4 hf0).
  rewrite transposed_list_nat_undoes_itself in h4.
  simpl.
  rewrite h4 at 1.
  erewrite transposed_list_nat_functional.
  f_equal.
  apply proof_irrelevance.
  apply proof_irrelevance.

  Grab Existential Variables.
  rewrite <- (S_pred _ _ h1); auto.
  intro h4.
  rewrite transposed_list_nat_eq_nil_iff in h4.
  pose proof (lt_length _ _ h1); contradiction.
Qed.

  
Lemma nth_lt_removelast_list_to_set_seg :
  forall l (pfnd:NoDup l)
         (pflt:0 < length l)
         (pfe:list_to_set l = seg (length l))
         (r:nat)
         (pfr:r < length (removelast_list_to_set_seg l pfnd (length l) pflt pfe)),
    r = lind nat_eq_dec l (pred (length l))
               (iff2 (list_to_set_in_iff l (pred (length l)))
                     (in_eq_set _ _ (eq_sym pfe) _
                                (in_seg _ _ (lt_pred_n _ (length l) pflt)))) \/
    nth_lt (removelast_list_to_set_seg l pfnd (length l) pflt pfe) r pfr =
    nth_lt l r
           (lt_trans _ _ _
                     (lt_eq_trans
                          _ _ _
                          pfr
                          (length_removelast_list_to_set_seg l pfnd (length l) pflt pfe))
                     (lt_pred_n _ (length l) pflt)).
  intros l hnd hlt he r hr.
  unfold removelast_list_to_set_seg.
  erewrite nth_lt_removelast.
  unfold transposed_list_nat.
  rewrite nth_lt_map_seg'.
  unfold transpose_functor.
  (*One day clean this up with a match goal or tactic!!!*)
  destruct (eq_nat_dec r (lind nat_eq_dec _ _   (iff2 (list_to_set_in_iff l (pred (length l)))
               (in_eq_set (seg (length l)) (list_to_set l) 
                          (eq_sym he) (pred (length l)) (in_pred_seg (length l) hlt))))) as [ha | ha]; subst.
  rewrite beq_nat_same.
  left.
  apply lind_functional3.
  pose proof ha as ha'.
  rewrite <- neq_nat_compat in ha'.
  rewrite ha'.
  destruct (eq_nat_dec r (pred (length l))) as [hc | hc]; subst.
  rewrite beq_nat_same.
  right.
  pose proof hr as h2.
  rewrite length_removelast_list_to_set_seg in h2.
  omega.
  pose proof hc as hc'.
  rewrite <- neq_nat_compat in hc'.
  rewrite hc'.
  right.
  apply nth_lt_functional3.
  apply nat_eq_dec.

  Grab Existential Variables.
  rewrite length_transposed_list_nat.
  rewrite length_removelast_list_to_set_seg in hr; omega.
Qed.


Inductive permuted_lists {T:Type} :
  list T -> list T ->  Prop :=
| permuted_lists_nil : permuted_lists nil nil 
| permuted_lists_intro :
    forall (l l':list T)
           (l0:list nat) r s
           (pfr:r < length l0) (pfs:s < length l0)
           (pfp:permuted_seg_list l0)
           (pfe:length l0 = length l'),
      permuted_lists
        l'
        (map_in_dep (fun i (pfi:In i l0) =>
                       let pfilt := lt_permuted_seg_list _ _ pfp pfi in
                       let pflt := lt_eq_trans _ _ _ pfilt pfe in
                       nth_lt l' i pflt)).
