(*Copyright (C) 2016-2019, Daniel Wyckoff, except for the 
  functions labeld "Schepler", which were copied and pasted from
  Daniel Schepler's "Zorn's Lemma."*)

(*This file is part of BooleanAlgebrasIntro2.

BooleanAlgebrasIntro2 is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

BooleanAlgebrasIntro2 is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with BooleanAlgebrasIntro2.  If not, see <http://www.gnu.org/licenses/>.*)

(*Version 0.3 bool2 -- certified -- Coq 8.4 pl4.*)

Require Export Relation_Definitions.
Require Import Relation_Definitions_Implicit.
Require Import Classical_Wf. 
Require Import Description.
Require Import FunctionalExtensionality.
Require Import Classical.
Require Import SetUtilities.
Require Import Relations.
Require Import LogicUtilities.
Require Import ZornsLemma.
Require Import Arith.
Require Import Equality.
Require Import ArithUtilities.
Require Import DecidableDec.
Require Import EqDec.
Require Import FunctionProperties. 

Section WellOrder.

Variable T:Type.

(*Schepler*)
Record well_order (R:relation T) : Prop := {
  wo_well_founded: well_founded R;
  wo_total_strict_order: total_strict_order R
}.

(*Schepler*)
Lemma wo_irrefl: forall R:relation T, well_order R ->
  (forall x:T, ~ R x x).
Proof.
intuition.
assert (forall y:T, Acc R y -> y <> x).
intros.
induction H1.
intuition.
rewrite H3 in H2.
apply H2 with x.
trivial.
trivial.

pose proof (wo_well_founded R H).
unfold well_founded in H2.
pose proof (H1 x (H2 x)).
auto.
Qed.

(*Schepler*)
Lemma wo_antisym: forall R:relation T, well_order R ->
  (forall x y:T, R x y -> ~ R y x).
Proof.
intuition.
assert (forall z:T, Acc R z -> z <> x /\ z <> y). 
intros.
induction H2. 
intuition.
rewrite H4 in H3.
pose proof (H3 y H1).
tauto.
rewrite H4 in H3.
pose proof (H3 x H0).
tauto.

pose proof (wo_well_founded R H).
unfold well_founded in H3.
pose proof (H2 x (H3 x)).
tauto.
Qed.

(*Schepler*)
Lemma wo_transitive: forall R:relation T, well_order R -> transitive R.
Proof.
intros.
unfold transitive.
intros.
case (wo_total_strict_order R H x z).
trivial.
intro.
case H2.
intro.
rewrite H3 in H0.
pose proof (wo_antisym R H y z).
contradict H0.
auto.

intro.

assert (forall a:T, Acc R a -> a <> x /\ a <> y /\ a <> z).
intros.
induction H4.
intuition.
rewrite H2 in H5.
pose proof (H5 z H3).
tauto.
rewrite H2 in H5.
pose proof (H5 x H0).
tauto.
rewrite H2 in H5.
pose proof (H5 y H1).
tauto.
rewrite H2 in H5.
pose proof (H5 z H3).
tauto.
rewrite H2 in H5.
pose proof (H5 x H0).
tauto.
rewrite H2 in H5.
pose proof (H5 y H1).
tauto.

pose proof (wo_well_founded R H).
unfold well_founded in H5.
pose proof (H4 x (H5 x)).
tauto.
Qed.



(*Wyckoff*)
Definition unq_minimal_element_property (R : relation T) :=
  forall S : Ensemble T,
    Inhabited S -> exists! x : T, Inn S x /\ (forall y : T, Inn S y -> ~ R y x).

(*Wyckoff*)
Lemma wo_implies_umep : 
  forall {R:Relation T},
    well_order R -> unq_minimal_element_property R.
intros R h1.
destruct h1 as [h1 h2].
pose proof (WF_implies_MEP _ _ h1) as h3.
red in h3. red.
intros S h4. specialize (h3 S h4).
destruct h3 as [x [h5 h6]].
exists x. red. split. split; auto.
intros x' h7. destruct h7 as [h7 h8].
red in h2.
specialize (h2 x x'). destruct h2 as [h2 | h2].
specialize (h8 _ h5). contradiction.
destruct h2 as [h2 | h2]; auto.
specialize (h6 _ h7). contradiction.
Qed.


(*Wyckoff*)
Definition min_set (R:relation T) (pfw:well_order R) 
           (A:Ensemble T) (pfi:Inhabited A) : T :=
  (proj1_sig (constructive_definite_description _ ((wo_implies_umep pfw) A pfi))).

(*Wyckoff*)
Lemma min_set_compat :
  forall (R:relation T) (pfw:well_order R) 
         (A:Ensemble T) (pfi:Inhabited A),
    let m := min_set R pfw A pfi in
     Inn A m /\ (forall y : T, Inn A y -> ~ R y m).
intros R h1 A pfi m. unfold m, min_set.
destruct constructive_definite_description. simpl.
assumption.
Qed.

(*Wyckoff*)
Lemma min_set_compat' : 
  forall (R:relation T) (pfw:well_order R) 
         (A:Ensemble T) (pfi:Inhabited A) (y:T),
    R y (min_set R pfw A pfi) -> 
    ~ Inn A y.
intros R h1 A h2 y h3. intro h6.
pose proof (min_set_compat R h1 A h2) as h4.
destruct h4 as [h4 h5].
specialize (h5 _ h6).
contradiction.
Qed.


(*Wyckoff*)
Lemma min_set_full_compat : 
  forall (pfdt:type_eq_dec T) (R:relation T) (pfw:well_order R) (pfi:Inhabited (Full_set T)) 
         (y:T),
    {min_set R pfw _ pfi = y} + {R (min_set R pfw _ pfi) y}.
intros h0 R h1 h2 y. pose proof h1 as h1'.
destruct (h0 y (min_set R h1 (Full_set T) h2)) as [h3 | h4].
left; auto.
destruct h1 as [h1a h1b].
red in h1b. 
pose proof (h1b y  (min_set R {| wo_well_founded := h1a; wo_total_strict_order := h1b |}
      (Full_set T) h2)) as h5.
right.
pose proof (wo_antisym).
pose proof (min_set_compat  R
            {| wo_well_founded := h1a; wo_total_strict_order := h1b |}
            (Full_set T) h2) as h6. simpl in h6.
destruct h6 as [h6 h7].
specialize (h7 _ (Full_intro _ y)).
tauto.
Qed.

(*Wyckoff*)
Lemma min_set_full_compat' : 
  forall (pfdt:type_eq_dec T)
         (R:relation T) (pfw:well_order R) (pfi:Inhabited (Full_set T)) 
         (y:T),
    min_set R pfw _ pfi = y \/ R (min_set R pfw _ pfi) y.
intros h0 R h1 h2 y.
destruct (min_set_full_compat h0 R h1 h2 y).
left; auto. right; auto.
Qed.



(*Wyckoff*)
Definition min_elt (R:Relation T) (pfw:well_order R) :=
  min_set R pfw (Full_set T).

(*Wyckoff*)
Lemma ex_wo_max_set_unq : 
  forall (R:Relation T),
    well_order R -> 
    forall (S:Ensemble T),
      ex_max_set R S -> ex_unq_max_set R S.
intros R h1 S h2.
pose proof h1 as h1'.
red in h2. red.
destruct h2 as [m h2].
destruct h2 as [h2 h3].
destruct h1 as [h1a h1b].
red in h1b.
exists m.
red. split. split; auto.
intros x h4. destruct h4 as [h4 h5].
specialize (h1b m x). 
destruct h1b as [h1b | h1b].
specialize (h3 _ h4).
destruct h3 as [h3 | h6].
pose proof (wo_antisym _ h1') as h6.
specialize (h6 _ _ h1b). contradiction.
subst; auto. 
destruct h1b as [h1b | h1b]; auto.
specialize (h5 _ h2).
destruct h5 as [h5 | h6].
pose proof (wo_antisym _ h1') as h6.
specialize (h6 _ _ h5). contradiction.
assumption.
Qed.

(*Wyckoff*)
Definition wo_max_set (R:Relation T) (pfw:well_order R) 
           (S:Ensemble T) (pfe:ex_max_set R S) : T :=
(proj1_sig (constructive_definite_description _ (ex_wo_max_set_unq R pfw S pfe))).



(*Wyckoff*)
Lemma wo_max_set_compat : 
  forall (R:Relation T) (pfw:well_order R) 
         (S:Ensemble T) (pfe:ex_max_set R S),
    let m := wo_max_set R pfw S pfe in 
    Inn S m /\ (forall w : T, Inn S w -> R w m \/ w = m). 
intros R h1 S h2. simpl.
unfold wo_max_set.
destruct constructive_definite_description as [m h3]; simpl.
assumption.
Qed.


(*Wyckoff*)              
Lemma ex_max_set_or_grows : 
  forall (R:Relation T) (pfw:total_strict_order R) (S:Ensemble T),
    {ex_max_set R S} +
    {forall x:T, Inn S x -> Inhabited (afterset' R S x)}.
intros R h0 S.
destruct (classic_dec (ex_max_set R S)) as [h1 | h2].
left; auto. right. 
intros x h3.
unfold ex_max_set in h2.
pose proof (not_ex_all_not _ _ h2 x) as h4. clear h2.
simpl in h4.
apply not_and_or in h4.
destruct h4 as [h4 | h5]. contradiction.  
apply not_all_ex_not in h5.
destruct h5 as [y h5]. 
apply Inhabited_intro with y. constructor.
split; auto.
apply imply_to_and in h5.
destruct h5 as [h5 h6].
apply not_or_and in h6.
destruct h6 as [h6 h7].
red in h0.
specialize (h0 x y).
apply neq_sym in h7.  
tauto.
Qed.

(*Wyckoff*)
Definition wo_max_elt (R:Relation T) (pfw:well_order R) 
           (pfe:ex_max_set R (Full_set T)) :=
  wo_max_set R pfw (Full_set T) pfe.

(*Wyckoff*)
Lemma ex_max_set_or_grows_full :
  forall (R:Relation T) (pfw:total_strict_order R),
    {ex_max_set R (Full_set T)} +
    {forall x:T, Inhabited (afterset' R (Full_set T) x)}.
intros R h0.
destruct (ex_max_set_or_grows R h0 (Full_set _)) as [h1 | h2].
left; auto.
right.
intro x.
apply h2; auto. constructor.
Qed.


End WellOrder.

Implicit Arguments total_strict_order [[T]].
Implicit Arguments well_order [[T]].
Implicit Arguments wo_well_founded [[T] [R]].
Implicit Arguments wo_transitive [[T] [R]].
Implicit Arguments wo_total_strict_order [[T] [R]].
Implicit Arguments wo_irrefl [[T] [R]].
Implicit Arguments wo_antisym [[T] [R]].
Arguments min_elt [T] _ _ _.


Section Covers.
Variable T:Type.

(*Wyckoff*)
(* Named without vowels to avoid name-clashing with 
similar definition in the standard library.*) 
Definition cvrs  {T:Type} (R:Relation T) (x y:T) :=
  R y x /\ ~ (exists z:T, R y z /\ R z x).

(*Wyckoff*)
Definition ex_cover'
  {T:Type} (R:Relation T) (x:T) :=
  exists y:T, cvrs R x y.

(*Wyckoff*)
Definition ex_cover {T:Type} (R:Relation T) (x:T) :=
  exists! y:T, cvrs R x y.

(*Wyckoff*)
Definition covered_elt {T:Type} (R:Relation T) (x:T) (pf:ex_cover R x) := 
  (proj1_sig ((constructive_definite_description _  pf))).

(*Wyckoff*)
Lemma covered_elt_compat : 
  forall (R:Relation T) (x:T) (pf:ex_cover R x),
    let y := covered_elt _ _ pf in
    cvrs R x y.
intros R x pf y. unfold y, covered_elt.
destruct constructive_definite_description. simpl.
assumption.
Qed.

(*Wyckoff*)
Lemma ex_cover_prime_compat :
  forall (R:Relation T),
    total_strict_order R ->
    forall (x:T),
      ex_cover R x <-> ex_cover' R x.
intros R h0 x.
split; intro h1. 
red in h1. destruct h1 as [y [h1 h2]].
red. exists y. assumption.
red in h1. destruct h1 as[y [h1 h2]].
red. exists y. red. split.
split; auto.
intros y' h3. destruct h3 as [h3 h4].
red in h0.
specialize (h0 y y').
destruct h0 as [h0 | h0].
contradict h2. exists y'. split; auto.
destruct h0 as [h0 | h0]. subst; auto.
contradict h4. exists y. split; auto.
Qed.

(*Wyckoff*)
Lemma covered_elt_lt : 
  forall (R:Relation T) (x:T) (pf:ex_cover R x),
    R (covered_elt _ _ pf) x.
intros R x h1.
pose proof (covered_elt_compat _ _ h1) as h2.
destruct h2; auto.
Qed. 

(*Wyckoff*)
Lemma tso_lt_covered_elt : 
  forall (R:Relation T),
    total_strict_order R ->
    forall (a x:T) (pf:ex_cover R x),
      R a x -> a <> covered_elt _ _ pf -> R a (covered_elt _ _ pf).
intros R h0 a x h1 h2 h3. 
red in h0. specialize (h0 a (covered_elt R x h1)).
destruct h0 as [h0 | h0]. assumption.
destruct h0 as [h0 | h0]. contradiction.
pose proof (covered_elt_compat R x h1) as h4. simpl in h4.
destruct h4 as [h4 h5].
pose proof (not_ex_all_not _ (fun z:T =>  R (covered_elt R x h1) z /\ R z x) h5) as h6.
specialize (h6 a). simpl in h6.
contradict h6.
split; auto.
Qed.

(*Wyckoff*)
Lemma not_ex_cover_min_elt : 
  forall (R:relation T) (pfw:well_order R) (pfi:Inhabited (Full_set T)),
    ~ex_cover R (min_elt R pfw pfi).
intros R h1 h2 h3.
red in h3.
destruct h3 as [y h3].
red in h3. destruct h3 as [h3 h4].
red in h3.
destruct h3 as [h3 h5].
pose proof (min_set_compat _ R h1 (Full_set T) h2) as h6.
simpl in h6.
destruct h6 as [h6 h7].
specialize (h7 _ (Full_intro _ y)).
contradiction.
Qed.



(*Wyckoff*)
Inductive ex_cover_n (R:Relation T) : 
  T-> nat -> Prop :=
| ex_cover_n_zero : forall (x:T),
                    ex_cover R x -> ex_cover_n R x 0
| ex_cover_n_S : 
    forall (x:T) (n:nat) (pf:ex_cover R x),
      ex_cover_n R x n ->
      ex_cover_n R (covered_elt _ _ pf) n ->
      ex_cover_n R x (S n).

(*Wyckoff*)
Definition ex_cover_nat (R:Relation T) (x:T) := 
  forall (n:nat),
    ex_cover_n R x n.


(*Wyckoff*)
Definition exists_final_seg (R:Relation T) (x:T) : Prop := 
  exists (n:nat),
    ex_cover_n R x n /\ ~ex_cover_n R x (S n).

(*Wyckoff*)
Lemma ex_cover_n_bound : 
  forall (R:Relation T) (x:T) (n:nat),
    ex_cover_n R x n ->
    forall m:nat, m < n -> ex_cover_n R x m. 
intros R x n h1.
induction h1 as [x h2 | x n h3 h4 h5].
intros; omega.
intros m h6. 
destruct (zerop n) as [h7 | h8]. subst. 
assert (h7:m = 0). omega. subst.
assumption.
destruct (zerop m) as [h9 | h10]. subst. constructor. assumption.
pose proof (S_pred _ _ h10) as h11. rewrite h11 in h6.
apply lt_S_n in h6. 
pose proof (h5 _ h6) as h12.
pose proof (IHh1_1 _ h6) as h13.
pose proof (ex_cover_n_S _ _ _ _ h12 h13) as h14.
rewrite <- (S_pred _ _ h10) in h14.
assumption.
Qed.

(*Wyckoff*)
Definition isolated_from (R:Relation T) (x:T) :=
  [n:nat | ~ex_cover_n R x n].

(*Wyckoff*)
Lemma isolated_from_inhabited : 
  forall (R:Relation T) (x:T), 
    ~ex_cover_nat R x ->
    Inhabited (isolated_from R x).
intros R x h1.
unfold ex_cover_nat in h1.
apply not_all_ex_not in h1.
destruct h1 as [n]. apply Inhabited_intro with n.
constructor. assumption.
Qed.

(*Wyckoff*)
Definition nat_translatable_isolated_from_nat : 
  forall (R:Relation T) (x:T),
    Inhabited (isolated_from R x) ->
    nat_translatable (isolated_from R x).
intros R x h1.
rewrite nat_translatable_iff'; auto.
intros n h2.  
destruct h2 as [h2]. constructor.
intro h3. 
assert (h4:n < S n). auto with arith.
pose proof (ex_cover_n_bound _ _ _ h3 _ h4).
contradiction.
Qed.

(*Wyckoff*)
Definition nat_translatable_isolated_from_nat' : 
  forall (R:Relation T) (x:T),
    Inhabited (isolated_from R x) ->
    nat_translatable' (isolated_from R x).
intros. rewrite <- nat_translatable_prime_compat.
apply nat_translatable_isolated_from_nat; auto.
Qed.

(*Wyckoff*)
Lemma max_ex_cover_n_ex : 
  forall (R:Relation T) (x:T), 
    ~ex_cover_nat R x -> ex_cover R x ->
    exists! (n:nat),
      ex_cover_n R x n /\ ~ex_cover_n R x (S n).
intros R x h2 ha.
pose proof (isolated_from_inhabited R x h2) as h3.
pose proof (nat_translatable_isolated_from_nat _ _ h3) as h4.
apply (nat_translatable_boundary' (ex_cover_n R x)); auto.
constructor; auto.
Qed.

(*Wyckoff*)
Definition max_ex_cover_n (R:Relation T) (x:T)
           (pfnat:~ex_cover_nat R x) (pf:ex_cover R x) :=
  proj1_sig (constructive_definite_description
               _ (max_ex_cover_n_ex R x pfnat pf)).

(*Wyckoff*)
Lemma max_ex_cover_n_compat : 
  forall (R:Relation T) (x:T)
         (pfnat:~ex_cover_nat R x) (pf:ex_cover R x),
    let n := max_ex_cover_n R x pfnat pf in
      ex_cover_n R x n /\ ~ex_cover_n R x (S n).
intros R x h1 h2.
simpl. unfold max_ex_cover_n.
destruct constructive_definite_description as [n h3].
simpl.
assumption.
Qed.

End Covers.

Section WellOrderedInduction.

Definition well_ordered_induction_fun_aux_step
           {T U:Type} (R:T->T->Prop)
           (f_S:T->U->U) (f_sup:{x:T & {y:T | R y x}->U}->U) :
  (forall x : T, (forall y : T, R y x -> U) -> U).
intros x h5. 
destruct (classic_dec (ex_cover R x)) as [h6 | h7]. 
pose proof (covered_elt_lt _ R x h6) as h8.
specialize (h5 _ h8). 
refine (f_S x h5).
refine (f_sup (existT _ x (fun y':{y:T | R y x} =>
                             (h5 _ (proj2_sig y'))))).
Defined.


End WellOrderedInduction.


Section WellOrder'.

(*Wyckoff*)
Lemma total_strict_order_exist_rel :
  forall {T:Type} (R:relation T),
    total_strict_order R ->
    forall (P:T->Prop),
      total_strict_order (exist_rel R P).
intros T R h1 P.
red. intros x y.
destruct x as [x h2], y as [y h3].
unfold exist_rel; simpl.
red in h1. specialize (h1 x y).
destruct h1 as [h1 | h1]. left; auto.
destruct h1 as [h1 | h1]. subst. right. left. apply proj1_sig_injective; auto.
right; auto.
Qed.

(*Wyckoff*)
Lemma well_order_exist_rel :
  forall {T:Type} (R:relation T),
    well_order R ->
    forall (P:T->Prop),
      well_order (exist_rel R P).
intros T R h1 P.
destruct h1 as [h1 h2].
constructor.
apply well_founded_exist_rel; auto.
apply total_strict_order_exist_rel; auto.
Qed.


(*Wyckoff*)
Lemma min_set_exist_rel_eq : 
  forall {T:Type} (R:Relation T) (pfw:well_order R) (b:T)
         (pfi:Inhabited (Full_set T))
         (pfi':Inhabited (Full_set ({t:T | R t b})))
         (pfr: R (min_set _ R pfw _ pfi) b),
    exist  (fun a : T => R a b) _ pfr = 
    min_set _
            (exist_rel R (fun a : T => R a b))
            (well_order_exist_rel R pfw (fun a:T => R a b)) _ pfi'.
intros T R h1 b h2 h3 h4.
pose proof (wo_total_strict_order
              (well_order_exist_rel R h1 (fun a:T => R a b))) as h5.
red in h5.
specialize (h5 (exist (fun a : T => R a b)
                      (min_set _ R h1 (Full_set T) h2) h4)
               (min_set _ (exist_rel R (fun a : T => R a b))
                        (well_order_exist_rel R h1 (fun a : T => R a b))
                        (Full_set {t : T | R t b}) h3)).
destruct h5 as [h5 | h5]. 
pose proof (min_set_compat  _ (exist_rel R (fun a : T => R a b))
            (well_order_exist_rel R h1 (fun a : T => R a b))
            (Full_set {t : T | R t b}) h3) as h6.
simpl in h6.
destruct h6 as [h6 h7].
specialize (h7 (exist (fun a : T => R a b)  _ h4)).
hfirst h7.
constructor.
specialize (h7 hf).
contradiction.
destruct h5 as [h5 | h5]; auto.
pose proof (min_set_compat _ R h1 (Full_set T) h2) as h6.
simpl in h6.
destruct h6 as [h6 h7]. clear h6.
unfold exist_rel in h5. simpl in h5.
apply proj1_sig_injective; simpl.
gtermr.
specialize (h7 y).
hfirst h7.
constructor.
specialize (h7 hf).
contradiction.
Qed.


End WellOrder'.



Section WOArith.
(*Wyckoff*)
Lemma wo_nat : well_order lt.
constructor.
apply lt_wf.
red. intros m n.
pose proof (lt_eq_lt_dec m n); tauto.
Qed.

(*Wyckoff*)
Definition min_set_nat (A:Ensemble nat) (pfi:Inhabited A) :=
  min_set _ lt wo_nat A pfi.

(*Wyckoff*)
Lemma min_set_nat_compat :
  forall (A:Ensemble nat) (pfi:Inhabited A)
         (n:nat),
    Inn A n ->
    Inn A (min_set _ lt wo_nat A pfi) /\
    ~n < min_set _ lt wo_nat A pfi.
  intros A h1 n h2.
  pose proof (min_set_compat _ lt wo_nat A h1) as h3.
  simpl in h3.
  destruct h3 as [h3 h4].
  specialize (h4 n h2).
  split; auto.
Qed.


(*Wyckoff*)
Lemma min_set_nat_nin : 
  forall (A:Ensemble nat) (pfi:Inhabited A),
    {min_set_nat A pfi = 0} + {~Inn A (pred (min_set_nat A pfi))}.
intros A h1.
destruct (zerop (min_set_nat A h1)) as [h2 | h3].
left; auto. right.
intro h4. unfold min_set_nat in h4.
pose proof (min_set_compat _ lt wo_nat A h1) as h5. simpl in h5.
destruct h5 as [h5 h6].
specialize (h6 _ h4).
unfold min_set_nat in h3. omega.
Qed.


(*Wyckoff*)
Lemma min_set_nat_max_set_compat : 
  forall (A:Ensemble nat) (pfi:Inhabited A),
    glb_nat A (min_set_nat A pfi).
intros A h1.
pose proof (min_set_compat _ lt wo_nat A h1) as h3. simpl in h3.
destruct h3 as [h3 h4].
red. split.
red.
intros n h2. 
unfold min_set_nat. 
specialize (h4 _ h2).
omega.
intros b' h5. 
red in h5. 
specialize (h5 _ h3).
assumption.
Qed.


End WOArith.



Section WellOrderMinimum.

Variable T:Type.
Variable R:relation T.
Hypothesis well_ord: well_order R.

(*Schepler*)
Definition WO_minimum:
  forall S:Ensemble T, Inhabited S ->
    { x:T | Inn S x /\ forall y:T, Inn S y -> y = x \/ R x y }.
refine (fun S H => constructive_definite_description _ _).
pose proof (WF_implies_MEP T R (wo_well_founded well_ord)).
unfold minimal_element_property in H0.
pose proof (H0 S H).

destruct H1.
destruct H1.
exists x.
red.
split.
split.
assumption.
intros.
case (wo_total_strict_order well_ord x y).
tauto.
intro.
case H4.
auto.
intro.
contradict H5.
auto.

intros.
destruct H3.
case (wo_total_strict_order well_ord x x').
intro.
pose proof (H4 x H1).
case H6.
trivial.
intro.
contradict H7.
auto.

intro.
case H5.
trivial.
intro.
contradict H6.
auto.
Defined.

End WellOrderMinimum.

Implicit Arguments WO_minimum [[T]].

Section WellOrderConstruction.

Variable T:Type.

(*Schepler*)
Definition restriction_relation (R:relation T) (S:Ensemble T) :
  relation ({z:T | Inn S z}) :=
  fun (x y:{z:T | Inn S z}) => R (proj1_sig x) (proj1_sig y).

Record partial_WO : Type := {
  pwo_S: Ensemble T;
  pwo_R: relation T;
  pwo_R_lives_on_S: forall (x y:T),
                      pwo_R x y ->
                      Inn pwo_S x /\ Inn pwo_S y;
  pwo_wo: well_order (restriction_relation pwo_R pwo_S)
}.

(*Schepler*)
(*The last condition below says that S WO1 is a downward closed
  subset of S WO2 *)
Record partial_WO_ord (WO1 WO2:partial_WO) : Prop := {
  pwo_S_incl: Included (pwo_S WO1) (pwo_S WO2);
  pwo_restriction: forall x y:T,
                     Inn (pwo_S WO1) x ->
                     Inn (pwo_S WO1) y ->
    (pwo_R WO1 x y <-> pwo_R WO2 x y);
  pwo_downward_closed: forall x y:T,
                         Inn (pwo_S WO1) y -> Inn (pwo_S WO2) x ->
    pwo_R WO2 x y -> Inn (pwo_S WO1) x
}.

(*Schepler*)
Lemma partial_WO_preord : preorder partial_WO_ord.
Proof.
constructor.
unfold reflexive.
intro.
destruct x.
constructor; simpl.
auto with sets.
split.
trivial.
trivial.
auto.

unfold transitive.
destruct x.   
destruct y.
destruct z.
intros.
destruct H.
destruct H0.

simpl in pwo_S_incl0; simpl in pwo_restriction0;
  simpl in pwo_downward_closed0; simpl in pwo_S_incl1;
  simpl in pwo_restriction1; simpl in pwo_downward_closed1.
constructor; simpl.
auto with sets.
intros.
apply iff_trans with (pwo_R1 x y).
apply pwo_restriction0; auto with sets.
apply pwo_restriction1; auto with sets.

intros.
apply pwo_downward_closed0 with y; trivial.
apply pwo_downward_closed1 with y; trivial.
auto with sets.
apply <- (pwo_restriction1 x y); trivial.
auto with sets.
apply pwo_downward_closed1 with y; trivial.
auto with sets.
Qed.

End WellOrderConstruction.

Section WO_implies_AC.

(*Schepler*)
Lemma WO_implies_AC: forall (A B:Type) (R: A -> B -> Prop)
  (WO:relation B), well_order WO ->
  (forall x:A, exists y:B, R x y) ->
  exists f:A->B, forall x:A, R x (f x).
Proof.
intros.
assert (forall a:A, Inhabited [ b:B | R a b ]).
intro.
pose proof (H0 a).
destruct H1.
exists x.
constructor; assumption.

exists (fun a:A => proj1_sig
  (WO_minimum WO H [ b:B | R a b ] (H1 a))).
intro.
destruct @WO_minimum.
simpl.
destruct a.
destruct H2.
assumption.
Qed.

End WO_implies_AC.

